#include "MatchingCurve.h"

#include "MappingProjectionParameters.h"

MatchingCurve::
MatchingCurve()
{ 
  curve=NULL; 
  projectionParameters=NULL; 
  numberOfLinesForNormalBlend=3;
  gridLine=-100;  // put bogus value
  curvePosition=0.;
  curveDirection=0;
} 

MatchingCurve::
~MatchingCurve()
{
  if( curve!=NULL && curve->decrementReferenceCount()==0 )
    delete curve;
  delete projectionParameters;
}


MatchingCurve::
MatchingCurve(const MatchingCurve & mc)
{
  curve=NULL; 
  projectionParameters=NULL; 
  numberOfLinesForNormalBlend=3;
  gridLine=-100;  // put bogus value
  curvePosition=0.;
  curveDirection=0;

  *this=mc;
}

void MatchingCurve::
setCurve( Mapping & curve_ )
{
  curve=&curve_;
  if(  projectionParameters==NULL )
    projectionParameters=new MappingProjectionParameters;
}

MatchingCurve& MatchingCurve::
operator=(const MatchingCurve & mc )
{
  x[0]=mc.x[0]; x[1]=mc.x[1]; x[2]=mc.x[2];
  curvePosition=mc.curvePosition; 
  gridLine=mc.gridLine;
  curveDirection=mc.curveDirection; 

  if( curve!=NULL && curve->decrementReferenceCount()==0 )
    delete curve;
  curve=mc.curve;     
  if( curve!=NULL ) curve->incrementReferenceCount();
  
  // make a deep copy of the projectionParameters so we don't have to worry about ref counting.
  if( mc.projectionParameters!=NULL )
  {
    if( projectionParameters==NULL )
      projectionParameters=new MappingProjectionParameters;

    // *projectionParameters=*mc.projectionParameters;   // do we need to copy??
  }
  
  numberOfLinesForNormalBlend=mc.numberOfLinesForNormalBlend;  
  return *this;
}

