#include "Ogmg.h"
#include "SparseRep.h"

#define  FOR_3(i1,i2,i3,I1,I2,I3)\
  I1Base=I1.getBase(); I2Base=I2.getBase(); I3Base=I3.getBase();\
  I1Bound=I1.getBound(); I2Bound=I2.getBound(); I3Bound=I3.getBound();\
  for( i3=I3Base; i3<=I3Bound; i3++ )  \
  for( i2=I2Base; i2<=I2Bound; i2++ )  \
  for( i1=I1Base; i1<=I1Bound; i1++ )

#define FOR_M()\
 for( int m3=-1; m3<=m3b; m3++ )\
 for( int m2=-1; m2<=1; m2++ )\
 for( int m1=-1; m1<=1; m1++ )


void Ogmg::
assignBoundaryConditionCoefficients( realMappedGridFunction & coeff, int grid, int level, 
                                     int sideToCheck /* =-1 */, int axisToCheck /* =-1 */ )
// =============================================================================================
// /Description:
//     Fill the boundary conditions into the coefficient matrix.
//  This function is called by buildPredefinedCoefficientMatrix and operatorAveraging.
//  
// =============================================================================================
{
  if( debug & 4 )
    fprintf(debugFile,"\n +++++++++++++++ assignBoundaryConditionCoefficients (side,axis)=(%i,%i) ++++++++\n",
	   sideToCheck,axisToCheck);
  
  MappedGrid & mg = mgcg.multigridLevel[level][grid];
//    realArray & coeffa = coeff;
  MappedGridOperators & op = * coeff.getOperators();
  const int numberOfDimensions=mg.numberOfDimensions();

  Index Iv[3], &I1=Iv[0], &I2=Iv[1], I3=Iv[2];
  Index Jv[3], &J1=Jv[0], &J2=Jv[1], J3=Jv[2];
  Index Ig1,Ig2,Ig3;

  RealArray & a = bcParams.a;
  a.redim(2);

  const int orderOfExtrapolation= orderOfAccuracy==2 ? 3 : 4;  // 5 **** use extrap order 4 for 4th order
    
  BoundaryConditionParameters extrapParams;
  extrapParams.orderOfExtrapolation=orderOfExtrapolation; // orderOfAccuracy+1; 

  // const IntegerArray & bc = mg.boundaryCondition();
  int isv[3], &is1=isv[0], &is2=isv[1], &is3=isv[2];

  const int axisStart=axisToCheck==-1 ? 0 : axisToCheck;
  const int axisEnd  =axisToCheck==-1 ? mgcg.numberOfDimensions()-1 : axisToCheck;

  const int sideStart=sideToCheck==-1 ? 0 : sideToCheck;
  const int sideEnd  =sideToCheck==-1 ? 1 : sideToCheck;


  for( int axis=axisStart; axis<=axisEnd; axis++ )
  {
    for( int side=sideStart; side<=sideEnd; side++ )
    {
      if( mg.boundaryCondition(side,axis)<=0 )
	continue;
	
	
      if( bc(side,axis,grid)==OgmgParameters::dirichlet )
      {
//            if( orderOfAccuracy==2 || 
//  	      // *wdh* 030521 level==(mgcg.numberOfMultigridLevels()-1) ||
//                 parameters.fourthOrderBoundaryConditionOption==0 ||
//                 (orderOfAccuracy==4 && level!=0 && !parameters.useEquationForDirichletOnLowerLevels) )

	bool useEquationOnGhost = useEquationOnGhostLineForDirichletBC(mg,level);

	if( orderOfAccuracy==2 || !useEquationOnGhost )
	{
	  coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::boundary1+side+2*axis,
						   extrapParams);
	}
	else
	{
	  // === Use the equation to 2nd order on the boundary ===

	  if( equationToSolve==OgesParameters::laplaceEquation )
	  {
	    getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
	    op.setOrderOfAccuracy(2);

	    // ***this is wrong*** --> corners
	    // ** op.coefficients(MappedGridOperators::laplacianOperator,coeff,I1,I2,I3); // efficient version

           
	    // ***** 030606: to fix: what if there are interp points on the edge --> we need to extrap points
	    // adjacent to them *****


	    realArray tempCoeff(coeff.dimension(0),I1,I2,I3);
	    op.assignCoefficients(MappedGridOperators::laplacianOperator,tempCoeff,I1,I2,I3); // efficient version
	    op.setOrderOfAccuracy(4);
	    getGhostIndex(mg.gridIndexRange(),side,axis,Ig1,Ig2,Ig3);

              
	    const int m3b= mg.numberOfDimensions()==2 ? -1 : 1;
	    const int ee=0; 
	    int I1Base,I2Base,I3Base;
	    int I1Bound,I2Bound,I3Bound;
	    int i1,i2,i3;
	    is1=is2=is3=0;
	    isv[axis]=1-2*side;
	      
	    FOR_M()
	    {
	      // copy the second order equation into the correct positions of the 4th order stencil
	      const int index2=(m1+1)+3*(m2+1+3*(m3+1));  // stencil width=3

	      int index4=mg.numberOfDimensions()==2 ? (m1+2)+5*(m2+2)          :        // stencil width==4
		(m1+2)+5*(m2+2+5*(m3+2));
	      // remember we are shifted to the ghost line :
	      index4 += axis==0 ? 1-2*side : axis==1 ? 5*(1-2*side) : 25*(1-2*side);
		
	      FOR_3(i1,i2,i3,I1,I2,I3)
	      {
		int ig1=i1-is1, ig2=i2-is2, ig3=i3-is3;
		coeff(index4,ig1,ig2,ig3)=tempCoeff(index2,i1,i2,i3);
		coeff.sparse->setClassify(SparseRepForMGF::ghost1,ig1,ig2,ig3,ee);
	      }
		
	    }

	    if( debug & 8 )
	    {
	      printf("\n>>>>>>>>>>>>>assignBoundaryConditionCoefficients: \n"
		     " Dirichlet: fill in 2nd-order equation on the ghost points : level=%i, grid=%i side=%i "
		     "axis=%i\n",level,grid,side,axis);
	      // coeff(all,Ig1,Ig2,Ig3).display("Eqn to 2nd order on the boundary");
	    }
	      
	    // **** end points on dirichlet sides: use extrapolation (otherwise the same eqn appears twice!)

	    for( int dir=0; dir<mg.numberOfDimensions()-1; dir++ )
	    {
	      const int axisp = (axis+dir+1) % mg.numberOfDimensions(); // adjacent side
	      for( int side2=0; side2<=1; side2++ )
	      {
		if( bc(side2,axisp,grid)==OgmgParameters::dirichlet )
		{
		  if( mg.boundaryCondition(side2,axisp)<=0 )
		  {
		    display(bc,"bc");
		    display(mg.boundaryCondition(),"mg.boundaryCondition()");
		    Overture::abort("Unepected Error");
		  }
		    
		  J1=Ig1, J2=Ig2, J3=Ig3;
		  Jv[axisp]= side2==0 ? Jv[axisp].getBase() : Jv[axisp].getBound();
		  // extrapolate points  coeff(.,J1,J2,J3) in the direction axis
		  op.setExtrapolationCoefficients(coeff,ee,J1,J2,J3,orderOfExtrapolation); // in GenericMGOP
		}
	      }
	    }
	      

	  }
	  else
	  {
	    Overture::abort();
	  }
	    
	}

	coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::dirichlet,BCTypes::boundary1+side+2*axis);
	if( orderOfAccuracy==4 )
	{
	  extrapParams.ghostLineToAssign=2;
	  coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::boundary1+side+2*axis,
						   extrapParams); 
	  extrapParams.ghostLineToAssign=1;
	}
      }
      else if( bc(side,axis,grid)==OgmgParameters::neumann )
      {
        if( orderOfAccuracy==4 &&
	    level>0 && parameters.lowerLevelNeumannFirstGhostLineBC==OgmgParameters::useEquationToSecondOrder )
	{
	  // apply symmetry on two lines since this corresponds to the homogeneous Neumann BC to second order
          printf(" ****** assignBCCoeff: level=%i use double symmetry for neuman BC *****\n",level);
	  
	  extrapParams.ghostLineToAssign=1;
          coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::evenSymmetry,BCTypes::boundary1+side+2*axis,
						   extrapParams); 
	  extrapParams.ghostLineToAssign=2;
          coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::evenSymmetry,BCTypes::boundary1+side+2*axis,
						   extrapParams); 
          extrapParams.ghostLineToAssign=1;
	}
	else
	{
	  coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::neumann,BCTypes::boundary1+side+2*axis);
	  if( orderOfAccuracy==4 )
	  {
	    bool useEquationOnGhost = useEquationOnGhostLineForNeumannBC(mg,level);
	    if( useEquationOnGhost )
	    {
	      printf(" ******assignBoundaryConditionCoefficients: WARNING: not using eqn on ghost for neumann BC"
		     " at level =%i ******\n",level);
	    }
	    extrapParams.ghostLineToAssign=2;
	    if( false )
	    {
	      coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::boundary1+side+2*axis,
						       extrapParams); 
	    }
	    else
	    {
	      coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::evenSymmetry,BCTypes::boundary1+side+2*axis,
						       extrapParams); 
	    }
	  }
	}
	extrapParams.ghostLineToAssign=1; // reset

	// getGhostIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);

	// Range all;
	// display(coeff(all,I1,I2,I3),"coeff on ghost line after adding a neumann BC",debugFile,"%8.2e ");
	  
      }
      else if( bc(side,axis,grid)==OgmgParameters::mixed )
      {
	a(0)=boundaryConditionData(0,side,axis,grid);  // coeff of u
	a(1)=boundaryConditionData(1,side,axis,grid);  // coeff of du/dn
	printf(" predefined: mixed BC: a0=%e a1=%e \n",a(0),a(1));
	  
	if( orderOfAccuracy==4 &&
	    level>0 && parameters.lowerLevelNeumannFirstGhostLineBC==OgmgParameters::useEquationToSecondOrder )
	{
	  // apply mixed BC to second order -- this will reduce to an even symmetry condition if mixed==neumann
          printf(" ****** assignBCCoeff: level=%i use mixedSecondOrder for mixed BC *****\n",level);
	  
	  getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
              
          // don't apply on end points if adjacent face is a dirichlet BC
          int dir;
	  for( dir=0; dir<numberOfDimensions; dir++ )
	  {
	    if( dir!=axis )
	    {
	      int na=Iv[dir].getBase();
	      int nb=Iv[dir].getBound();

	      if( mg.boundaryCondition(0,dir)>0  && boundaryCondition(0,dir,grid)==OgmgParameters::extrapolate ) na++;
	      if( mg.boundaryCondition(1,dir)>0  && boundaryCondition(1,dir,grid)==OgmgParameters::extrapolate ) nb--;
	      Iv[dir]=Range(na,nb);
	    }
	  }

          const real a0=a(0);
	  const real a1=a(1);

	  const int ee=0; 
	  int I1Base,I2Base,I3Base;
	  int I1Bound,I2Bound,I3Bound;
	  int i1,i2,i3;
	  is1=is2=is3=0;
	  isv[axis]=1-2*side;
         
          real dx[3]={1.,1.,1.};//
          real dr[3]={1.,1.,1.};//
          real t1;
          const bool isRectangular = mg.isRectangular();
	  if( isRectangular )
	  {
	    mg.getDeltaX(dx);
	    t1=(2.*dx[axis]*(a0/a1));
	  }
	  else
	  {
	    dr[0]=mg.gridSpacing(0);
	    dr[1]=mg.gridSpacing(1);
	    dr[2]=mg.gridSpacing(2);
	  }
	  
          const realArray & rsxy = isRectangular ? coeff : mg.inverseVertexDerivative();
#define RX(i1,i2,i3,m,n) rsxy(i1,i2,i3,(m)+numberOfDimensions*(n))

	  FOR_3(i1,i2,i3,I1,I2,I3) // **** could optimise this loop *****
	  {
            if( !isRectangular )
	    {
	      if( numberOfDimensions==2 )
	      {
                t1=2.*dr[axis]*(a0/(a1*sqrt(SQR(RX(i1,i2,i3,axis,0))+ SQR(RX(i1,i2,i3,axis,1)) )));
	      }
	      else 
	      {
                t1=2.*dr[axis]*(a0/(a1*sqrt(SQR(RX(i1,i2,i3,axis,0))+ 
                                            SQR(RX(i1,i2,i3,axis,1))+
                                            SQR(RX(i1,i2,i3,axis,2)) )));
	      }
	    }
#undef RX	    
//      u(i1-  is1,i2-  is2,i3-  is3)=u(i1+  is1,i2+  is2,i3+  is3)-   t1
//      u(i1-2*is1,i2-2*is2,i3-2*is3)=u(i1+2*is1,i2+2*is2,i3+2*is3)-2.*t1

            // first ghost line
	    int ig1=i1-is1, ig2=i2-is2, ig3=i3-is3;
	    coeff(0,ig1,ig2,ig3)= 1.;
	    coeff(1,ig1,ig2,ig3)= t1;
	    coeff(2,ig1,ig2,ig3)=-1.;
	    coeff.sparse->setCoefficientIndex(0, ee,ig1,ig2,ig3, ee,ig1,ig2,ig3);
	    coeff.sparse->setCoefficientIndex(1, ee,ig1,ig2,ig3, ee,i1,i2,i3);
	    coeff.sparse->setCoefficientIndex(2, ee,ig1,ig2,ig3, ee,i1+is1,i2+is2,i3+is3);

            // second ghost line
            ig1=i1-2*is1, ig2=i2-2*is2, ig3=i3-2*is3;
	    coeff(0,ig1,ig2,ig3)= 1.;
	    coeff(1,ig1,ig2,ig3)= 2.*t1;
	    coeff(2,ig1,ig2,ig3)=-1.;
	    coeff.sparse->setCoefficientIndex(0, ee,ig1,ig2,ig3, ee,ig1,ig2,ig3);
	    coeff.sparse->setCoefficientIndex(1, ee,ig1,ig2,ig3, ee,i1,i2,i3);
	    coeff.sparse->setCoefficientIndex(2, ee,ig1,ig2,ig3, ee,i1+2*is1,i2+2*is2,i3+2*is3);
	  }

	  // **** end points on dirichlet sides: use extrapolation (otherwise the same eqn appears twice!)
          getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);
	  for( int dir=0; dir<numberOfDimensions-1; dir++ )
	  {
	    const int axisp = (axis+dir+1) % numberOfDimensions; // adjacent side
	    for( int side2=0; side2<=1; side2++ )
	    {
	      if( boundaryCondition(side2,axisp,grid)==OgmgParameters::extrapolate ) // bc(side2,axisp,grid)==dirichlet )
	      {
		if( mg.boundaryCondition(side2,axisp)<=0 )
		{
		  display(bc,"bc");
		  display(mg.boundaryCondition(),"mg.boundaryCondition()");
		  Overture::abort("Unepected Error");
		}
		    
                J1=I1, J2=I2, J3=I3;
      	        Jv[axisp]= side2==0 ? Jv[axisp].getBase() : Jv[axisp].getBound();
                for( int ghost=1; ghost<=2; ghost++ )
		{
		  Jv[axis]=Iv[axis]-ghost*isv[axis];

		  // extrapolate points  coeff(.,J1,J2,J3) in the direction axis
		  op.setExtrapolationCoefficients(coeff,ee,J1,J2,J3,orderOfExtrapolation); // in GenericMGOP

		}
	      }
	    }
	  }



	}
	else
	{

	  coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::mixed,BCTypes::boundary1+side+2*axis,bcParams);

	  if( orderOfAccuracy==4 )
	  {
	    bool useEquationOnGhost = useEquationOnGhostLineForNeumannBC(mg,level);
	    if( useEquationOnGhost )
	    {
	      // *** fix this *** use mixedToSecondOrder instead of the neumann condition above and the symmetry below

	      printf(" ******assignBoundaryConditionCoefficients: WARNING: not using eqn on ghost for mixed BC"
		     " at level =%i ******\n",level);
	    }

	    extrapParams.ghostLineToAssign=2;
	    if( level==0 || !parameters.useSymmetryForNeumannOnLowerLevels ) 
	    {
	      extrapParams.orderOfExtrapolation=4;
	      coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::boundary1+side+2*axis,
						       extrapParams); 
	    }
	    else
	    {
	      coeff.applyBoundaryConditionCoefficients(0,0,BCTypes::evenSymmetry,BCTypes::boundary1+side+2*axis,
						       extrapParams); 
	    }
	  }
	}
	
	extrapParams.ghostLineToAssign=1;
	  
      }
      else if( bc(side,axis,grid)>0 )
      {
	printf("Ogmg::assignBoundaryConditionCoefficients:ERROR: unknown bc=%i for grid=%i side=%i axis=%i\n",
	       bc(side,axis,grid),grid,side,axis);
	throw "error";
      }
    }
  }


}
