#include "Ogmg.h"
#include "OGPolyFunction.h"
#include "OGTrigFunction.h"
#include "display.h"
#include "PlotStuff.h"
#include "ParallelUtility.h"

#undef ForBoundary
#define ForBoundary(side,axis)  \
       for( axis=0; axis<cg.numberOfDimensions(); axis++ ) \
         for( side=0; side<=1; side++ )

int Ogmg::
coarseGridSolverTest()
// ===============================================================================================
// test the accuracy of the coarse grid equations
// ===============================================================================================
{
  const int level = mgcg.numberOfMultigridLevels()-1;  // coarse level
  
  realCompositeGridFunction & u = uMG.multigridLevel[level];
  realCompositeGridFunction & f = fMG.multigridLevel[level];
  CompositeGrid & cg = *u.getCompositeGrid();

  int twilightZone=2;  // 1=trig, 2=poly
  
  real fx=1., fy=1., fz=1.;
  OGTrigFunction tzTrig(fx,fy,fz);  // create an exact solution

  if( directSolver.isSolverIterative() ) 
  {
    real tol=REAL_EPSILON*100.;
    directSolver.set(OgesParameters::THEpreconditioner,OgesParameters::incompleteLUPreconditioner);
    directSolver.set(OgesParameters::THErelativeTolerance,max(tol,REAL_EPSILON*10.));
    directSolver.set(OgesParameters::THEmaximumNumberOfIterations,10000);
  } 


  const int maxDegree= twilightZone==2 ? orderOfAccuracy : 0;
  for( int degree=0; degree<=maxDegree; degree++ )
  {
    printf("\n-------------------------------------------------------------------------------------\n");
    if( twilightZone==2 )
      cout << "test a polynomial of degree " << degree << "...\n";

    int degreeOfSpacePolynomial=degree;
  
    OGPolyFunction tzPoly(degreeOfSpacePolynomial,cg.numberOfDimensions());      // create an exact solution
    OGFunction & tz = twilightZone==1 ? (OGFunction&)tzTrig : (OGFunction&)tzPoly;

    const int numberOfComponents=1;
  
    RealArray spatialCoefficientsForTZ(5,5,5,numberOfComponents);  
    spatialCoefficientsForTZ=0.;
    RealArray timeCoefficientsForTZ(5,numberOfComponents);      
    timeCoefficientsForTZ=0.;
    timeCoefficientsForTZ(0,0)=1.;
    if( degreeOfSpacePolynomial==0 )
    {
      spatialCoefficientsForTZ(0,0,0)=1.; 
    }
    else if( degreeOfSpacePolynomial==2 )
    {
      spatialCoefficientsForTZ(1,0,0)=-.5;
      spatialCoefficientsForTZ(0,1,0)=-.5;
      spatialCoefficientsForTZ(2,0,0)=1.;  // x^2
      spatialCoefficientsForTZ(0,2,0)=1.;
      if( cg.numberOfDimensions()==3 )
      {
	spatialCoefficientsForTZ(0,0,1)=-.5;
	spatialCoefficientsForTZ(0,0,2)=.5;
      }
    }
    else if( degreeOfSpacePolynomial==1 )
    {
      spatialCoefficientsForTZ(1,0,0)=1.; 
      spatialCoefficientsForTZ(0,1,0)=1.;
      if( cg.numberOfDimensions()==3 )
	spatialCoefficientsForTZ(0,0,1)=-.25;
    }
    else if( degreeOfSpacePolynomial==4 )
    {
      spatialCoefficientsForTZ(1,0,0)=-.5;
      spatialCoefficientsForTZ(0,1,0)= .5;

      spatialCoefficientsForTZ(2,0,0)= .3;  // x^2
      spatialCoefficientsForTZ(0,2,0)=-.3;

      spatialCoefficientsForTZ(3,1,0)= .1;  // x^2
      spatialCoefficientsForTZ(1,3,0)=-.1;

      spatialCoefficientsForTZ(4,0,0)=.2;  // x^4
      spatialCoefficientsForTZ(0,4,0)=.1;  // y^4

      if( cg.numberOfDimensions()==3 )
      {
	spatialCoefficientsForTZ(0,0,1)= .6;
	spatialCoefficientsForTZ(0,0,2)=-.3;
	spatialCoefficientsForTZ(0,0,4)= .1;

      }
    
    }
    else
    {
      spatialCoefficientsForTZ(0,0,0)=1.; 
    }
    
    tzPoly.setCoefficients( spatialCoefficientsForTZ,timeCoefficientsForTZ );  


    Index I1,I2,I3;
    Index I1b,I2b,I3b,I1g,I2g,I3g;

  // assign the rhs for all the equations (including BC's)
    int grid;
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      mg.update(MappedGrid::THEcenter | MappedGrid::THEvertex );
    
      getIndex(mg.dimension(),I1,I2,I3);
      if( equationToSolve==OgesParameters::heatEquationOperator )
      {
	f[grid](I1,I2,I3)=equationCoefficients(0,grid)*tz(mg,I1,I2,I3)+
	  equationCoefficients(1,grid)*tz.laplacian(mg,I1,I2,I3); 
      }
      else
      {
	f[grid](I1,I2,I3)=tz.laplacian(mg,I1,I2,I3);
      }
      
      // Assign Boundary Conditions 
      int side,axis;
      ForBoundary(side,axis)
      {
	getBoundaryIndex(mg.gridIndexRange(),side,axis,I1b,I2b,I3b);
	getGhostIndex   (mg.gridIndexRange(),side,axis,I1g,I2g,I3g);
	if( mg.boundaryCondition(side,axis)==OgmgParameters::dirichlet )
	{
	  f[grid](I1b,I2b,I3b)=tz(mg,I1b,I2b,I3b);
	  f[grid](I1g,I2g,I3g)=0.; // extrap
	}
	else
	{
	  if( mg.isRectangular() )
	  {
	    if( axis==axis1 )
	      f[grid](I1g,I2g,I3g)=tz.x(mg,I1b,I2b,I3b)*(2*side-1.); // normal is 2*side-1
	    else if( axis==axis2 )
	      f[grid](I1g,I2g,I3g)=tz.y(mg,I1b,I2b,I3b)*(2*side-1.);
	    else
	      f[grid](I1g,I2g,I3g)=tz.z(mg,I1b,I2b,I3b)*(2*side-1.);
	  }
	  else
	  {
	    mg.update(MappedGrid::THEvertexBoundaryNormal );
	    RealDistributedArray & normal = mg.vertexBoundaryNormal(side,axis);
	    if( cg.numberOfDimensions()==2 )
	      f[grid](I1g,I2g,I3g)=normal(I1b,I2b,I3b,0)*tz.x(mg,I1b,I2b,I3b)
		+normal(I1b,I2b,I3b,1)*tz.y(mg,I1b,I2b,I3b);
	    else 
	      f[grid](I1g,I2g,I3g)=normal(I1b,I2b,I3b,0)*tz.x(mg,I1b,I2b,I3b)
		+normal(I1b,I2b,I3b,1)*tz.y(mg,I1b,I2b,I3b)
		+normal(I1b,I2b,I3b,2)*tz.z(mg,I1b,I2b,I3b);
	  }
	  
	  if( mg.boundaryCondition(side,axis)==OgmgParameters::mixed )
	  {
	    f[grid](I1g,I2g,I3g)=bcParams.a(0)*tz(mg,I1b,I2b,I3b)+bcParams.a(1)*f[grid](I1g,I2g,I3g);
	  }
	}
      }
    }
  

    u=0.;   // initial guess

  // ************************
  // ********Solve***********
  // ************************
    directSolver.solve( u,f );

    printf("coarseGridSolver: max residual=%8.2e (iterations=%i) ***\n",directSolver.getMaximumResidual(),
	   directSolver.getNumberOfIterations());

    RealCompositeGridFunction e(cg);
    e=0.;
    RealArray error(cg.numberOfComponentGrids());
    error=0.;
    real uMax=0.;

    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )     
    {
      MappedGrid & mg = cg[grid];
      
      getIndex(mg.gridIndexRange(),I1,I2,I3,1);         // include ghost points 
      const realArray & uExact=tz(mg,I1,I2,I3);
      where( mg.mask()(I1,I2,I3)!=0 )
      {
	uMax=max(uMax,max(fabs(uExact)));
	
	e[grid](I1,I2,I3)=fabs(u[grid](I1,I2,I3)-uExact);
	error(grid)=max(fabs(e[grid](I1,I2,I3)));
      }
    }

    if( twilightZone==2 )
    {
      printf(" coarseGridSolverTest: exact solution is a polynomial of degree %i\n",degreeOfSpacePolynomial);
    }
    else
    {
      printf(" coarseGridSolverTest: exact solution is a trigonometric polynomial\n");
    }
  
    for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )     
    {
      printf("coarseGridSolverTest: Maximum error on grid: %15s = %8.2e, max relative error=%8.2e\n",
	     (const char *)cg[grid].mapping().getName(Mapping::mappingName),error(grid),error(grid)/uMax);
    }
    printf("-------------------------------------------------------------------------------------\n");
    
  }
  
//      if( Ogmg::debug & 8 )
//        e.display("error including ghost points","%6.2e ");


  return 0;
}








//--------------------------------------------------------------------------------------
//  Test the smoothers
//--------------------------------------------------------------------------------------
int Ogmg::
smoothTest(GenericGraphicsInterface & ps, int plotOption )
{

//    uMG.reference(u0);  
//    fMG.reference(f0);
  realCompositeGridFunction & u0 = uMG;
  realCompositeGridFunction & f0 = fMG;

  printF("\n"
         " ============== Ogmg::smoothTest =================\n");

//   printf("u0.numberOfGrids = %i \n",u0.numberOfGrids());
//   printf("u0.multigridLevel[0].numberOfGrids = %i \n",u0.multigridLevel[0].numberOfGrids());

//   printf("uMG.numberOfGrids = %i \n",uMG.numberOfGrids());
//   printf("uMG.multigridLevel[0].numberOfGrids = %i \n",uMG.multigridLevel[0].numberOfGrids());

  parameters.numberOfSubSmooths=1;
  

  for( int level=0; level<mgcg.numberOfMultigridLevels()-1; level++ )
  {
    cout << "====================Level = " << level << "======================\n";
    realCompositeGridFunction & u = u0.multigridLevel[level];
    realCompositeGridFunction & f = f0.multigridLevel[level];
    realCompositeGridFunction & df = defectMG.multigridLevel[level];
    CompositeGrid & cg = mgcg.multigridLevel[level];
    
    u=0.;  // initial guess
    f=1.;  
    Index I1,I2,I3;
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      getIndex(mg.dimension(),I1,I2,I3);
      bool ok=true;
      #ifdef USE_PPP
	realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u[grid],uLocal);
	realSerialArray fLocal; getLocalArrayWithGhostBoundaries(f[grid],fLocal);
	intSerialArray maskLocal; getLocalArrayWithGhostBoundaries(mg.mask(),maskLocal);
	ok=ParallelUtility::getLocalArrayBounds(u[grid],uLocal,I1,I2,I3,1);
      #else
	realSerialArray & uLocal = u[grid];
	realSerialArray & fLocal = f[grid];
	intSerialArray & maskLocal = mg.mask();
      #endif

      if( ok )
      {
	where( maskLocal(I1,I2,I3)==0 )
	{
	  uLocal(I1,I2,I3)=123456789.;  // put bogus values at unused points
	}
      }
      
      // fill in the BC's into f
      for( int axis=axis1; axis<cg.numberOfDimensions(); axis++ )
      {
	for( int side=Start; side<=End; side++ )
	{
	  if( mg.boundaryCondition(side,axis) ==OgmgParameters::dirichlet  )
	  {
	    getBoundaryIndex(mg.extendedIndexRange(),side,axis,I1,I2,I3);
	    ok=ParallelUtility::getLocalArrayBounds(u[grid],uLocal,I1,I2,I3,1);
	    if( !ok ) continue;
	    fLocal(I1,I2,I3)=0.;   // NOTE: must be zero to work at level > 0
	  }
	  else if( mg.boundaryCondition(side,axis) ==OgmgParameters::neumann )  
	  {
	    getGhostIndex(mg.extendedIndexRange(),side,axis,I1,I2,I3);
	    ok=ParallelUtility::getLocalArrayBounds(u[grid],uLocal,I1,I2,I3,1);
	    if( !ok ) continue;
	    fLocal(I1,I2,I3)=0.;   // NOTE: must be zero to work at level > 0
	  }
	  else if( mg.boundaryCondition()(side,axis)>0 )
	  {
	    cout << "smoothTest::ERROR: unknown BC \n";
	    throw "error";
	  }
	}
      }
      if( Ogmg::debug & 8 )
	display(f[grid],"Here is f for the smooth test",Ogmg::debugFile);

    }
    
    defect(level);
    real maximumDefectOld=max(df);
    int numIterations=25;
    for( int it=0; it<numIterations; it++ )
    {
      int numberOfSmoothingSteps=parameters.numberOfSmooths(0,level)+parameters.numberOfSmooths(1,level);

      smooth(level,numberOfSmoothingSteps,it);

      defect(level);

      real maximumDefect=max(fabs(df));

      if( Ogmg::debug & 32 )
        df.display("defect");

      printF(" smoothTest: it= %4i,  max(defect) = %8.4e, defect/defectOld = %6.3f\n",
	     it,maximumDefect,maximumDefect/(maximumDefectOld+FLT_MIN));

      maximumDefectOld=maximumDefect;
      if( maximumDefect<1.e-5 )
        break;

      if( plotOption )
      {
        PlotStuffParameters psp;

	aString answer;
	aString menu[]=
	{
	  "solution",
	  "error",
	  "defect",
	  "grid",
	  "rhs (for Ogmg)",
	  "exit",
	  ""
	};
    
	for( ;; )
	{
	  ps.getMenuItem(menu,answer,"choose an option");
	  if( answer=="exit" )
	  {
	    break;
	  }
	  else if( answer=="solution" )
	  {
	    psp.set(GI_TOP_LABEL,"Solution u"); 
	    PlotIt::contour(ps,u,psp);
	  }
	  else if( answer=="defect" )
	  {
	    psp.set(GI_TOP_LABEL,"defect"); 
	    PlotIt::contour(ps,df,psp);
	  }
	  else if( answer=="rhs" )
	  {
	    psp.set(GI_TOP_LABEL,"fMG"); 
	    PlotIt::contour(ps,f,psp);
	  }
	  else if( answer=="grid" )
	  {
	    psp.set(GI_TOP_LABEL,"grid"); 
	    PlotIt::plot(ps,cg,psp);
	  }
      
	}
      }
      
    }
  }
  return 0;
}

//--------------------------------------------------------------------------------------
//  Test the coarse to fine transfer operator
//--------------------------------------------------------------------------------------
int Ogmg::
coarseToFineTest()
{

//    uMG.reference(u);  
//    fMG.reference(f);

  realCompositeGridFunction & u = uMG;
  realCompositeGridFunction & f = fMG;
  

  const int degreeSpace=2;
  OGPolyFunction exact(degreeSpace,mgcg.numberOfDimensions());
  const int numberOfComponents=1;
  RealArray spatialCoefficientsForTZ(5,5,5,numberOfComponents);  
  spatialCoefficientsForTZ=0.;
  RealArray timeCoefficientsForTZ(5,numberOfComponents);      
  timeCoefficientsForTZ=0.;
  timeCoefficientsForTZ(0,0)=1.;

  Index I1,I2,I3;
  int grid;
  aString buff;
  
  // debug=31;  // ***********************
  int debugcf=debug;

  useForcingAsBoundaryConditionOnAllLevels=true; // use f so define the dirichlet BC on all levels (normally only level=0)
  

  mgcg.update(MappedGrid::THEcenter | MappedGrid::THEvertex );
  
  for( int level=0; level<mgcg.numberOfMultigridLevels()-1; level++ )
  {
    cout << "====================Level = " << level << "======================\n";
    realCompositeGridFunction & uCoarse = u.multigridLevel[level+1];
    realCompositeGridFunction & uFine   = u.multigridLevel[level];

    realCompositeGridFunction & fFine   = f.multigridLevel[level];

    CompositeGrid & cgCoarse = mgcg.multigridLevel[level+1];
    CompositeGrid & cgFine = mgcg.multigridLevel[level];
    
//    uCoarse.display("uCoarse");
    

//      CompositeGridOperators & op = *uFine.getOperators();
  
//      op.setTwilightZoneFlow(TRUE);           // set twilight-zone flow for level 0 only
//      op.setTwilightZoneFlowFunction(exact);

    const int maxDegree=2;
    for( int degree=0; degree<=maxDegree; degree++ )
    {
      cout << "**********************************************\n";
      cout << "test a polynomial of degree " << degree << "...\n";
      spatialCoefficientsForTZ=0.;
      spatialCoefficientsForTZ(0,0,0,0)=1.;
      if( degree>0 )
      {
	spatialCoefficientsForTZ(0,0,0,0)=1.;
	spatialCoefficientsForTZ(1,0,0,0)=1.;
	spatialCoefficientsForTZ(0,1,0,0)=2.;
	if( mgcg.numberOfDimensions()==3 )
	  spatialCoefficientsForTZ(0,0,1,0)=-1.;
      }
      if( degree>1 )
      {
	spatialCoefficientsForTZ(2,0,0,0)=-1.;
	spatialCoefficientsForTZ(0,2,0,0)=1.5;
	if( mgcg.numberOfDimensions()==3 )
	  spatialCoefficientsForTZ(0,0,2,0)=3.;
      }
      exact.setCoefficients( spatialCoefficientsForTZ,timeCoefficientsForTZ );  // for u
      uFine=0.;
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	MappedGrid & mg = cgCoarse[grid];
	getIndex(mg.dimension(),I1,I2,I3);
	uCoarse[grid](I1,I2,I3)=exact(mg,I1,I2,I3,0);
	where( mg.mask()(I1,I2,I3)==0 )
	  uCoarse[grid](I1,I2,I3)=123456789.;  // put bogus values in un-used points.

        // For predefined equations we should set the RHS for dirichlet BC's
        MappedGrid & mgFine = cgFine[grid];
	for( int axis=axis1; axis<mgcg.numberOfDimensions(); axis++ )
	{
	  for( int side=Start; side<=End; side++ )
	  {
	    if( mgFine.boundaryCondition(side,axis)>0 ) //  ==dirichlet  )
	    {
	      getBoundaryIndex(mgFine.gridIndexRange(),side,axis,I1,I2,I3);
              //if( level>0 )
	      //  uFine[grid](I1,I2,I3)=exact(mgFine,I1,I2,I3,0);
              // else
	        fFine[grid](I1,I2,I3)=exact(mgFine,I1,I2,I3,0);
	    }
	  }
	}
	
      }
      if( debugcf & 16 )
	uCoarse.display(sPrintF(buff,"coarseToFine: BEFORE:uCoarse level=%i degree=%i",level,degree),NULL,"%7.1e ");
      if( debugcf & 16 )
	uFine.display(sPrintF(buff,"coarseToFine: BEFORE:uFine level=%i degree=%i",level,degree),NULL,"%7.1e ");

      coarseToFine( level );

      if( debugcf & 16 )
	uFine.display(sPrintF(buff,"coarseToFine: AFTER:uFine level=%i degree=%i",level,degree),NULL,"%7.1e ");

      real error;
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	MappedGrid & mg = cgFine[grid];
	getIndex(extendedGridIndexRange(mg),I1,I2,I3);
	where( mg.mask()(I1,I2,I3)!=0 )
	  error= max(fabs(uFine[grid](I1,I2,I3)-exact(mg,I1,I2,I3,0)));
	printf("Maximum error in coarse to fine on grid=%i is %e \n",grid,error);
	if( debugcf & 16 )
	{
	  realArray err(I1,I2,I3);
	  err=0;
	  where( mg.mask()(I1,I2,I3)!=0 )
	    err=fabs(uFine[grid](I1,I2,I3)-exact(mg,I1,I2,I3,0));
	  display(err,sPrintF(buff,"coarseToFine Error level=%i grid=%i degree=%i",level,grid,degree),NULL,"%6.1e ");
	}
      }
    }
  }
  useForcingAsBoundaryConditionOnAllLevels=false;
  
  return 0;
}

//--------------------------------------------------------------------------------------
//  Test the fine to coarse transfer operator
//--------------------------------------------------------------------------------------
int Ogmg::
fineToCoarseTest()
{

  realCompositeGridFunction & u = uMG;
  realCompositeGridFunction & f = fMG;

//    uMG.reference(u);  
//    fMG.reference(f);

//  realCompositeGridFunction defect(mgcg);
//  defectMG.reference(defect);  

  const int degreeSpace=2;
  OGPolyFunction exact(degreeSpace,mgcg.numberOfDimensions());
  const int numberOfComponents=1;
  RealArray spatialCoefficientsForTZ(5,5,5,numberOfComponents);  
  spatialCoefficientsForTZ=0.;
  RealArray timeCoefficientsForTZ(5,numberOfComponents);      
  timeCoefficientsForTZ=0.;
  timeCoefficientsForTZ(0,0)=1.;

  mgcg.update(MappedGrid::THEcenter | MappedGrid::THEvertex );

  Index I1,I2,I3;
  int grid;
  for( int level=0; level<mgcg.numberOfMultigridLevels()-1; level++ )
  {
    cout << "====================Level = " << level << "======================\n";
    realCompositeGridFunction & fCoarse      = fMG.multigridLevel[level+1];
    realCompositeGridFunction & defectFine   = defectMG.multigridLevel[level];
    CompositeGrid & cgCoarse                 = mgcg.multigridLevel[level+1];
    CompositeGrid & cgFine                   = mgcg.multigridLevel[level];
    
    for( int degree=0; degree<=2; degree++ )
    {
      cout << "*****************************************************************\n";
      cout << "test a polynomial of degree " << degree << "...\n";
      spatialCoefficientsForTZ=0.;
      spatialCoefficientsForTZ(0,0,0,0)=1.;
      if( degree>0 )
      {
	spatialCoefficientsForTZ(0,0,0,0)=1.;
	spatialCoefficientsForTZ(1,0,0,0)=1.;
	spatialCoefficientsForTZ(0,1,0,0)=2.;
	if( mgcg.numberOfDimensions()==3 )
	  spatialCoefficientsForTZ(0,0,1,0)=-1.;
      }
      if( degree>1 )
      {
	spatialCoefficientsForTZ(2,0,0,0)=-1.;
	spatialCoefficientsForTZ(0,2,0,0)=1.5;
	if( mgcg.numberOfDimensions()==3 )
	  spatialCoefficientsForTZ(0,0,2,0)=3.;
      }
      exact.setCoefficients( spatialCoefficientsForTZ,timeCoefficientsForTZ );  // for u
    
      fCoarse=0.;
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	MappedGrid & mg = cgFine[grid];
	getIndex(mg.dimension(),I1,I2,I3);
	defectFine[grid](I1,I2,I3)=exact(mg,I1,I2,I3,0);
	where( mg.mask()(I1,I2,I3) ==0 )
	  defectFine[grid](I1,I2,I3)=123456789.;  // put bogus values in un-used points.
      }

      if( level==1 ) // *** testing ***
      {
	// int grid=1, gridI=0; 
	int grid=0, gridI=1; 

//  	display(defectFine[grid],sPrintF("Before interp: defectFine, grid=%i, level=%i, degree=%i",grid,level,degree),
//                    debugFile,"%8.2e ");

        intArray & ip = cgFine.interpolationPoint[grid];
	intArray & il = cgFine.interpoleeLocation[grid];
	intArray & varWidth = cgFine.variableInterpolationWidth[grid];
        realArray & u = defectFine[gridI];

	for( int i=0 ; i<cgFine.numberOfInterpolationPoints(grid); i++ )
	{
          int width = varWidth(i);
	  int i1=il(i,0), i2=il(i,1), i3=0;
          real sum=0.;
	  for( int iw1=0; iw1<width; iw1++ )
	  for( int iw2=0; iw2<width; iw2++ )
	  {
	    sum+=u(i1+iw1,i2+iw2,i3);
	  }
          if( fabs(sum)>10. )
	  {
   	    fprintf(debugFile,"*** i=%i ip=(%i,%i) il=(%i,%i) sum=%e \n",i,ip(i,0),ip(i,1),i1,i2,sum);
	    for( int iw1=0; iw1<width; iw1++ )
	      for( int iw2=0; iw2<width; iw2++ )
	      {
		fprintf(debugFile,"...u(%i,%i) = %e \n",i1+iw1,i2+iw2,u(i1+iw1,i2+iw2,i3));
		
	      }
	  
	  }
	  
	}
	fprintf(debugFile," ******************* done checking interpolation ********\n");
	
//         interpolate( defectFine );
       
//  	display(defectFine[grid],sPrintF("After interp: defectFine, grid=%i, level=%i, degree=%i",grid,level,degree),
//                    debugFile,"%8.2e ");

      }
      
       
     if( debug & 16 )
	defectFine.display("Here is defectFine before fineToCoarse");

      fineToCoarse( level );

      real error;
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	MappedGrid & mg = cgCoarse[grid];
	getIndex(extendedGridIndexRange(mg),I1,I2,I3,-1);  // -1: not not check boundary (dirichlet condition)
	where( mg.mask()(I1,I2,I3) > 0 )
	  error= max(fabs(fCoarse[grid](I1,I2,I3)-exact(mg,I1,I2,I3,0)));
	printf("Maximum error in fine to coarse on grid=%i is %e \n",grid,error);
	// if( error>.1 || debug & 8 )
	if( (grid==1 && level==1) || debug & 8 )
	{
	  realArray err = evaluate(fabs(fCoarse[grid](I1,I2,I3)-exact(mg,I1,I2,I3,0)));
	  where( mg.mask()(I1,I2,I3) <= 0 )
            err=0.;

          display(defectFine[grid],sPrintF("defectFine on grid=%i, level=%i, degree=%i",grid,level,degree),
                  debugFile,"%8.2e ");
	  display(fCoarse[grid],sPrintF("fCoarse on grid=%i, level=%i, degree=%i",grid,level,degree),
                  debugFile,"%8.2e ");

	  display(err,sPrintF("The error in fCoarse on grid=%i, level=%i, degree=%i",grid,level,degree),
                  debugFile,"%6.1e ");
	}
      }
      if( debug & 16 )
      {
	fCoarse.display("Here is fCoarse after fineToCoarse");
      }
    }

  }
  
  return 0;
}


int Ogmg::
bcTest()
// =======================================================================================
// /Description:
//    Test the boundary conditions.
// =======================================================================================
{

  realCompositeGridFunction & u0 = uMG;
  realCompositeGridFunction & f0 = fMG;
//    uMG.reference(u0);  
//    fMG.reference(f0);

  printf("bcTest: u0.numberOfGrids = %i \n",u0.numberOfGrids());
  printf("bcTest: u0.multigridLevel[0].numberOfGrids = %i \n",u0.multigridLevel[0].numberOfGrids());

  printf("bcTest: uMG.numberOfGrids = %i \n",uMG.numberOfGrids());
  printf("bcTest: uMG.multigridLevel[0].numberOfGrids = %i \n",uMG.multigridLevel[0].numberOfGrids());

  

  for( int level=0; level<mgcg.numberOfMultigridLevels(); level++ )
  {
    cout << "====================Level = " << level << "======================\n";
    realCompositeGridFunction & u = u0.multigridLevel[level];
    realCompositeGridFunction & f = f0.multigridLevel[level];
    CompositeGrid & cg = mgcg.multigridLevel[level];
    
    u=2.; 
    f=1.;  
    Index I1,I2,I3;
    for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      getIndex(mg.dimension(),I1,I2,I3);
      where( mg.mask()(I1,I2,I3)==0 )
      {
	u[grid](I1,I2,I3)=789.;  // put bogus values at unused points
      }

      // fill in the BC's into f
      int extra=1;
      for( int axis=axis1; axis<cg.numberOfDimensions(); axis++ )
      {
	for( int side=Start; side<=End; side++ )
	{
	  if( boundaryCondition(side,axis,grid)==OgmgParameters::equation )  // this is a neumann like BC
	  {
	    getGhostIndex(mg.extendedIndexRange(),side,axis,I1,I2,I3,1,extra);
            u[grid](I1,I2,I3)=123.;  
	    f[grid](I1,I2,I3)=0.;   
	  }
	  else 
	  { // extrapolation
	    getGhostIndex(mg.extendedIndexRange(),side,axis,I1,I2,I3,1,extra);
            u[grid](I1,I2,I3)=123.;  
	    f[grid](I1,I2,I3)=0.;   
	  }
	}
      }
    }
    
    u.display("bcTest: u before applyBoundaryConditions","%9.2e");
    applyBoundaryConditions( level,u,f );
    u.display("bcTest: u after applyBoundaryConditions","%9.2e");
    
  }
  return 0;
}

