#include "Ogmg.h"
#include "ParallelUtility.h"

// Define an Index from a base and bound and optional stride
Index IndexBB(int base, int bound, int stride )
{
//  return Range(base,bound,stride);
  return Index(base,(bound-base+stride)/stride,stride);
}

// Change the stride on an Index
Index IndexBB(Index I, const int stride )
{
  return Index(I.getBase(),(I.getBound()-I.getBase()+stride)/stride,stride);
}


//\begin{>>OgmgInclude.tex}{\subsection{getNumberOfIterations}} 
int Ogmg::
chooseBestSmoother()
// ======================================================================================
// /Description:
//     Attempt to guess which smoother will work best on each grid.
//\end{OgmgInclude.tex}
// ======================================================================================
{
//   int level=0;
//   CompositeGrid & cg = mgcg;
  
//   for( int grid=0; grid<c.numberOfComponentGrids(); grid++ )
//   {
//   }
  

  return 0;
}





//----------------------------------------------------------------
//  Return the "mean" value of a grid function
//  This is a special definition of the mean, used by Ogmg
//  The mean is just the sum of all points with mask!=0
//----------------------------------------------------------------
real Ogmg::
getMean(realCompositeGridFunction & u)
{
  CompositeGrid & c = (CompositeGrid&)(*u.gridCollection);
  real mean=0.;
  Index I1,I2,I3;
  for( int grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    getIndex(c[grid].extendedIndexRange(),I1,I2,I3);
    where( c[grid].mask()(I1,I2,I3)!=0 )
    {
      mean+=sum(u[grid](I1,I2,I3));
    }
  }
  return mean;
}

//---------------------------------------------------------------
// Set the "mean" of a grid function 
// The mean is just the sum of all points with mask!=0
//--------------------------------------------------------------
void Ogmg::
setMean(realCompositeGridFunction & u, const real meanValue, int level)
{
  if( !parameters.assignMeanValueForSingularProblem ) return;

  CompositeGrid & c = (CompositeGrid&)(*u.gridCollection);
  real mean=0.;
  int count=0;
  Index I1,I2,I3;
  int grid;
  for( grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    getIndex(c[grid].extendedIndexRange(),I1,I2,I3);
    intArray mask = c[grid].mask()(I1,I2,I3)!=0;
    where( mask )
    {
      mean+=sum(u[grid](I1,I2,I3));
      // doesn't work: count++;
    }
    count+=sum(mask);   
  }
  if( Ogmg::debug & 2 )
  {
    printF("%*.1s Ogmg::setMean: level=%i, actual-mean/count=%g, count=%i, meanValue - actual-mean=%8.2e\n",
           level*4," ",level,mean/max(1,count),count,meanValue-mean);
  }
  real value=(meanValue-mean)/max(1,count);
  for( grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    getIndex(c[grid].dimension(),I1,I2,I3);  // shift all points ****
    u[grid](I1,I2,I3)+=value;
  }
  u.periodicUpdate();  
}

#define l2normOpt EXTERN_C_NAME(l2normopt)
#define l2AndMaxNormOpt EXTERN_C_NAME(l2andmaxnormopt)
#define maxNormOpt EXTERN_C_NAME(maxnormopt)
#define l2ErrorOpt EXTERN_C_NAME(l2erroropt)

extern "C"
{

  void l2normOpt( const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
		  const int &nd3a, const int &nd3b,
		  const int &n1a, const int &n1b, 
		  const int &n2a, const int &n2b, 
		  const int &n3a, const int &n3b, 
		  const real & u, const int & mask, real & norm, int & count );

  void l2AndMaxNormOpt( const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
		  const int &nd3a, const int &nd3b,
		  const int &n1a, const int &n1b, 
		  const int &n2a, const int &n2b, 
		  const int &n3a, const int &n3b, 
		  const real & u, const int & mask, real & norm, int & count, real & uMax );

  void maxNormOpt( const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
		  const int &nd3a, const int &nd3b,
		  const int &n1a, const int &n1b, 
		  const int &n2a, const int &n2b, 
		  const int &n3a, const int &n3b, 
		  const real & u, const int & mask, real & uMax );

  void l2ErrorOpt( const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
		  const int &nd3a, const int &nd3b,
		  const int &n1a, const int &n1b, 
		  const int &n2a, const int &n2b, 
		  const int &n3a, const int &n3b, 
		  const real & u, const real & v, const int & mask, real & norm, int & count );
}

//---------------------------------------------------------
//  Return    sqrt(  (SUM u^2) / number of points )
//---------------------------------------------------------
real Ogmg::
l2Norm(const realCompositeGridFunction & u )
{
  CompositeGrid & c = (CompositeGrid&)(*u.gridCollection);
  real returnValue=0.;
  int count=0;
  Index I1,I2,I3;
  for( int grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    MappedGrid & mg =c[grid];
    getIndex(mg.gridIndexRange(),I1,I2,I3);
    if( parameters.useOptimizedVersion )
    {
      const IntegerArray & d = mg.dimension();
      int countg;
      real uSquared;
      l2normOpt( d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
		 I1.getBase(),I1.getBound(),I2.getBase(),I2.getBound(),I3.getBase(),I3.getBound(),
		 *getDataPointer(u[grid]),  *getDataPointer(mg.mask()), uSquared, countg);
      returnValue+=uSquared;
      count+=countg;
    }
    else
    {
      const intArray & mask = evaluate( mg.mask()(I1,I2,I3)!=0 );
      count+=sum(mask);
      where( mask )
      {
	returnValue+=sum(u[grid](I1,I2,I3)*u[grid](I1,I2,I3));
      }
    }
  }
  
  returnValue=sqrt(returnValue/max(1,count));
  return returnValue;
}


real Ogmg::
l2Norm(const realMappedGridFunction & u )
{
  MappedGrid & mg = *u.getMappedGrid();
  real returnValue=0.;
  int count=0;
  if( parameters.useOptimizedVersion )
  {
    #ifdef USE_PPP
      const realSerialArray & uLocal = u.getLocalArrayWithGhostBoundaries();
      const intArray & maskd = mg.mask();
      const intSerialArray & maskLocal = maskd.getLocalArrayWithGhostBoundaries();
      IntegerArray gid; gid = mg.gridIndexRange();
      for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
      {
	gid(0,axis) = max(gid(0,axis),maskLocal.getBase(axis) +maskd.getGhostBoundaryWidth(axis));
	gid(1,axis) = min(gid(1,axis),maskLocal.getBound(axis)-maskd.getGhostBoundaryWidth(axis));
      }
    #else
      const realSerialArray & uLocal = u;
      const intSerialArray & maskLocal = mg.mask();
      const IntegerArray & gid = mg.gridIndexRange();
    #endif

    l2normOpt( maskLocal.getBase(0),maskLocal.getBound(0),
	       maskLocal.getBase(1),maskLocal.getBound(1),
	       maskLocal.getBase(2),maskLocal.getBound(2),
	       gid(0,0),gid(1,0),gid(0,1),gid(1,1),gid(0,2),gid(1,2),
	       *getDataPointer(uLocal),  *getDataPointer(maskLocal), returnValue, count );

    #ifdef USE_PPP
      returnValue=ParallelUtility::getSum(returnValue);
      count=ParallelUtility::getSum(count);
    #endif

    returnValue=sqrt(returnValue/max(1,count));

    // printf(" l2Norm: count=%i, returnValue=%8.2e\n",count,returnValue);
    
  }
  else
  {
    Index I1,I2,I3;
    getIndex(mg.gridIndexRange(),I1,I2,I3);
    const intArray & mask = evaluate( mg.mask()(I1,I2,I3)!=0 );
    count+=sum(mask);
    where( mask )
    {
      returnValue+=sum(u(I1,I2,I3)*u(I1,I2,I3));
    }
    returnValue=sqrt(returnValue/max(1,count));
  }
  
  return returnValue;
}

#undef l2normOpt 

//---------------------------------------------------------
//  Return    max |u| 
//---------------------------------------------------------
real Ogmg::
maxNorm(const realCompositeGridFunction & u )
{
  CompositeGrid & c = (CompositeGrid&)(*u.gridCollection);
  real returnValue=0.;
  for( int grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    MappedGrid & mg = c[grid];
    if( parameters.useOptimizedVersion )
    {
      #ifdef USE_PPP
        const realSerialArray & uLocal = u[grid].getLocalArrayWithGhostBoundaries();
        const intArray & maskd = mg.mask();
        const intSerialArray & maskLocal = maskd.getLocalArrayWithGhostBoundaries();
        IntegerArray gid; gid = mg.gridIndexRange();
        for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
	{
	  gid(0,axis) = max(gid(0,axis),maskLocal.getBase(axis) +maskd.getGhostBoundaryWidth(axis));
	  gid(1,axis) = min(gid(1,axis),maskLocal.getBound(axis)-maskd.getGhostBoundaryWidth(axis));
	}
      #else
        const realSerialArray & uLocal = u[grid];
        const intSerialArray & maskLocal = mg.mask();
        const IntegerArray & gid = mg.gridIndexRange();
      #endif    

      real maxNorm=0.;
      maxNormOpt( maskLocal.getBase(0),maskLocal.getBound(0),
                  maskLocal.getBase(1),maskLocal.getBound(1),
                  maskLocal.getBase(2),maskLocal.getBound(2),
		  gid(0,0),gid(1,0),gid(0,1),gid(1,1),gid(0,2),gid(1,2),
		 *getDataPointer(uLocal),  *getDataPointer(maskLocal), maxNorm );
      #ifdef USE_PPP
        maxNorm=ParallelUtility::getMaxValue(maxNorm);
      #endif
      returnValue=max(returnValue,maxNorm);
    }
    else
    {
     Index I1,I2,I3;
      getIndex(mg.gridIndexRange(),I1,I2,I3);
      where( mg.mask()(I1,I2,I3)!=0 )
      {
	returnValue=max(returnValue,max(fabs(u[grid](I1,I2,I3))));
      }
    }
    
  }
  return returnValue;

}

real Ogmg::
maxNorm(const realMappedGridFunction & u )
{
  MappedGrid & mg = *u.getMappedGrid();
  real returnValue=0.;
  if( parameters.useOptimizedVersion )
  {
    #ifdef USE_PPP
      const realSerialArray & uLocal = u.getLocalArrayWithGhostBoundaries();
      const intArray & maskd = mg.mask();
      const intSerialArray & maskLocal = maskd.getLocalArrayWithGhostBoundaries();
      IntegerArray gid; gid = mg.gridIndexRange();
      for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
	{
	  gid(0,axis) = max(gid(0,axis),maskLocal.getBase(axis) +maskd.getGhostBoundaryWidth(axis));
	  gid(1,axis) = min(gid(1,axis),maskLocal.getBound(axis)-maskd.getGhostBoundaryWidth(axis));
	}
    #else
      const realSerialArray & uLocal = u;
      const intSerialArray & maskLocal = mg.mask();
      const IntegerArray & gid = mg.gridIndexRange();
    #endif

    real maxNorm=0.;
    maxNormOpt( maskLocal.getBase(0),maskLocal.getBound(0),
		maskLocal.getBase(1),maskLocal.getBound(1),
		maskLocal.getBase(2),maskLocal.getBound(2),
                gid(0,0),gid(1,0),gid(0,1),gid(1,1),gid(0,2),gid(1,2),
		*getDataPointer(uLocal),  *getDataPointer(maskLocal), returnValue );
    #ifdef USE_PPP
      returnValue=ParallelUtility::getMaxValue(returnValue);
    #endif
  }
  else
  {
    Index I1,I2,I3;
    getIndex(mg.gridIndexRange(),I1,I2,I3);
    where( mg.mask()(I1,I2,I3)!=0 )
    {
      returnValue=max(returnValue,max(fabs(u(I1,I2,I3))));
    }
  }
  
  return returnValue;

}

//---------------------------------------------------------
//  Return    sqrt(  (SUM (u-v)^2) / number of points )
//---------------------------------------------------------
real Ogmg::
l2Error(const realCompositeGridFunction & u, const realCompositeGridFunction & v )
{
  CompositeGrid & c = (CompositeGrid&)(*u.gridCollection);
  real returnValue=0.;
  int count=0;
  Index I1,I2,I3;
  for( int grid=0; grid<c.numberOfComponentGrids(); grid++ )
  {
    MappedGrid & mg =c[grid];
    if( parameters.useOptimizedVersion )
    {
      #ifdef USE_PPP
        const realSerialArray & uLocal = u[grid].getLocalArrayWithGhostBoundaries();
        const realSerialArray & vLocal = v[grid].getLocalArrayWithGhostBoundaries();
        const intArray & maskd = mg.mask();
        const intSerialArray & maskLocal = maskd.getLocalArrayWithGhostBoundaries();
        IntegerArray gid; gid = mg.gridIndexRange();
        for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
        {
  	gid(0,axis) = max(gid(0,axis),maskLocal.getBase(axis) +maskd.getGhostBoundaryWidth(axis));
  	gid(1,axis) = min(gid(1,axis),maskLocal.getBound(axis)-maskd.getGhostBoundaryWidth(axis));
        }
      #else
        const realSerialArray & uLocal = u[grid];
        const realSerialArray & vLocal = v[grid];
        const intSerialArray & maskLocal = mg.mask();
        const IntegerArray & gid = mg.gridIndexRange();
      #endif

      int countg=0;
      real uSquared;
      l2ErrorOpt( maskLocal.getBase(0),maskLocal.getBound(0),
		  maskLocal.getBase(1),maskLocal.getBound(1),
		  maskLocal.getBase(2),maskLocal.getBound(2),
		  gid(0,0),gid(1,0),gid(0,1),gid(1,1),gid(0,2),gid(1,2),
		  *getDataPointer(uLocal), *getDataPointer(vLocal), 
		  *getDataPointer(maskLocal), uSquared, countg);
      returnValue+=uSquared;
      count+=countg;
    }
    else
    {
      getIndex(mg.gridIndexRange(),I1,I2,I3);
      const intArray & mask = evaluate( mg.mask()(I1,I2,I3)!=0 );
      count+=sum(mask);
      realArray w; 
      w = u[grid](I1,I2,I3)-v[grid](I1,I2,I3);
      where( mask )
      {
	returnValue+=sum(w*w);
      }
    }
  }
  #ifdef USE_PPP
    returnValue=ParallelUtility::getSum(returnValue);
    count=ParallelUtility::getSum(count);
  #endif

  returnValue=sqrt(returnValue/max(1,count));
  return returnValue;
}

