#include "Ogmg.h"
#include "display.h"

// static int *nipn=NULL, *ndipn=NULL, **ipn=NULL;  // add to class and destroy when done
      

#define smoothRedBlackOpt EXTERN_C_NAME(smoothredblackopt)
#define smoothJacobiOpt EXTERN_C_NAME(smoothjacobiopt)
#define getInterpNeighbours EXTERN_C_NAME(getinterpneighbours)

extern "C"
{
  void smoothJacobiOpt( const int &nd,  const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
                       const int &nd3a, const int &nd3b,
		       const int &n1a, const int &n1b, const int &n1c,
                       const int &n2a, const int &n2b, const int &n2c,
                       const int &n3a, const int &n3b, const int &n3c,
                       const int &ndc, const real & f, const real & c,
                       const real & u, const real & v, const int & mask, const int & option, 
                       const int & order, const int & sparseStencil,
                       const real & cc, const real & varCoeff, const real & dx, const real & omega, const int & bc,
                       const int &np, const int &ndip, const int &ip, const int & ipar );

  void smoothRedBlackOpt( const int &nd,  const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
                       const int &nd3a, const int &nd3b,
		       const int &n1a, const int &n1b, const int &n1c,
                       const int &n2a, const int &n2b, const int &n2c,
                       const int &n3a, const int &n3b, const int &n3c, 
                       const int &ndc, const real & f, const real & c,
                       const real & u, const int & mask, const int & option, 
                       const int & order, const int & sparseStencil,
                       const real & cc, const real & varCoeff, const real & dx, const real & omega,
                       const int & useLocallyOptimalOmega, const real & variableOmegaScaleFactor );

    

  void getInterpNeighbours( const int &nd,  const int & nd1a, const int &nd1b, const int &nd2a, const int &nd2b,
	  		    const int &nd3a, const int &nd3b,
			    const int &n1a, const int &n1b,
			    const int &n2a, const int &n2b,
			    const int &n3a, const int &n3b,
			    const int &mask, int &mask2,
			    const int & nip,const int &ndip,const int &ip, 
			    int &nipn, const int &ndipn, int &ipn, const int &width, const int & ierr );

}


//\begin{>>OgmgInclude.tex}{\subsection{smoothBoundary}}
void Ogmg::
smoothBoundary(int level, 
               int grid, 
               int bcOption[6], 
               int numberOfLayers /* =1 */, 
               int numberOfIterations /* =1 */ )
// ======================================================================================
//   /Description:
//       Smooth points near boundaries.
//
// /bcOption[6] (input): smooth (side,axis) if bcOption[side+2*axis]=1
//\end{OgmgInclude.tex} 
// ======================================================================================
{
  if( numberOfLayers<1 || numberOfIterations<1 )
    return;

  if( level>parameters.numberOfLevelsForBoundarySmoothing )
    return;

  real time0=getCPU();
  
  if( debug & 8 )
    printf("Smooth the boundary of grid=%i level=%i, numberOfLayers=%i\n",grid,level,numberOfLayers);
  

  realMappedGridFunction & u = uMG.multigridLevel[level][grid];
  realMappedGridFunction & f = fMG.multigridLevel[level][grid];
  realMappedGridFunction & c =  level==0 ? cMG[grid] : cMG.multigridLevel[level][grid];
  realArray & defect = defectMG.multigridLevel[level][grid];
  MappedGrid & mg = mgcg.multigridLevel[level][grid];  
  const intArray & mask = mg.mask();

//  Index Iav[3], &I1a=Iav[0], &I2a=Iav[1], &I3a=Iav[2];
//  getIndex(mg.gridIndexRange(),I1a,I2a,I3a); //

  int nab[6]={0,0,0,0,0,0};  //


  const IntegerArray & mgBC=mg.boundaryCondition();
  int bc[6]={0,0,0,0,0,0};
  for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
  {
    for( int side=0; side<=1; side++ )
    {
      // nab[side+2*axis]=mg.gridIndexRange(side,axis);
      nab[side+2*axis]=mg.indexRange(side,axis);
      
      if( mgBC(side,axis)>0 )
      {
//  	bc[side+2*axis]= bcOption==0 ? boundaryCondition(side,axis,grid)==equation : 
//  	  bcOption==1 ? boundaryCondition(side,axis,grid)==extrapolate : 1;
	bc[side+2*axis]= bcOption[side+2*axis];
        if( boundaryCondition(side,axis,grid)==OgmgParameters::extrapolate )
	{
	  nab[side+2*axis]+= 1-2*side;  // do not smooth ON the dirichlet boundaries.
	}
	
      }
      else
      {
	bc[side+2*axis]=0;
      }
    }
  }
  

  const IntegerArray & d = mg.dimension();
  const int ndc=c.getLength(0);

  const bool rectangular=(*c.getOperators()).isRectangular() && ( level < mgcg.numberOfMultigridLevels()-numberOfExtraLevels ); 

  real dx[3]={1.,1.,1.};
  int sparseStencil=general;
  if( mg.isRectangular() ) // ***
  {
    if( equationToSolve!=OgesParameters::userDefined ) // ***  defect is used on the boundary
    {
      if( equationToSolve==OgesParameters::divScalarGradOperator ||
	  equationToSolve==OgesParameters::variableHeatEquationOperator ||
	  equationToSolve==OgesParameters::divScalarGradHeatEquationOperator )
      {
	sparseStencil=level==0 ? sparseVariableCoefficients : variableCoefficients;
      }
      else
      {
	sparseStencil=level==0 ? sparseConstantCoefficients : constantCoeff;
      }
    }
    else if( rectangular && assumeSparseStencilForRectangularGrids )
      sparseStencil=sparse;

    mg.getDeltaX(dx);
  }

  real *vp = getDataPointer(defect); // temp space for jacobi
  real *up=getDataPointer(u);
  real *vcp = varCoeff!=NULL ? getDataPointer((*varCoeff).multigridLevel[level][grid]) : up;

  int ipar[10];
  ipar[0]=numberOfLayers;

  int option=3;  // Gauss-Seidel on the boundary
  real omega=1.; // .8;


//    const int numLevels=parameters.numberOfLevelsForBoundarySmoothing;
//    const int gid=level+numLevels*grid;

//    if( parameters.numberOfInterpolationLayersToSmooth>0 && level<numLevels && 
//        mgcg.multigridLevel[level].numberOfInterpolationPoints(grid)>0 && (nipn==NULL || ipn[gid]==NULL) )
//    {
//      // *** allocate space and/or determine interpolation neighbours that should be smoothed. ***

//      const int nip = mgcg.multigridLevel[level].numberOfInterpolationPoints(grid);
//      if( nip>0 )
//      {
    
//        const intArray & ip = mgcg.multigridLevel[level].interpolationPoint[grid];
//        const int width=parameters.numberOfInterpolationLayersToSmooth;

//        if( nipn==NULL )
//        {
//          const int numArrays=numLevels*mgcg.multigridLevel[0].numberOfComponentGrids();
//  	nipn  = new int [numArrays];
//  	ndipn = new int [numArrays];
//  	ipn   = new int * [numArrays];
//          for( int i=0; i<numArrays; i++ )
//  	{
//  	  nipn[i]=0;
//  	  ndipn[i]=0;
//            ipn[i]=NULL;
//  	}
//        }

//        const int maxNumberOfNeighbours = nip*int(pow(2*width+1,mg.numberOfDimensions())-1 +.5); // this is normally too many **FIX**

//        ndipn[gid]=maxNumberOfNeighbours;


//        assert( ipn[gid]==NULL );
      
//        ipn[gid] = new int [ndipn[gid]*mg.numberOfDimensions()];  // we store (i1,i2) or (i1,i2,i3)

//        IntegerArray mask2(mask.dimension(0),mask.dimension(1),mask.dimension(2)); // no need to initialize, done below
      
//        const IntegerArray & d = mg.dimension();
//        const IntegerArray & gir = mg.gridIndexRange();
      
//        getInterpNeighbours( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
//                             gir(0,0),gir(1,0),gir(0,1),gir(1,1),gir(0,2),gir(1,2),
//  			   *getDataPointer(mask),*getDataPointer(mask2), nip,ip.getLength(0),*getDataPointer(ip), 
//  			   nipn[gid],ndipn[gid],ipn[gid][0], width );
      
//        assert( nipn[gid]<= ndipn[gid] );

//        if( debugFile!=NULL && Ogmg::debug & 1 )
//        {
//            fprintf(debugFile,"\n ************ INFO for Smoothing Interpolation Neighbours grid=%i level=%i width=%i *****************\n",
//                   grid,level,width);

//  	fprintf(debugFile,"nip=%i, ip=",nip);
//          int i;
//  	for( i=0; i<nip; i++ )
//  	{
//     	  fprintf(debugFile,"(%3i,%3i)",ip(i,0),ip(i,1));
//            if( i%10 == 9 ) fprintf(debugFile,"\n  ");
//  	}
	
//           fprintf(debugFile,"\n\n nipn=%i, ndipn=%i, ipn=",nipn[gid],ndipn[gid]);      
//           int *pipn = ipn[gid];
//           const int ndipn0=ndipn[gid];
//  #define IPN(i,m) pipn[(i)+ndipn0*(m)]
	 
//  	 for( i=0; i<nipn[gid]; i++ )
//  	 {
//  	   fprintf(debugFile,"(%3i,%3i)",IPN(i,0),IPN(i,1));
//  	   if( i%10 == 9 ) fprintf(debugFile,"\n  ");
//  	 }
//        }
//      }
    
//    }
  

  if( debug & 4 )
  {
    printf("smoothBoundary: layers=%i nab= [%i,%i][%i,%i][%i,%i] bc=[%i,%i][%i,%i][%i,%i] "
	   " omega=%8.2e \n",numberOfLayers,
	   nab[0],nab[1],nab[2],nab[3],nab[4],nab[5],bc[0],bc[1],bc[2],bc[3],bc[4],bc[5],omega);
  }

  const int np=0,ndip=1,ip=0;
  const int n1c=1, n2c=1, n3c=1;
  for( int iter=0; iter<numberOfIterations; iter++ )
  {
    if( false ) // flip between 3 and 5 sometimes helps...
      ipar[0]=numberOfLayers - 2*(iter % 2);
      

    smoothJacobiOpt( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
		     nab[0],nab[1],n1c,nab[2],nab[3],n2c,nab[4],nab[5],n3c, ndc,
		     *getDataPointer(f),
		     *getDataPointer(c),
		     *up, *vp,
		     *getDataPointer(mask), 
		     option, orderOfAccuracy, sparseStencil, constantCoefficients(0,grid,level), *vcp, dx[0], omega,
		     bc[0],np,ndip,ip, ipar[0] );


    
//      if( parameters.numberOfInterpolationLayersToSmooth>0 && level<numLevels && nipn!=NULL && nipn[gid]>0 )
//      {
//        if( debug & 4 )
//  	printf("Smooth interp neighbours: level=%i grid=%i, number of neighbours=nipn[gid]=%i \n",
//  	       level,grid,nipn[gid]);

//        int optionInterp=4; // smooth a list of points with Gauss-Seidel
//        smoothJacobiOpt( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
//  		       nab[0],nab[1],n1c,nab[2],nab[3],n2c,nab[4],nab[5],n3c, ndc,
//  		       *getDataPointer(f),
//  		       *getDataPointer(c),
//  		       *up, *vp,
//  		       *getDataPointer(mask), 
//  		       optionInterp, orderOfAccuracy, sparseStencil, constantCoefficients(0,grid,level), *vcp, dx[0], omega,
//  		       bc[0], nipn[gid],ndipn[gid],ipn[gid][0], ipar[0] );
//      }
    

    applyBoundaryConditions( level,grid,u,f ); 
  }



  tm[timeForBoundarySmooth]+=getCPU()-time0;
}

//\begin{>>OgmgInclude.tex}{\subsection{smoothBoundary}}
void Ogmg::
smoothInterpolationNeighbours(int level, int grid )
// ======================================================================================
//   /Description:
//      Smooth neighbours of interpolation points
//
// /bcOption[6] (input): smooth (side,axis) if bcOption[side+2*axis]=1
//\end{OgmgInclude.tex} 
// ======================================================================================
{
  if( parameters.numberOfInterpolationLayersToSmooth<1 || parameters.numberOfInterpolationSmoothIterations<1 ||
      level>=parameters.numberOfLevelsForInterpolationSmoothing )
    return;

  real time0=getCPU();
  
  const int numberOfLayers=parameters.numberOfInterpolationLayersToSmooth;
  const int numberOfIterations=parameters.numberOfInterpolationSmoothIterations;
  
  if( debug & 8 )
    printf("Smooth interpolation neighbours of grid=%i level=%i, numberOfLayers=%i\n",grid,level,numberOfLayers);
  

  realMappedGridFunction & u = uMG.multigridLevel[level][grid];
  realMappedGridFunction & f = fMG.multigridLevel[level][grid];
  realMappedGridFunction & c =  level==0 ? cMG[grid] : cMG.multigridLevel[level][grid];
  realArray & defect = defectMG.multigridLevel[level][grid];
  MappedGrid & mg = mgcg.multigridLevel[level][grid];  
  const intArray & mask = mg.mask();

  const IntegerArray & d = mg.dimension();
  const IntegerArray & gir = mg.gridIndexRange();
  
  const int ndc=c.getLength(0);

  const bool rectangular=(*c.getOperators()).isRectangular() && ( level < mgcg.numberOfMultigridLevels()-numberOfExtraLevels ); 

  real dx[3]={1.,1.,1.};
  int sparseStencil=general;
  if( mg.isRectangular() ) // ***
  {
    if( equationToSolve!=OgesParameters::userDefined ) // ***  defect is used on the boundary
    {
      if( equationToSolve==OgesParameters::divScalarGradOperator ||
	  equationToSolve==OgesParameters::variableHeatEquationOperator ||
	  equationToSolve==OgesParameters::divScalarGradHeatEquationOperator )
      {
	sparseStencil=level==0 ? sparseVariableCoefficients : variableCoefficients;
      }
      else
      {
	sparseStencil=level==0 ? sparseConstantCoefficients : constantCoeff;
      }
    }
    else if( rectangular && assumeSparseStencilForRectangularGrids )
      sparseStencil=sparse;

    mg.getDeltaX(dx);
  }

  real *vp = getDataPointer(defect); // temp space for jacobi
  real *up=getDataPointer(u);
  real *vcp = varCoeff!=NULL ? getDataPointer((*varCoeff).multigridLevel[level][grid]) : up;


  const int numLevels=parameters.numberOfLevelsForInterpolationSmoothing;
  const int gid=level+numLevels*grid;

  if( parameters.numberOfInterpolationLayersToSmooth>0 && level<numLevels && 
      mgcg.multigridLevel[level].numberOfInterpolationPoints(grid)>0 && (nipn==NULL || ipn[gid]==NULL) )
  {
    // *** allocate space and/or determine interpolation neighbours that should be smoothed. ***

    const int nip = mgcg.multigridLevel[level].numberOfInterpolationPoints(grid);
    if( nip>0 )
    {
    
      const intArray & ip = mgcg.multigridLevel[level].interpolationPoint[grid];
      const int width=parameters.numberOfInterpolationLayersToSmooth;

      if( nipn==NULL )
      {
        const int numArrays=numLevels*mgcg.multigridLevel[0].numberOfComponentGrids();

	numberOfIBSArrays=numArrays;
	nipn  = new int [numArrays];
	ndipn = new int [numArrays];
	ipn   = new int * [numArrays];
        for( int i=0; i<numArrays; i++ )
	{
	  nipn[i]=0;
	  ndipn[i]=0;
          ipn[i]=NULL;
	}
      }

      // ** const int maxNumberOfNeighbours = nip*int(pow(2*width+1,mg.numberOfDimensions())-1 +.5); // this is normally too many **FIX**

      // The total number of neighbours of interpolation points is not very much larger than the current number
      // The worst case seems to be when the interp points are arranged in a circle
      //     2D :     New/Old = [ 2*pi*(r+dr) ]/[ 2*pi*r ] = 1+dr/r 
      //     3D :    New/Old = [ (4/3)pi*(r+dr)^2 ]/[ (4/3)pi*(r)^2 ] = (1+dr/r)^2

      const int maxNumberOfNeighbours = int( width*max(nip*(1.1 +4*4*4/nip)+20.,100.) ); 

      ndipn[gid]=maxNumberOfNeighbours;


      assert( ipn[gid]==NULL );
      
      ipn[gid] = new int [ndipn[gid]*mg.numberOfDimensions()];  // we store (i1,i2) or (i1,i2,i3)

      IntegerArray mask2(mask.dimension(0),mask.dimension(1),mask.dimension(2)); // no need to initialize, done below
      
      const IntegerArray & d = mg.dimension();
      
  
      const IntegerArray & mir = mg.indexRange(); 
      // valid neighbours live here (N.B. Do not include periodic boundaries or Dirichlet boundaries)
      int ir[6]={0,0,0,0,0,0};
      for( int side=0; side<=1; side++ )
      {
	for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
	{
	  ir[side+2*axis]=mir(side,axis);
	  if( boundaryCondition(side,axis,grid)==OgmgParameters::extrapolate ) // dirichlet BC
	    ir[side+2*axis]+= 1-2*side;
	}
      }
      int ierr=0;
      getInterpNeighbours( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
                           ir[0],ir[1],ir[2],ir[3],ir[4],ir[5],
			   *getDataPointer(mask),*getDataPointer(mask2), nip,ip.getLength(0),*getDataPointer(ip), 
			   nipn[gid],ndipn[gid],ipn[gid][0], width, ierr );
      
      int numberOfTrys=0;
      while( ierr!=0 )
      {
        numberOfTrys++;
	printf("Ogmg:smoothInterpolationNeighbours:WARNING: Not enough space was allocated for interpolation neighbours\n"
               "    level=%i grid=%i ...numberOfInterpolationPoints=%i ...estimated number of neighbours=%i \n",
                  level,grid,nip,ndipn[gid]);

	ndipn[gid]= int( ndipn[gid]*1.5+.5);
	printf("  ....increasing estimate of neighbours to %i\n",ndipn[gid]);

        delete [] ipn[gid];
	ipn[gid] = new int [ndipn[gid]*mg.numberOfDimensions()];  

	getInterpNeighbours( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
			     ir[0],ir[1],ir[2],ir[3],ir[4],ir[5],
			     *getDataPointer(mask),*getDataPointer(mask2), nip,ip.getLength(0),*getDataPointer(ip), 
			     nipn[gid],ndipn[gid],ipn[gid][0], width, ierr );
	
        assert( numberOfTrys<5 );
      }
      if( debug & 4 )
	printf("****Ogmg:smoothInterpolationNeighbours: level=%i grid=%i nip=%i, nipn=%i "
	       "(ndipn=%i, excess=%i excess/nipn=%6.2f)\n",
	       level,grid,nip,nipn[gid],ndipn[gid],ndipn[gid]-nipn[gid],(ndipn[gid]-nipn[gid])/real(nipn[gid]));
      
      
      if( debugFile!=NULL && Ogmg::debug & 16 )
      {
	fprintf(debugFile,"\n ****** INFO for Smoothing Interpolation Neighbours grid=%i level=%i width=%i ********\n",
		grid,level,width);

	fprintf(debugFile,"nip=%i, ip=",nip);
        int i;
	for( i=0; i<nip; i++ )
	{
   	  fprintf(debugFile,"(%3i,%3i)",ip(i,0),ip(i,1));
          if( i%10 == 9 ) fprintf(debugFile,"\n  ");
	}
	
         fprintf(debugFile,"\n\n nipn=%i, ndipn=%i, ipn=",nipn[gid],ndipn[gid]);      
         int *pipn = ipn[gid];
         const int ndipn0=ndipn[gid];
#define IPN(i,m) pipn[(i)+ndipn0*(m)]
	 
	 for( i=0; i<nipn[gid]; i++ )
	 {
	   fprintf(debugFile,"(%3i,%3i)",IPN(i,0),IPN(i,1));
	   if( i%10 == 9 ) fprintf(debugFile,"\n  ");
	 }
	 fprintf(debugFile,"\n");
	 
      }
    }
    
  }
  

  if( nipn!=NULL && nipn[gid]>0 )
  {
    int option=4;  // interpolation neighbours with Gauss-Seidel
    real omega=1.; // .8;

    int ipar[10];
    ipar[0]=numberOfLayers;
    int bc[6]={0,0,0,0,0,0}; 

    const int n1c=1, n2c=1, n3c=1;
    if( debug & 4 )
      printf("Smooth interp neighbours: level=%i grid=%i, number of neighbours=nipn[gid]=%i \n",
	     level,grid,nipn[gid]);

    if( false && Ogmg::debug & 1 )
    {
      char buff[100];
      display(u,sPrintF(buff,"smoothIN:level=%i,grid=%iHere is the solution BEFORE smoothIN, it %i",
			level,grid,0),debugFile,"%9.1e");
    }
    for( int iter=0; iter<numberOfIterations; iter++ )
    {
      smoothJacobiOpt( mg.numberOfDimensions(), d(0,0),d(1,0),d(0,1),d(1,1),d(0,2),d(1,2),
		       gir(0,0),gir(1,0),n1c,gir(0,1),gir(1,1),n2c,gir(0,2),gir(1,2),n3c, ndc,
		       *getDataPointer(f),
		       *getDataPointer(c),
		       *up, *vp,
		       *getDataPointer(mask), 
		       option, orderOfAccuracy, sparseStencil, constantCoefficients(0,grid,level), *vcp, dx[0], omega,
		       bc[0], nipn[gid],ndipn[gid],ipn[gid][0], ipar[0] );

      // maybe not applyBoundaryConditions( level,grid,u,f ); // this is needed if interp boundary hits physical boundary (hcic.order4)
      
      if( false && Ogmg::debug & 1 )
      {
	char buff[100];
	display(u,sPrintF(buff,"smoothIN:level=%i,grid=%iHere is the solution AFTER smoothIN, it %i",
			  level,grid,iter),debugFile,"%9.1e");
      }
      
    }
  }
  


  tm[timeForInterpolationSmooth]+=getCPU()-time0;
}

