#ifndef EXPOSED_POINTS_H
#define EXPOSED_POINTS_H

#include "Overture.h"

// This class is used to interpolate "exposed" points for moving grid computations.

class OGFunction;

class ExposedPoints
{
public:

  // There are different types of exposed points
  enum ExposedPointTypeEnum
  {
    exposedPointIsNeededForDiscretization,
    exposedDiscretization
  };

 enum InterpolationQualityEnum
 {
   canInterpolateQuality1=0,     // best quality, interpolates to correct order
   canInterpolateQuality2,       // 2nd best best quality, interpolates to correct order minus 1
   canInterpolateQuality3,       // 3rd best quality, interpolates to correct order minus 2
   canInterpolateWithExtrapolation, // bad quality -- should replace by a better quality result
   canInterpolateQualityBad,     // bad quality -- should replace by a better quality result
   canInterpolateQualityVeryBad, // worst quality-- should replace by a better quality result
   canNotInterpolate,             // unable to interpolate at all
   numberOfInterpolationQualityTypes
 };

ExposedPoints();
~ExposedPoints();

// choose the type of exposed points that are wanted (do this before calling initialize)
void setExposedPointType( const ExposedPointTypeEnum type );

// build the lists of exposed points; determine how to interpolate
int initialize(CompositeGrid& cg1, CompositeGrid & cg2,
	       int stencilWidth = -1  );

// interpolate values at the exposed points
int interpolate(realCompositeGridFunction & u1,
		OGFunction *TZFlow  =NULL,
		real t =0. );

// indicate whether interpolation neighbours are already assigned.
void setAssumeInterpolationNeighboursAreAssigned( const bool trueOrFalse=true );

// return the total number of exposed points (if grid=-1), otherwise return the number of exposed points on a grid
int getNumberOfExposedPoints(const int grid=-1 ) const;

static int debug;
static int info;

protected:

ExposedPointTypeEnum exposedPointType;

int numberOfExposedPoints;
int assumeInterpolationNeighboursAreAssigned; // default for stencilWidth=5

intArray ia_;
IntegerArray numberPerGrid;
realArray x_;

// realArray uInterpolated_;

int *numDonor;

intArray *ib;

intArray exposedInterpoleeGrid,
  exposedInterpoleeLocation,
  exposedInterpolationPoint,
  exposedVariableInterpolationWidth,
  exposedInterpolationQuality;

realArray exposedInterpolationCoordinates;

static int 
getInterpolationStencil(CompositeGrid & cg1,
                        const int grid2,
                        const int numToCheck, intArray & ia, intArray & ibg, realArray & r2, 
                        intArray & interpolationQuality,
                        intArray & interpoleeGrid,
                        intArray & interpoleeLocation,
                        intArray & interpolationPoint,
                        intArray & variableInterpolationWidth,
                        realArray & interpolationCoordinates );

static int 
checkForBetterQualityInterpolation( CompositeGrid & cg1, const int gridI, const int numToCheck, 
                                    intArray & ia, intArray & ic, 
                                    const realArray & x, 
				    intArray & interpolationQuality,
				    intArray & interpoleeGrid,
				    intArray & interpoleeLocation,
				    intArray & interpolationPoint,
				    intArray & variableInterpolationWidth,
				    realArray & interpolationCoordinates );

static int 
interpolatePoints(const realCompositeGridFunction & u,
		  realArray & uInterpolated_, 
		  int *numDonor, const intArray & ia_, const intArray *ib, 
		  const intArray & interpoleeGrid_,
		  const intArray & interpoleeLocation_,
		  const intArray & interpolationPoint_,
		  const intArray & variableInterpolationWidth_,
		  const realArray & interpolationCoordinates_,
		  const Range & R0 =nullRange,           
		  const Range & R1 =nullRange,
		  const Range & R2 =nullRange,
		  const Range & R3 =nullRange,
		  const Range & R4 =nullRange );

};



#endif
