! ------------------------------------------------------------------------------------
!  This file contains bpp macros that define the coefficient matricies for 
!  conservative operators: 
!     div( tensor grad), div( scalar grad),  Dx( scalar Dx), Dx( scalar Dy) ...
!  All these different operators are handled by the same macro below by appropriately 
!  defining the arguments a11, a12, etc. that appear in the macros. 
!   
!  This file, conservativeCoefficientMatrixMacros.h, was written by op/doc/consCoeff.maple
! ------------------------------------------------------------------------------------
! *** order of accuracy 2  ***

#beginMacro getDivTensorGradCoeffOrder2Dim2(coeff,dr)
 coeff(ma2(-1,-1, 0))=1./8.*(a12m1zz+a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-1, 0))=a22mh2/dr(1)**2*jac
 coeff(ma2( 1,-1, 0))=1./8.*(-a12p1zz-a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 0, 0))=a11mh2/dr(0)**2*jac
 coeff(ma2( 0, 0, 0))=((-a11ph2-a11mh2)/dr(0)**2+(-a22ph2-a22mh2)/dr(1)**2)*jac
 coeff(ma2( 1, 0, 0))=a11ph2/dr(0)**2*jac
 coeff(ma2(-1, 1, 0))=1./8.*(-a12m1zz-a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 1, 0))=a22ph2/dr(1)**2*jac
 coeff(ma2( 1, 1, 0))=1./8.*(a12p1zz+a12zp1z)/(dr(0)*dr(1))*jac
#endMacro


#beginMacro getDivTensorGradCoeffOrder2Dim3(coeff,dr)
 coeff(ma3(-1,-1,-1))=0.
 coeff(ma3( 0,-1,-1))=1./8.*(a23zm1z+a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-1))=0.
 coeff(ma3(-1, 0,-1))=1./8.*(a13m1zz+a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-1))=a33mh2/dr(2)**2*jac
 coeff(ma3( 1, 0,-1))=1./8.*(-a13p1zz-a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 1,-1))=0.
 coeff(ma3( 0, 1,-1))=1./8.*(-a23zp1z-a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-1))=0.
 coeff(ma3(-1,-1, 0))=1./8.*(a12m1zz+a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-1, 0))=a22mh2/dr(1)**2*jac
 coeff(ma3( 1,-1, 0))=1./8.*(-a12p1zz-a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 0, 0))=a11mh2/dr(0)**2*jac
 coeff(ma3( 0, 0, 0))=((-a11ph2-a11mh2)/dr(0)**2+(-a22ph2-a22mh2)/dr(1)**2+(-a33ph2-a33mh2)/dr(2)**2)*jac
 coeff(ma3( 1, 0, 0))=a11ph2/dr(0)**2*jac
 coeff(ma3(-1, 1, 0))=1./8.*(-a12m1zz-a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 1, 0))=a22ph2/dr(1)**2*jac
 coeff(ma3( 1, 1, 0))=1./8.*(a12p1zz+a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-1, 1))=0.
 coeff(ma3( 0,-1, 1))=1./8.*(-a23zm1z-a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 1))=0.
 coeff(ma3(-1, 0, 1))=1./8.*(-a13m1zz-a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 1))=a33ph2/dr(2)**2*jac
 coeff(ma3( 1, 0, 1))=1./8.*(a13p1zz+a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 1, 1))=0.
 coeff(ma3( 0, 1, 1))=1./8.*(a23zp1z+a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 1))=0.
#endMacro

! *** order of accuracy 4  ***

#beginMacro getDivTensorGradCoeffOrder4Dim2(coeff,dr)
 coeff(ma2(-2,-2, 0))=0.
 coeff(ma2(-1,-2, 0))=1./48.*(-a12zm2z-a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-2, 0))=1./24.*(-a22mh2-a22m3h2)/dr(1)**2*jac
 coeff(ma2( 1,-2, 0))=1./48.*(a12zm2z+a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-2, 0))=0.
 coeff(ma2(-2,-1, 0))=1./48.*(-a12m2zz-a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-1, 0))=1./48.*(10.*a12m1zz+10.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-1, 0))=1./24.*(24.*a22mh4+a22ph2+6.*a22mh2+a22m3h2)/dr(1)**2*jac
 coeff(ma2( 1,-1, 0))=1./48.*(-10.*a12p1zz-10.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-1, 0))=1./48.*(a12p2zz+a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2(-2, 0, 0))=1./24.*(-a11mh2-a11m3h2)/dr(0)**2*jac
 coeff(ma2(-1, 0, 0))=1./24.*(24.*a11mh4+a11ph2+6.*a11mh2+a11m3h2)/dr(0)**2*jac
 coeff(ma2( 0, 0, 0))=(1./24.*(-24.*a11ph4-24.*a11mh4-6.*a11ph2-6.*a11mh2)/dr(0)**2+1./24.*(-24.*a22ph4-24.*a22mh4-6.*a22ph2-6.*a22mh2)/dr(1)**2)*jac
 coeff(ma2( 1, 0, 0))=1./24.*(24.*a11ph4+6.*a11ph2+a11mh2+a11p3h2)/dr(0)**2*jac
 coeff(ma2( 2, 0, 0))=1./24.*(-a11ph2-a11p3h2)/dr(0)**2*jac
 coeff(ma2(-2, 1, 0))=1./48.*(a12m2zz+a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 1, 0))=1./48.*(-10.*a12m1zz-10.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 1, 0))=1./24.*(24.*a22ph4+6.*a22ph2+a22mh2+a22p3h2)/dr(1)**2*jac
 coeff(ma2( 1, 1, 0))=1./48.*(10.*a12p1zz+10.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 1, 0))=1./48.*(-a12p2zz-a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2(-2, 2, 0))=0.
 coeff(ma2(-1, 2, 0))=1./48.*(a12zp2z+a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 2, 0))=1./24.*(-a22ph2-a22p3h2)/dr(1)**2*jac
 coeff(ma2( 1, 2, 0))=1./48.*(-a12zp2z-a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 2, 0))=0.
#endMacro


#beginMacro getDivTensorGradCoeffOrder4Dim3(coeff,dr)
 coeff(ma3(-2,-2,-2))=0.
 coeff(ma3(-1,-2,-2))=0.
 coeff(ma3( 0,-2,-2))=0.
 coeff(ma3( 1,-2,-2))=0.
 coeff(ma3( 2,-2,-2))=0.
 coeff(ma3(-2,-1,-2))=0.
 coeff(ma3(-1,-1,-2))=0.
 coeff(ma3( 0,-1,-2))=1./48.*(-a23zzm2-a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-2))=0.
 coeff(ma3( 2,-1,-2))=0.
 coeff(ma3(-2, 0,-2))=0.
 coeff(ma3(-1, 0,-2))=1./48.*(-a13m1zz-a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-2))=1./24.*(-a33mh2-a33m3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-2))=1./48.*(a13zzm2+a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-2))=0.
 coeff(ma3(-2, 1,-2))=0.
 coeff(ma3(-1, 1,-2))=0.
 coeff(ma3( 0, 1,-2))=1./48.*(a23zzm2+a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-2))=0.
 coeff(ma3( 2, 1,-2))=0.
 coeff(ma3(-2, 2,-2))=0.
 coeff(ma3(-1, 2,-2))=0.
 coeff(ma3( 0, 2,-2))=0.
 coeff(ma3( 1, 2,-2))=0.
 coeff(ma3( 2, 2,-2))=0.
 coeff(ma3(-2,-2,-1))=0.
 coeff(ma3(-1,-2,-1))=0.
 coeff(ma3( 0,-2,-1))=1./48.*(-a23zzm1-a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-1))=0.
 coeff(ma3( 2,-2,-1))=0.
 coeff(ma3(-2,-1,-1))=0.
 coeff(ma3(-1,-1,-1))=0.
 coeff(ma3( 0,-1,-1))=1./48.*(10.*a23zm1z+10.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-1))=0.
 coeff(ma3( 2,-1,-1))=0.
 coeff(ma3(-2, 0,-1))=1./48.*(-a13m2zz-a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-1))=1./48.*(10.*a13m1zz+10.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-1))=1./24.*(24.*a33mh4+a33ph2+6.*a33mh2+a33m3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-1))=1./48.*(-10.*a13p1zz-10.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-1))=1./48.*(a13p2zz+a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 1,-1))=0.
 coeff(ma3(-1, 1,-1))=0.
 coeff(ma3( 0, 1,-1))=1./48.*(-10.*a23zp1z-10.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-1))=0.
 coeff(ma3( 2, 1,-1))=0.
 coeff(ma3(-2, 2,-1))=0.
 coeff(ma3(-1, 2,-1))=0.
 coeff(ma3( 0, 2,-1))=1./48.*(a23zzm1+a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-1))=0.
 coeff(ma3( 2, 2,-1))=0.
 coeff(ma3(-2,-2, 0))=0.
 coeff(ma3(-1,-2, 0))=1./48.*(-a12zm2z-a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-2, 0))=1./24.*(-a22mh2-a22m3h2)/dr(1)**2*jac
 coeff(ma3( 1,-2, 0))=1./48.*(a12zm2z+a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-2, 0))=0.
 coeff(ma3(-2,-1, 0))=1./48.*(-a12m2zz-a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-1, 0))=1./48.*(10.*a12m1zz+10.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-1, 0))=1./24.*(24.*a22mh4+a22ph2+6.*a22mh2+a22m3h2)/dr(1)**2*jac
 coeff(ma3( 1,-1, 0))=1./48.*(-10.*a12p1zz-10.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-1, 0))=1./48.*(a12p2zz+a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-2, 0, 0))=1./24.*(-a11mh2-a11m3h2)/dr(0)**2*jac
 coeff(ma3(-1, 0, 0))=1./24.*(24.*a11mh4+a11ph2+6.*a11mh2+a11m3h2)/dr(0)**2*jac
 coeff(ma3( 0, 0, 0))=(1./24.*(-24.*a11ph4-24.*a11mh4-6.*a11ph2-6.*a11mh2)/dr(0)**2+1./24.*(-24.*a22ph4-24.*a22mh4-6.*a22ph2-6.*a22mh2)/dr(1)**2+1./24.*(-24.*a33ph4-24.*a33mh4-6.*a33ph2-6.*a33mh2)/dr(2)**2)*jac
 coeff(ma3( 1, 0, 0))=1./24.*(24.*a11ph4+6.*a11ph2+a11mh2+a11p3h2)/dr(0)**2*jac
 coeff(ma3( 2, 0, 0))=1./24.*(-a11ph2-a11p3h2)/dr(0)**2*jac
 coeff(ma3(-2, 1, 0))=1./48.*(a12m2zz+a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 1, 0))=1./48.*(-10.*a12m1zz-10.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 1, 0))=1./24.*(24.*a22ph4+6.*a22ph2+a22mh2+a22p3h2)/dr(1)**2*jac
 coeff(ma3( 1, 1, 0))=1./48.*(10.*a12p1zz+10.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 1, 0))=1./48.*(-a12p2zz-a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-2, 2, 0))=0.
 coeff(ma3(-1, 2, 0))=1./48.*(a12zp2z+a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 2, 0))=1./24.*(-a22ph2-a22p3h2)/dr(1)**2*jac
 coeff(ma3( 1, 2, 0))=1./48.*(-a12zp2z-a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 2, 0))=0.
 coeff(ma3(-2,-2, 1))=0.
 coeff(ma3(-1,-2, 1))=0.
 coeff(ma3( 0,-2, 1))=1./48.*(a23zzp1+a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 1))=0.
 coeff(ma3( 2,-2, 1))=0.
 coeff(ma3(-2,-1, 1))=0.
 coeff(ma3(-1,-1, 1))=0.
 coeff(ma3( 0,-1, 1))=1./48.*(-10.*a23zm1z-10.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 1))=0.
 coeff(ma3( 2,-1, 1))=0.
 coeff(ma3(-2, 0, 1))=1./48.*(a13m2zz+a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 1))=1./48.*(-10.*a13m1zz-10.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 1))=1./24.*(24.*a33ph4+6.*a33ph2+a33mh2+a33p3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 1))=1./48.*(10.*a13p1zz+10.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 1))=1./48.*(-a13zzp1-a13p2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 1, 1))=0.
 coeff(ma3(-1, 1, 1))=0.
 coeff(ma3( 0, 1, 1))=1./48.*(10.*a23zp1z+10.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 1))=0.
 coeff(ma3( 2, 1, 1))=0.
 coeff(ma3(-2, 2, 1))=0.
 coeff(ma3(-1, 2, 1))=0.
 coeff(ma3( 0, 2, 1))=1./48.*(-a23zzp1-a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 1))=0.
 coeff(ma3( 2, 2, 1))=0.
 coeff(ma3(-2,-2, 2))=0.
 coeff(ma3(-1,-2, 2))=0.
 coeff(ma3( 0,-2, 2))=0.
 coeff(ma3( 1,-2, 2))=0.
 coeff(ma3( 2,-2, 2))=0.
 coeff(ma3(-2,-1, 2))=0.
 coeff(ma3(-1,-1, 2))=0.
 coeff(ma3( 0,-1, 2))=1./48.*(a23zzp2+a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 2))=0.
 coeff(ma3( 2,-1, 2))=0.
 coeff(ma3(-2, 0, 2))=0.
 coeff(ma3(-1, 0, 2))=1./48.*(a13zzp2+a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 2))=1./24.*(-a33ph2-a33p3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 2))=1./48.*(-a13p1zz-a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 2))=0.
 coeff(ma3(-2, 1, 2))=0.
 coeff(ma3(-1, 1, 2))=0.
 coeff(ma3( 0, 1, 2))=1./48.*(-a23zzp2-a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 2))=0.
 coeff(ma3( 2, 1, 2))=0.
 coeff(ma3(-2, 2, 2))=0.
 coeff(ma3(-1, 2, 2))=0.
 coeff(ma3( 0, 2, 2))=0.
 coeff(ma3( 1, 2, 2))=0.
 coeff(ma3( 2, 2, 2))=0.
#endMacro

! *** order of accuracy 6  ***

#beginMacro getDivTensorGradCoeffOrder6Dim2(coeff,dr)
 coeff(ma2(-3,-3, 0))=0.
 coeff(ma2(-2,-3, 0))=0.
 coeff(ma2(-1,-3, 0))=1./1440.*(6.*a12m1zz+6.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-3, 0))=1./5760.*(10.*a22m3h4+27.*a22mh2+27.*a22m5h2)/dr(1)**2*jac
 coeff(ma2( 1,-3, 0))=1./1440.*(-6.*a12zm3z-6.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-3, 0))=0.
 coeff(ma2( 3,-3, 0))=0.
 coeff(ma2(-3,-2, 0))=0.
 coeff(ma2(-2,-2, 0))=1./1440.*(5.*a12zm2z+5.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-2, 0))=1./1440.*(-64.*a12zm2z-64.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-2, 0))=1./5760.*(-270.*a22mh4-270.*a22m3h4-27.*a22ph2-27.*a22m5h2-135.*a22mh2-135.*a22m3h2)/dr(1)**2*jac
 coeff(ma2( 1,-2, 0))=1./1440.*(64.*a12zm2z+64.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-2, 0))=1./1440.*(-5.*a12p2zz-5.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 3,-2, 0))=0.
 coeff(ma2(-3,-1, 0))=1./1440.*(6.*a12zm1z+6.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2,-1, 0))=1./1440.*(-64.*a12zm1z-64.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-1, 0))=1./1440.*(380.*a12m1zz+380.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-1, 0))=1./5760.*(1530.*a22mh4+135.*a22ph2+135.*a22m3h2+540.*a22mh2+5760.*a22mh6+270.*a22ph4+270.*a22m3h4)/dr(1)**2*jac
 coeff(ma2( 1,-1, 0))=1./1440.*(-380.*a12p1zz-380.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-1, 0))=1./1440.*(64.*a12zm1z+64.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 3,-1, 0))=1./1440.*(-6.*a12zm1z-6.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-3, 0, 0))=1./5760.*(10.*a11m3h4+27.*a11mh2+27.*a11m5h2)/dr(0)**2*jac
 coeff(ma2(-2, 0, 0))=1./5760.*(-270.*a11mh4-270.*a11m3h4-27.*a11ph2-27.*a11m5h2-135.*a11mh2-135.*a11m3h2)/dr(0)**2*jac
 coeff(ma2(-1, 0, 0))=1./5760.*(5760.*a11mh6+270.*a11ph4+270.*a11m3h4+1530.*a11mh4+135.*a11ph2+135.*a11m3h2+540.*a11mh2)/dr(0)**2*jac
 coeff(ma2( 0, 0, 0))=(1./5760.*(-10.*a11m3h4-10.*a11p3h4-1530.*a11mh4-1530.*a11ph4-540.*a11mh2-540.*a11ph2-5760.*a11mh6-5760.*a11ph6)/dr(0)**2+1./5760.*(-5760.*a22mh6-5760.*a22ph6-10.*a22m3h4-10.*a22p3h4-1530.*a22mh4-1530.*a22ph4-540.*a22mh2-540.*a22ph2)/dr(1)**2)*jac
 coeff(ma2( 1, 0, 0))=1./5760.*(5760.*a11ph6+270.*a11mh4+270.*a11p3h4+1530.*a11ph4+135.*a11mh2+135.*a11p3h2+540.*a11ph2)/dr(0)**2*jac
 coeff(ma2( 2, 0, 0))=1./5760.*(-270.*a11ph4-270.*a11p3h4-27.*a11mh2-27.*a11p5h2-135.*a11ph2-135.*a11p3h2)/dr(0)**2*jac
 coeff(ma2( 3, 0, 0))=1./5760.*(10.*a11p3h4+27.*a11ph2+27.*a11p5h2)/dr(0)**2*jac
 coeff(ma2(-3, 1, 0))=1./1440.*(-6.*a12zp1z-6.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2, 1, 0))=1./1440.*(64.*a12zp1z+64.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 1, 0))=1./1440.*(-380.*a12m1zz-380.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 1, 0))=1./5760.*(270.*a22p3h4+1530.*a22ph4+135.*a22mh2+135.*a22p3h2+5760.*a22ph6+540.*a22ph2+270.*a22mh4)/dr(1)**2*jac
 coeff(ma2( 1, 1, 0))=1./1440.*(380.*a12p1zz+380.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 1, 0))=1./1440.*(-64.*a12zp1z-64.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 3, 1, 0))=1./1440.*(6.*a12zp1z+6.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-3, 2, 0))=0.
 coeff(ma2(-2, 2, 0))=1./1440.*(-5.*a12m2zz-5.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 2, 0))=1./1440.*(64.*a12m1zz+64.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 2, 0))=1./5760.*(-270.*a22p3h4-27.*a22mh2-27.*a22p5h2-135.*a22ph2-135.*a22p3h2-270.*a22ph4)/dr(1)**2*jac
 coeff(ma2( 1, 2, 0))=1./1440.*(-64.*a12p1zz-64.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 2, 0))=1./1440.*(5.*a12zp2z+5.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 3, 2, 0))=0.
 coeff(ma2(-3, 3, 0))=0.
 coeff(ma2(-2, 3, 0))=0.
 coeff(ma2(-1, 3, 0))=1./1440.*(-6.*a12m1zz-6.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 3, 0))=1./5760.*(27.*a22ph2+27.*a22p5h2+10.*a22p3h4)/dr(1)**2*jac
 coeff(ma2( 1, 3, 0))=1./1440.*(6.*a12p1zz+6.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 3, 0))=0.
 coeff(ma2( 3, 3, 0))=0.
#endMacro


#beginMacro getDivTensorGradCoeffOrder6Dim3(coeff,dr)
 coeff(ma3(-3,-3,-3))=0.
 coeff(ma3(-2,-3,-3))=0.
 coeff(ma3(-1,-3,-3))=0.
 coeff(ma3( 0,-3,-3))=0.
 coeff(ma3( 1,-3,-3))=0.
 coeff(ma3( 2,-3,-3))=0.
 coeff(ma3( 3,-3,-3))=0.
 coeff(ma3(-3,-2,-3))=0.
 coeff(ma3(-2,-2,-3))=0.
 coeff(ma3(-1,-2,-3))=0.
 coeff(ma3( 0,-2,-3))=0.
 coeff(ma3( 1,-2,-3))=0.
 coeff(ma3( 2,-2,-3))=0.
 coeff(ma3( 3,-2,-3))=0.
 coeff(ma3(-3,-1,-3))=0.
 coeff(ma3(-2,-1,-3))=0.
 coeff(ma3(-1,-1,-3))=0.
 coeff(ma3( 0,-1,-3))=1./1440.*(6.*a23zzm3+6.*a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-3))=0.
 coeff(ma3( 2,-1,-3))=0.
 coeff(ma3( 3,-1,-3))=0.
 coeff(ma3(-3, 0,-3))=0.
 coeff(ma3(-2, 0,-3))=0.
 coeff(ma3(-1, 0,-3))=1./1440.*(6.*a13zzm3+6.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-3))=1./5760.*(10.*a33m3h4+27.*a33mh2+27.*a33m5h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-3))=1./1440.*(-6.*a13p1zz-6.*a13zzm3)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-3))=0.
 coeff(ma3( 3, 0,-3))=0.
 coeff(ma3(-3, 1,-3))=0.
 coeff(ma3(-2, 1,-3))=0.
 coeff(ma3(-1, 1,-3))=0.
 coeff(ma3( 0, 1,-3))=1./1440.*(-6.*a23zp1z-6.*a23zzm3)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-3))=0.
 coeff(ma3( 2, 1,-3))=0.
 coeff(ma3( 3, 1,-3))=0.
 coeff(ma3(-3, 2,-3))=0.
 coeff(ma3(-2, 2,-3))=0.
 coeff(ma3(-1, 2,-3))=0.
 coeff(ma3( 0, 2,-3))=0.
 coeff(ma3( 1, 2,-3))=0.
 coeff(ma3( 2, 2,-3))=0.
 coeff(ma3( 3, 2,-3))=0.
 coeff(ma3(-3, 3,-3))=0.
 coeff(ma3(-2, 3,-3))=0.
 coeff(ma3(-1, 3,-3))=0.
 coeff(ma3( 0, 3,-3))=0.
 coeff(ma3( 1, 3,-3))=0.
 coeff(ma3( 2, 3,-3))=0.
 coeff(ma3( 3, 3,-3))=0.
 coeff(ma3(-3,-3,-2))=0.
 coeff(ma3(-2,-3,-2))=0.
 coeff(ma3(-1,-3,-2))=0.
 coeff(ma3( 0,-3,-2))=0.
 coeff(ma3( 1,-3,-2))=0.
 coeff(ma3( 2,-3,-2))=0.
 coeff(ma3( 3,-3,-2))=0.
 coeff(ma3(-3,-2,-2))=0.
 coeff(ma3(-2,-2,-2))=0.
 coeff(ma3(-1,-2,-2))=0.
 coeff(ma3( 0,-2,-2))=1./1440.*(5.*a23zzm2+5.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-2))=0.
 coeff(ma3( 2,-2,-2))=0.
 coeff(ma3( 3,-2,-2))=0.
 coeff(ma3(-3,-1,-2))=0.
 coeff(ma3(-2,-1,-2))=0.
 coeff(ma3(-1,-1,-2))=0.
 coeff(ma3( 0,-1,-2))=1./1440.*(-64.*a23zm1z-64.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-2))=0.
 coeff(ma3( 2,-1,-2))=0.
 coeff(ma3( 3,-1,-2))=0.
 coeff(ma3(-3, 0,-2))=0.
 coeff(ma3(-2, 0,-2))=1./1440.*(5.*a13m2zz+5.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-2))=1./1440.*(-64.*a13zzm2-64.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-2))=1./5760.*(-270.*a33mh4-270.*a33m3h4-27.*a33ph2-27.*a33m5h2-135.*a33mh2-135.*a33m3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-2))=1./1440.*(64.*a13zzm2+64.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-2))=1./1440.*(-5.*a13p2zz-5.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0,-2))=0.
 coeff(ma3(-3, 1,-2))=0.
 coeff(ma3(-2, 1,-2))=0.
 coeff(ma3(-1, 1,-2))=0.
 coeff(ma3( 0, 1,-2))=1./1440.*(64.*a23zp1z+64.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-2))=0.
 coeff(ma3( 2, 1,-2))=0.
 coeff(ma3( 3, 1,-2))=0.
 coeff(ma3(-3, 2,-2))=0.
 coeff(ma3(-2, 2,-2))=0.
 coeff(ma3(-1, 2,-2))=0.
 coeff(ma3( 0, 2,-2))=1./1440.*(-5.*a23zzm2-5.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-2))=0.
 coeff(ma3( 2, 2,-2))=0.
 coeff(ma3( 3, 2,-2))=0.
 coeff(ma3(-3, 3,-2))=0.
 coeff(ma3(-2, 3,-2))=0.
 coeff(ma3(-1, 3,-2))=0.
 coeff(ma3( 0, 3,-2))=0.
 coeff(ma3( 1, 3,-2))=0.
 coeff(ma3( 2, 3,-2))=0.
 coeff(ma3( 3, 3,-2))=0.
 coeff(ma3(-3,-3,-1))=0.
 coeff(ma3(-2,-3,-1))=0.
 coeff(ma3(-1,-3,-1))=0.
 coeff(ma3( 0,-3,-1))=1./1440.*(6.*a23zzm1+6.*a23zm3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3,-1))=0.
 coeff(ma3( 2,-3,-1))=0.
 coeff(ma3( 3,-3,-1))=0.
 coeff(ma3(-3,-2,-1))=0.
 coeff(ma3(-2,-2,-1))=0.
 coeff(ma3(-1,-2,-1))=0.
 coeff(ma3( 0,-2,-1))=1./1440.*(-64.*a23zzm1-64.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-1))=0.
 coeff(ma3( 2,-2,-1))=0.
 coeff(ma3( 3,-2,-1))=0.
 coeff(ma3(-3,-1,-1))=0.
 coeff(ma3(-2,-1,-1))=0.
 coeff(ma3(-1,-1,-1))=0.
 coeff(ma3( 0,-1,-1))=1./1440.*(380.*a23zm1z+380.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-1))=0.
 coeff(ma3( 2,-1,-1))=0.
 coeff(ma3( 3,-1,-1))=0.
 coeff(ma3(-3, 0,-1))=1./1440.*(6.*a13zzm1+6.*a13m3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0,-1))=1./1440.*(-64.*a13zzm1-64.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-1))=1./1440.*(380.*a13m1zz+380.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-1))=1./5760.*(5760.*a33mh6+270.*a33ph4+270.*a33m3h4+1530.*a33mh4+135.*a33ph2+135.*a33m3h2+540.*a33mh2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-1))=1./1440.*(-380.*a13p1zz-380.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-1))=1./1440.*(64.*a13p2zz+64.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0,-1))=1./1440.*(-6.*a13zzm1-6.*a13p3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-3, 1,-1))=0.
 coeff(ma3(-2, 1,-1))=0.
 coeff(ma3(-1, 1,-1))=0.
 coeff(ma3( 0, 1,-1))=1./1440.*(-380.*a23zp1z-380.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-1))=0.
 coeff(ma3( 2, 1,-1))=0.
 coeff(ma3( 3, 1,-1))=0.
 coeff(ma3(-3, 2,-1))=0.
 coeff(ma3(-2, 2,-1))=0.
 coeff(ma3(-1, 2,-1))=0.
 coeff(ma3( 0, 2,-1))=1./1440.*(64.*a23zzm1+64.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-1))=0.
 coeff(ma3( 2, 2,-1))=0.
 coeff(ma3( 3, 2,-1))=0.
 coeff(ma3(-3, 3,-1))=0.
 coeff(ma3(-2, 3,-1))=0.
 coeff(ma3(-1, 3,-1))=0.
 coeff(ma3( 0, 3,-1))=1./1440.*(-6.*a23zzm1-6.*a23zp3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3,-1))=0.
 coeff(ma3( 2, 3,-1))=0.
 coeff(ma3( 3, 3,-1))=0.
 coeff(ma3(-3,-3, 0))=0.
 coeff(ma3(-2,-3, 0))=0.
 coeff(ma3(-1,-3, 0))=1./1440.*(6.*a12m1zz+6.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-3, 0))=1./5760.*(10.*a22m3h4+27.*a22mh2+27.*a22m5h2)/dr(1)**2*jac
 coeff(ma3( 1,-3, 0))=1./1440.*(-6.*a12zm3z-6.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-3, 0))=0.
 coeff(ma3( 3,-3, 0))=0.
 coeff(ma3(-3,-2, 0))=0.
 coeff(ma3(-2,-2, 0))=1./1440.*(5.*a12zm2z+5.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-2, 0))=1./1440.*(-64.*a12zm2z-64.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-2, 0))=1./5760.*(-270.*a22mh4-270.*a22m3h4-27.*a22ph2-27.*a22m5h2-135.*a22mh2-135.*a22m3h2)/dr(1)**2*jac
 coeff(ma3( 1,-2, 0))=1./1440.*(64.*a12zm2z+64.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-2, 0))=1./1440.*(-5.*a12p2zz-5.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 3,-2, 0))=0.
 coeff(ma3(-3,-1, 0))=1./1440.*(6.*a12zm1z+6.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-2,-1, 0))=1./1440.*(-64.*a12zm1z-64.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-1, 0))=1./1440.*(380.*a12m1zz+380.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-1, 0))=1./5760.*(1530.*a22mh4+135.*a22ph2+135.*a22m3h2+540.*a22mh2+5760.*a22mh6+270.*a22ph4+270.*a22m3h4)/dr(1)**2*jac
 coeff(ma3( 1,-1, 0))=1./1440.*(-380.*a12p1zz-380.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-1, 0))=1./1440.*(64.*a12zm1z+64.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 3,-1, 0))=1./1440.*(-6.*a12zm1z-6.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-3, 0, 0))=1./5760.*(10.*a11m3h4+27.*a11mh2+27.*a11m5h2)/dr(0)**2*jac
 coeff(ma3(-2, 0, 0))=1./5760.*(-270.*a11mh4-270.*a11m3h4-27.*a11ph2-27.*a11m5h2-135.*a11mh2-135.*a11m3h2)/dr(0)**2*jac
 coeff(ma3(-1, 0, 0))=1./5760.*(5760.*a11mh6+270.*a11ph4+270.*a11m3h4+1530.*a11mh4+135.*a11ph2+135.*a11m3h2+540.*a11mh2)/dr(0)**2*jac
 coeff(ma3( 0, 0, 0))=(1./5760.*(-10.*a11m3h4-10.*a11p3h4-1530.*a11mh4-1530.*a11ph4-540.*a11mh2-540.*a11ph2-5760.*a11mh6-5760.*a11ph6)/dr(0)**2+1./5760.*(-5760.*a22mh6-5760.*a22ph6-10.*a22m3h4-10.*a22p3h4-1530.*a22mh4-1530.*a22ph4-540.*a22mh2-540.*a22ph2)/dr(1)**2+1./5760.*(-10.*a33m3h4-5760.*a33mh6-10.*a33p3h4-5760.*a33ph6-1530.*a33mh4-1530.*a33ph4-540.*a33mh2-540.*a33ph2)/dr(2)**2)*jac
 coeff(ma3( 1, 0, 0))=1./5760.*(5760.*a11ph6+270.*a11mh4+270.*a11p3h4+1530.*a11ph4+135.*a11mh2+135.*a11p3h2+540.*a11ph2)/dr(0)**2*jac
 coeff(ma3( 2, 0, 0))=1./5760.*(-270.*a11ph4-270.*a11p3h4-27.*a11mh2-27.*a11p5h2-135.*a11ph2-135.*a11p3h2)/dr(0)**2*jac
 coeff(ma3( 3, 0, 0))=1./5760.*(10.*a11p3h4+27.*a11ph2+27.*a11p5h2)/dr(0)**2*jac
 coeff(ma3(-3, 1, 0))=1./1440.*(-6.*a12zp1z-6.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-2, 1, 0))=1./1440.*(64.*a12zp1z+64.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 1, 0))=1./1440.*(-380.*a12m1zz-380.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 1, 0))=1./5760.*(270.*a22p3h4+1530.*a22ph4+135.*a22mh2+135.*a22p3h2+5760.*a22ph6+540.*a22ph2+270.*a22mh4)/dr(1)**2*jac
 coeff(ma3( 1, 1, 0))=1./1440.*(380.*a12p1zz+380.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 1, 0))=1./1440.*(-64.*a12zp1z-64.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 3, 1, 0))=1./1440.*(6.*a12zp1z+6.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-3, 2, 0))=0.
 coeff(ma3(-2, 2, 0))=1./1440.*(-5.*a12m2zz-5.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 2, 0))=1./1440.*(64.*a12m1zz+64.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 2, 0))=1./5760.*(-270.*a22p3h4-27.*a22mh2-27.*a22p5h2-135.*a22ph2-135.*a22p3h2-270.*a22ph4)/dr(1)**2*jac
 coeff(ma3( 1, 2, 0))=1./1440.*(-64.*a12p1zz-64.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 2, 0))=1./1440.*(5.*a12zp2z+5.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 3, 2, 0))=0.
 coeff(ma3(-3, 3, 0))=0.
 coeff(ma3(-2, 3, 0))=0.
 coeff(ma3(-1, 3, 0))=1./1440.*(-6.*a12m1zz-6.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 3, 0))=1./5760.*(27.*a22ph2+27.*a22p5h2+10.*a22p3h4)/dr(1)**2*jac
 coeff(ma3( 1, 3, 0))=1./1440.*(6.*a12p1zz+6.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 3, 0))=0.
 coeff(ma3( 3, 3, 0))=0.
 coeff(ma3(-3,-3, 1))=0.
 coeff(ma3(-2,-3, 1))=0.
 coeff(ma3(-1,-3, 1))=0.
 coeff(ma3( 0,-3, 1))=1./1440.*(-6.*a23zm3z-6.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3, 1))=0.
 coeff(ma3( 2,-3, 1))=0.
 coeff(ma3( 3,-3, 1))=0.
 coeff(ma3(-3,-2, 1))=0.
 coeff(ma3(-2,-2, 1))=0.
 coeff(ma3(-1,-2, 1))=0.
 coeff(ma3( 0,-2, 1))=1./1440.*(64.*a23zzp1+64.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 1))=0.
 coeff(ma3( 2,-2, 1))=0.
 coeff(ma3( 3,-2, 1))=0.
 coeff(ma3(-3,-1, 1))=0.
 coeff(ma3(-2,-1, 1))=0.
 coeff(ma3(-1,-1, 1))=0.
 coeff(ma3( 0,-1, 1))=1./1440.*(-380.*a23zm1z-380.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 1))=0.
 coeff(ma3( 2,-1, 1))=0.
 coeff(ma3( 3,-1, 1))=0.
 coeff(ma3(-3, 0, 1))=1./1440.*(-6.*a13zzp1-6.*a13m3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0, 1))=1./1440.*(64.*a13zzp1+64.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 1))=1./1440.*(-380.*a13m1zz-380.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 1))=1./5760.*(5760.*a33ph6+270.*a33mh4+270.*a33p3h4+1530.*a33ph4+135.*a33mh2+135.*a33p3h2+540.*a33ph2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 1))=1./1440.*(380.*a13p1zz+380.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 1))=1./1440.*(-64.*a13p2zz-64.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0, 1))=1./1440.*(6.*a13zzp1+6.*a13p3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-3, 1, 1))=0.
 coeff(ma3(-2, 1, 1))=0.
 coeff(ma3(-1, 1, 1))=0.
 coeff(ma3( 0, 1, 1))=1./1440.*(380.*a23zp1z+380.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 1))=0.
 coeff(ma3( 2, 1, 1))=0.
 coeff(ma3( 3, 1, 1))=0.
 coeff(ma3(-3, 2, 1))=0.
 coeff(ma3(-2, 2, 1))=0.
 coeff(ma3(-1, 2, 1))=0.
 coeff(ma3( 0, 2, 1))=1./1440.*(-64.*a23zzp1-64.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 1))=0.
 coeff(ma3( 2, 2, 1))=0.
 coeff(ma3( 3, 2, 1))=0.
 coeff(ma3(-3, 3, 1))=0.
 coeff(ma3(-2, 3, 1))=0.
 coeff(ma3(-1, 3, 1))=0.
 coeff(ma3( 0, 3, 1))=1./1440.*(6.*a23zp3z+6.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3, 1))=0.
 coeff(ma3( 2, 3, 1))=0.
 coeff(ma3( 3, 3, 1))=0.
 coeff(ma3(-3,-3, 2))=0.
 coeff(ma3(-2,-3, 2))=0.
 coeff(ma3(-1,-3, 2))=0.
 coeff(ma3( 0,-3, 2))=0.
 coeff(ma3( 1,-3, 2))=0.
 coeff(ma3( 2,-3, 2))=0.
 coeff(ma3( 3,-3, 2))=0.
 coeff(ma3(-3,-2, 2))=0.
 coeff(ma3(-2,-2, 2))=0.
 coeff(ma3(-1,-2, 2))=0.
 coeff(ma3( 0,-2, 2))=1./1440.*(-5.*a23zzp2-5.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 2))=0.
 coeff(ma3( 2,-2, 2))=0.
 coeff(ma3( 3,-2, 2))=0.
 coeff(ma3(-3,-1, 2))=0.
 coeff(ma3(-2,-1, 2))=0.
 coeff(ma3(-1,-1, 2))=0.
 coeff(ma3( 0,-1, 2))=1./1440.*(64.*a23zm1z+64.*a23zzp2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 2))=0.
 coeff(ma3( 2,-1, 2))=0.
 coeff(ma3( 3,-1, 2))=0.
 coeff(ma3(-3, 0, 2))=0.
 coeff(ma3(-2, 0, 2))=1./1440.*(-5.*a13m2zz-5.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 2))=1./1440.*(64.*a13zzp2+64.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 2))=1./5760.*(-270.*a33ph4-270.*a33p3h4-27.*a33mh2-27.*a33p5h2-135.*a33ph2-135.*a33p3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 2))=1./1440.*(-64.*a13p1zz-64.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 2))=1./1440.*(5.*a13p2zz+5.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0, 2))=0.
 coeff(ma3(-3, 1, 2))=0.
 coeff(ma3(-2, 1, 2))=0.
 coeff(ma3(-1, 1, 2))=0.
 coeff(ma3( 0, 1, 2))=1./1440.*(-64.*a23zp1z-64.*a23zzp2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 2))=0.
 coeff(ma3( 2, 1, 2))=0.
 coeff(ma3( 3, 1, 2))=0.
 coeff(ma3(-3, 2, 2))=0.
 coeff(ma3(-2, 2, 2))=0.
 coeff(ma3(-1, 2, 2))=0.
 coeff(ma3( 0, 2, 2))=1./1440.*(5.*a23zzp2+5.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 2))=0.
 coeff(ma3( 2, 2, 2))=0.
 coeff(ma3( 3, 2, 2))=0.
 coeff(ma3(-3, 3, 2))=0.
 coeff(ma3(-2, 3, 2))=0.
 coeff(ma3(-1, 3, 2))=0.
 coeff(ma3( 0, 3, 2))=0.
 coeff(ma3( 1, 3, 2))=0.
 coeff(ma3( 2, 3, 2))=0.
 coeff(ma3( 3, 3, 2))=0.
 coeff(ma3(-3,-3, 3))=0.
 coeff(ma3(-2,-3, 3))=0.
 coeff(ma3(-1,-3, 3))=0.
 coeff(ma3( 0,-3, 3))=0.
 coeff(ma3( 1,-3, 3))=0.
 coeff(ma3( 2,-3, 3))=0.
 coeff(ma3( 3,-3, 3))=0.
 coeff(ma3(-3,-2, 3))=0.
 coeff(ma3(-2,-2, 3))=0.
 coeff(ma3(-1,-2, 3))=0.
 coeff(ma3( 0,-2, 3))=0.
 coeff(ma3( 1,-2, 3))=0.
 coeff(ma3( 2,-2, 3))=0.
 coeff(ma3( 3,-2, 3))=0.
 coeff(ma3(-3,-1, 3))=0.
 coeff(ma3(-2,-1, 3))=0.
 coeff(ma3(-1,-1, 3))=0.
 coeff(ma3( 0,-1, 3))=1./1440.*(-6.*a23zm1z-6.*a23zzp3)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 3))=0.
 coeff(ma3( 2,-1, 3))=0.
 coeff(ma3( 3,-1, 3))=0.
 coeff(ma3(-3, 0, 3))=0.
 coeff(ma3(-2, 0, 3))=0.
 coeff(ma3(-1, 0, 3))=1./1440.*(-6.*a13zzp3-6.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 3))=1./5760.*(10.*a33p3h4+27.*a33ph2+27.*a33p5h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 3))=1./1440.*(6.*a13zzp3+6.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 3))=0.
 coeff(ma3( 3, 0, 3))=0.
 coeff(ma3(-3, 1, 3))=0.
 coeff(ma3(-2, 1, 3))=0.
 coeff(ma3(-1, 1, 3))=0.
 coeff(ma3( 0, 1, 3))=1./1440.*(6.*a23zzp3+6.*a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 3))=0.
 coeff(ma3( 2, 1, 3))=0.
 coeff(ma3( 3, 1, 3))=0.
 coeff(ma3(-3, 2, 3))=0.
 coeff(ma3(-2, 2, 3))=0.
 coeff(ma3(-1, 2, 3))=0.
 coeff(ma3( 0, 2, 3))=0.
 coeff(ma3( 1, 2, 3))=0.
 coeff(ma3( 2, 2, 3))=0.
 coeff(ma3( 3, 2, 3))=0.
 coeff(ma3(-3, 3, 3))=0.
 coeff(ma3(-2, 3, 3))=0.
 coeff(ma3(-1, 3, 3))=0.
 coeff(ma3( 0, 3, 3))=0.
 coeff(ma3( 1, 3, 3))=0.
 coeff(ma3( 2, 3, 3))=0.
 coeff(ma3( 3, 3, 3))=0.
#endMacro

! *** order of accuracy 8  ***

#beginMacro getDivTensorGradCoeffOrder8Dim2(coeff,dr)
 coeff(ma2(-4,-4, 0))=0.
 coeff(ma2(-3,-4, 0))=0.
 coeff(ma2(-2,-4, 0))=0.
 coeff(ma2(-1,-4, 0))=1./10080.*(-9.*a12m1zz-9.*a12zm4z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-4, 0))=1./322560.*(-63.*a22m3h2-225.*a22m7h2-63.*a22m5h2-225.*a22mh2)/dr(1)**2*jac
 coeff(ma2( 1,-4, 0))=1./10080.*(9.*a12p1zz+9.*a12zm4z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-4, 0))=0.
 coeff(ma2( 3,-4, 0))=0.
 coeff(ma2( 4,-4, 0))=0.
 coeff(ma2(-4,-3, 0))=0.
 coeff(ma2(-3,-3, 0))=0.
 coeff(ma2(-2,-3, 0))=1./10080.*(-7.*a12m2zz-7.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-3, 0))=1./10080.*(110.*a12m1zz+110.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-3, 0))=1./322560.*(560.*a22m3h6+225.*a22ph2+1512.*a22mh4+1512.*a22m5h4+630.*a22m3h2+1764.*a22mh2+1764.*a22m5h2+225.*a22m7h2)/dr(1)**2*jac
 coeff(ma2( 1,-3, 0))=1./10080.*(-110.*a12p1zz-110.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-3, 0))=1./10080.*(7.*a12p2zz+7.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 3,-3, 0))=0.
 coeff(ma2( 4,-3, 0))=0.
 coeff(ma2(-4,-2, 0))=0.
 coeff(ma2(-3,-2, 0))=1./10080.*(-7.*a12zm2z-7.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2,-2, 0))=1./10080.*(91.*a12zm2z+91.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-2, 0))=1./10080.*(-665.*a12m1zz-665.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-2, 0))=1./322560.*(-15120.*a22mh6-1512.*a22m5h4-7560.*a22mh4-7560.*a22m3h4-1764.*a22ph2-1764.*a22m5h2-1512.*a22ph4-6300.*a22mh2-6300.*a22m3h2-15120.*a22m3h6)/dr(1)**2*jac
 coeff(ma2( 1,-2, 0))=1./10080.*(665.*a12zm2z+665.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-2, 0))=1./10080.*(-91.*a12p2zz-91.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 3,-2, 0))=1./10080.*(7.*a12p3zz+7.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 4,-2, 0))=0.
 coeff(ma2(-4,-1, 0))=1./10080.*(-9.*a12zm1z-9.*a12m4zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-3,-1, 0))=1./10080.*(110.*a12zm1z+110.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2,-1, 0))=1./10080.*(-665.*a12zm1z-665.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1,-1, 0))=1./10080.*(3052.*a12m1zz+3052.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0,-1, 0))=1./322560.*(63.*a22p3h2+63.*a22m5h2+322560.*a22mh8+30240.*a22mh4+6300.*a22ph2+6300.*a22m3h2+15120.*a22m3h6+15120.*a22ph6+19530.*a22mh2+85680.*a22mh6+7560.*a22ph4+7560.*a22m3h4)/dr(1)**2*jac
 coeff(ma2( 1,-1, 0))=1./10080.*(-3052.*a12p1zz-3052.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2,-1, 0))=1./10080.*(665.*a12zm1z+665.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 3,-1, 0))=1./10080.*(-110.*a12zm1z-110.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 4,-1, 0))=1./10080.*(9.*a12zm1z+9.*a12p4zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-4, 0, 0))=1./322560.*(-225.*a11m7h2-225.*a11mh2-63.*a11m5h2-63.*a11m3h2)/dr(0)**2*jac
 coeff(ma2(-3, 0, 0))=1./322560.*(225.*a11m7h2+1512.*a11mh4+1764.*a11mh2+1764.*a11m5h2+225.*a11ph2+630.*a11m3h2+1512.*a11m5h4+560.*a11m3h6)/dr(0)**2*jac
 coeff(ma2(-2, 0, 0))=1./322560.*(-15120.*a11mh6-1512.*a11ph4-7560.*a11mh4-7560.*a11m3h4-1764.*a11ph2-1764.*a11m5h2-1512.*a11m5h4-6300.*a11mh2-6300.*a11m3h2-15120.*a11m3h6)/dr(0)**2*jac
 coeff(ma2(-1, 0, 0))=1./322560.*(85680.*a11mh6+7560.*a11ph4+7560.*a11m3h4+322560.*a11mh8+63.*a11p3h2+15120.*a11ph6+63.*a11m5h2+30240.*a11mh4+6300.*a11ph2+6300.*a11m3h2+19530.*a11mh2+15120.*a11m3h6)/dr(0)**2*jac
 coeff(ma2( 0, 0, 0))=(1./322560.*(-630.*a11m3h2-630.*a11p3h2-19530.*a11mh2-19530.*a11ph2-560.*a11p3h6-560.*a11m3h6-30240.*a11mh4-30240.*a11ph4-85680.*a11mh6-85680.*a11ph6-322560.*a11mh8-322560.*a11ph8)/dr(0)**2+1./322560.*(-630.*a22m3h2-630.*a22p3h2-322560.*a22mh8-322560.*a22ph8-30240.*a22mh4-30240.*a22ph4-560.*a22m3h6-19530.*a22mh2-19530.*a22ph2-85680.*a22mh6-85680.*a22ph6-560.*a22p3h6)/dr(1)**2)*jac
 coeff(ma2( 1, 0, 0))=1./322560.*(322560.*a11ph8+63.*a11p5h2+30240.*a11ph4+6300.*a11mh2+6300.*a11p3h2+63.*a11m3h2+19530.*a11ph2+15120.*a11p3h6+15120.*a11mh6+85680.*a11ph6+7560.*a11mh4+7560.*a11p3h4)/dr(0)**2*jac
 coeff(ma2( 2, 0, 0))=1./322560.*(-7560.*a11ph4-7560.*a11p3h4-1764.*a11mh2-1764.*a11p5h2-15120.*a11ph6-6300.*a11ph2-6300.*a11p3h2-15120.*a11p3h6-1512.*a11mh4-1512.*a11p5h4)/dr(0)**2*jac
 coeff(ma2( 3, 0, 0))=1./322560.*(1764.*a11ph2+1764.*a11p5h2+225.*a11p7h2+225.*a11mh2+630.*a11p3h2+1512.*a11p5h4+560.*a11p3h6+1512.*a11ph4)/dr(0)**2*jac
 coeff(ma2( 4, 0, 0))=1./322560.*(-225.*a11p7h2-225.*a11ph2-63.*a11p5h2-63.*a11p3h2)/dr(0)**2*jac
 coeff(ma2(-4, 1, 0))=1./10080.*(9.*a12zp1z+9.*a12m4zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-3, 1, 0))=1./10080.*(-110.*a12zp1z-110.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2, 1, 0))=1./10080.*(665.*a12zp1z+665.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 1, 0))=1./10080.*(-3052.*a12m1zz-3052.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 1, 0))=1./322560.*(322560.*a22ph8+63.*a22p5h2+30240.*a22ph4+6300.*a22mh2+6300.*a22p3h2+85680.*a22ph6+19530.*a22ph2+15120.*a22p3h6+63.*a22m3h2+7560.*a22mh4+7560.*a22p3h4+15120.*a22mh6)/dr(1)**2*jac
 coeff(ma2( 1, 1, 0))=1./10080.*(3052.*a12p1zz+3052.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 1, 0))=1./10080.*(-665.*a12zp1z-665.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 3, 1, 0))=1./10080.*(110.*a12zp1z+110.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 4, 1, 0))=1./10080.*(-9.*a12zp1z-9.*a12p4zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-4, 2, 0))=0.
 coeff(ma2(-3, 2, 0))=1./10080.*(7.*a12zp2z+7.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-2, 2, 0))=1./10080.*(-91.*a12m2zz-91.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 2, 0))=1./10080.*(665.*a12m1zz+665.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 2, 0))=1./322560.*(-1512.*a22mh4-15120.*a22ph6-7560.*a22p3h4-1764.*a22mh2-1764.*a22p5h2-6300.*a22ph2-6300.*a22p3h2-7560.*a22ph4-1512.*a22p5h4-15120.*a22p3h6)/dr(1)**2*jac
 coeff(ma2( 1, 2, 0))=1./10080.*(-665.*a12p1zz-665.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 2, 0))=1./10080.*(91.*a12p2zz+91.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 3, 2, 0))=1./10080.*(-7.*a12p3zz-7.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma2( 4, 2, 0))=0.
 coeff(ma2(-4, 3, 0))=0.
 coeff(ma2(-3, 3, 0))=0.
 coeff(ma2(-2, 3, 0))=1./10080.*(7.*a12zp3z+7.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma2(-1, 3, 0))=1./10080.*(-110.*a12m1zz-110.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 3, 0))=1./322560.*(1764.*a22ph2+1764.*a22p5h2+225.*a22p7h2+630.*a22p3h2+1512.*a22ph4+225.*a22mh2+1512.*a22p5h4+560.*a22p3h6)/dr(1)**2*jac
 coeff(ma2( 1, 3, 0))=1./10080.*(110.*a12p1zz+110.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 3, 0))=1./10080.*(-7.*a12p2zz-7.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma2( 3, 3, 0))=0.
 coeff(ma2( 4, 3, 0))=0.
 coeff(ma2(-4, 4, 0))=0.
 coeff(ma2(-3, 4, 0))=0.
 coeff(ma2(-2, 4, 0))=0.
 coeff(ma2(-1, 4, 0))=1./10080.*(9.*a12zp4z+9.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 0, 4, 0))=1./322560.*(-225.*a22ph2-225.*a22p7h2-63.*a22p3h2-63.*a22p5h2)/dr(1)**2*jac
 coeff(ma2( 1, 4, 0))=1./10080.*(-9.*a12zp4z-9.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma2( 2, 4, 0))=0.
 coeff(ma2( 3, 4, 0))=0.
 coeff(ma2( 4, 4, 0))=0.
#endMacro


#beginMacro getDivTensorGradCoeffOrder8Dim3(coeff,dr)
 coeff(ma3(-4,-4,-4))=0.
 coeff(ma3(-3,-4,-4))=0.
 coeff(ma3(-2,-4,-4))=0.
 coeff(ma3(-1,-4,-4))=0.
 coeff(ma3( 0,-4,-4))=0.
 coeff(ma3( 1,-4,-4))=0.
 coeff(ma3( 2,-4,-4))=0.
 coeff(ma3( 3,-4,-4))=0.
 coeff(ma3( 4,-4,-4))=0.
 coeff(ma3(-4,-3,-4))=0.
 coeff(ma3(-3,-3,-4))=0.
 coeff(ma3(-2,-3,-4))=0.
 coeff(ma3(-1,-3,-4))=0.
 coeff(ma3( 0,-3,-4))=0.
 coeff(ma3( 1,-3,-4))=0.
 coeff(ma3( 2,-3,-4))=0.
 coeff(ma3( 3,-3,-4))=0.
 coeff(ma3( 4,-3,-4))=0.
 coeff(ma3(-4,-2,-4))=0.
 coeff(ma3(-3,-2,-4))=0.
 coeff(ma3(-2,-2,-4))=0.
 coeff(ma3(-1,-2,-4))=0.
 coeff(ma3( 0,-2,-4))=0.
 coeff(ma3( 1,-2,-4))=0.
 coeff(ma3( 2,-2,-4))=0.
 coeff(ma3( 3,-2,-4))=0.
 coeff(ma3( 4,-2,-4))=0.
 coeff(ma3(-4,-1,-4))=0.
 coeff(ma3(-3,-1,-4))=0.
 coeff(ma3(-2,-1,-4))=0.
 coeff(ma3(-1,-1,-4))=0.
 coeff(ma3( 0,-1,-4))=1./10080.*(-9.*a23zm1z-9.*a23zzm4)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-4))=0.
 coeff(ma3( 2,-1,-4))=0.
 coeff(ma3( 3,-1,-4))=0.
 coeff(ma3( 4,-1,-4))=0.
 coeff(ma3(-4, 0,-4))=0.
 coeff(ma3(-3, 0,-4))=0.
 coeff(ma3(-2, 0,-4))=0.
 coeff(ma3(-1, 0,-4))=1./10080.*(-9.*a13zzm4-9.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-4))=1./322560.*(-225.*a33mh2-63.*a33m3h2-63.*a33m5h2-225.*a33m7h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-4))=1./10080.*(9.*a13zzm4+9.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-4))=0.
 coeff(ma3( 3, 0,-4))=0.
 coeff(ma3( 4, 0,-4))=0.
 coeff(ma3(-4, 1,-4))=0.
 coeff(ma3(-3, 1,-4))=0.
 coeff(ma3(-2, 1,-4))=0.
 coeff(ma3(-1, 1,-4))=0.
 coeff(ma3( 0, 1,-4))=1./10080.*(9.*a23zzm4+9.*a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-4))=0.
 coeff(ma3( 2, 1,-4))=0.
 coeff(ma3( 3, 1,-4))=0.
 coeff(ma3( 4, 1,-4))=0.
 coeff(ma3(-4, 2,-4))=0.
 coeff(ma3(-3, 2,-4))=0.
 coeff(ma3(-2, 2,-4))=0.
 coeff(ma3(-1, 2,-4))=0.
 coeff(ma3( 0, 2,-4))=0.
 coeff(ma3( 1, 2,-4))=0.
 coeff(ma3( 2, 2,-4))=0.
 coeff(ma3( 3, 2,-4))=0.
 coeff(ma3( 4, 2,-4))=0.
 coeff(ma3(-4, 3,-4))=0.
 coeff(ma3(-3, 3,-4))=0.
 coeff(ma3(-2, 3,-4))=0.
 coeff(ma3(-1, 3,-4))=0.
 coeff(ma3( 0, 3,-4))=0.
 coeff(ma3( 1, 3,-4))=0.
 coeff(ma3( 2, 3,-4))=0.
 coeff(ma3( 3, 3,-4))=0.
 coeff(ma3( 4, 3,-4))=0.
 coeff(ma3(-4, 4,-4))=0.
 coeff(ma3(-3, 4,-4))=0.
 coeff(ma3(-2, 4,-4))=0.
 coeff(ma3(-1, 4,-4))=0.
 coeff(ma3( 0, 4,-4))=0.
 coeff(ma3( 1, 4,-4))=0.
 coeff(ma3( 2, 4,-4))=0.
 coeff(ma3( 3, 4,-4))=0.
 coeff(ma3( 4, 4,-4))=0.
 coeff(ma3(-4,-4,-3))=0.
 coeff(ma3(-3,-4,-3))=0.
 coeff(ma3(-2,-4,-3))=0.
 coeff(ma3(-1,-4,-3))=0.
 coeff(ma3( 0,-4,-3))=0.
 coeff(ma3( 1,-4,-3))=0.
 coeff(ma3( 2,-4,-3))=0.
 coeff(ma3( 3,-4,-3))=0.
 coeff(ma3( 4,-4,-3))=0.
 coeff(ma3(-4,-3,-3))=0.
 coeff(ma3(-3,-3,-3))=0.
 coeff(ma3(-2,-3,-3))=0.
 coeff(ma3(-1,-3,-3))=0.
 coeff(ma3( 0,-3,-3))=0.
 coeff(ma3( 1,-3,-3))=0.
 coeff(ma3( 2,-3,-3))=0.
 coeff(ma3( 3,-3,-3))=0.
 coeff(ma3( 4,-3,-3))=0.
 coeff(ma3(-4,-2,-3))=0.
 coeff(ma3(-3,-2,-3))=0.
 coeff(ma3(-2,-2,-3))=0.
 coeff(ma3(-1,-2,-3))=0.
 coeff(ma3( 0,-2,-3))=1./10080.*(-7.*a23zzm3-7.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-3))=0.
 coeff(ma3( 2,-2,-3))=0.
 coeff(ma3( 3,-2,-3))=0.
 coeff(ma3( 4,-2,-3))=0.
 coeff(ma3(-4,-1,-3))=0.
 coeff(ma3(-3,-1,-3))=0.
 coeff(ma3(-2,-1,-3))=0.
 coeff(ma3(-1,-1,-3))=0.
 coeff(ma3( 0,-1,-3))=1./10080.*(110.*a23zzm3+110.*a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-3))=0.
 coeff(ma3( 2,-1,-3))=0.
 coeff(ma3( 3,-1,-3))=0.
 coeff(ma3( 4,-1,-3))=0.
 coeff(ma3(-4, 0,-3))=0.
 coeff(ma3(-3, 0,-3))=0.
 coeff(ma3(-2, 0,-3))=1./10080.*(-7.*a13zzm3-7.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-3))=1./10080.*(110.*a13m1zz+110.*a13zzm3)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-3))=1./322560.*(630.*a33m3h2+225.*a33ph2+1764.*a33mh2+1764.*a33m5h2+560.*a33m3h6+225.*a33m7h2+1512.*a33m5h4+1512.*a33mh4)/dr(2)**2*jac
 coeff(ma3( 1, 0,-3))=1./10080.*(-110.*a13p1zz-110.*a13zzm3)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-3))=1./10080.*(7.*a13zzm3+7.*a13p2zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0,-3))=0.
 coeff(ma3( 4, 0,-3))=0.
 coeff(ma3(-4, 1,-3))=0.
 coeff(ma3(-3, 1,-3))=0.
 coeff(ma3(-2, 1,-3))=0.
 coeff(ma3(-1, 1,-3))=0.
 coeff(ma3( 0, 1,-3))=1./10080.*(-110.*a23zzm3-110.*a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-3))=0.
 coeff(ma3( 2, 1,-3))=0.
 coeff(ma3( 3, 1,-3))=0.
 coeff(ma3( 4, 1,-3))=0.
 coeff(ma3(-4, 2,-3))=0.
 coeff(ma3(-3, 2,-3))=0.
 coeff(ma3(-2, 2,-3))=0.
 coeff(ma3(-1, 2,-3))=0.
 coeff(ma3( 0, 2,-3))=1./10080.*(7.*a23zp2z+7.*a23zzm3)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-3))=0.
 coeff(ma3( 2, 2,-3))=0.
 coeff(ma3( 3, 2,-3))=0.
 coeff(ma3( 4, 2,-3))=0.
 coeff(ma3(-4, 3,-3))=0.
 coeff(ma3(-3, 3,-3))=0.
 coeff(ma3(-2, 3,-3))=0.
 coeff(ma3(-1, 3,-3))=0.
 coeff(ma3( 0, 3,-3))=0.
 coeff(ma3( 1, 3,-3))=0.
 coeff(ma3( 2, 3,-3))=0.
 coeff(ma3( 3, 3,-3))=0.
 coeff(ma3( 4, 3,-3))=0.
 coeff(ma3(-4, 4,-3))=0.
 coeff(ma3(-3, 4,-3))=0.
 coeff(ma3(-2, 4,-3))=0.
 coeff(ma3(-1, 4,-3))=0.
 coeff(ma3( 0, 4,-3))=0.
 coeff(ma3( 1, 4,-3))=0.
 coeff(ma3( 2, 4,-3))=0.
 coeff(ma3( 3, 4,-3))=0.
 coeff(ma3( 4, 4,-3))=0.
 coeff(ma3(-4,-4,-2))=0.
 coeff(ma3(-3,-4,-2))=0.
 coeff(ma3(-2,-4,-2))=0.
 coeff(ma3(-1,-4,-2))=0.
 coeff(ma3( 0,-4,-2))=0.
 coeff(ma3( 1,-4,-2))=0.
 coeff(ma3( 2,-4,-2))=0.
 coeff(ma3( 3,-4,-2))=0.
 coeff(ma3( 4,-4,-2))=0.
 coeff(ma3(-4,-3,-2))=0.
 coeff(ma3(-3,-3,-2))=0.
 coeff(ma3(-2,-3,-2))=0.
 coeff(ma3(-1,-3,-2))=0.
 coeff(ma3( 0,-3,-2))=1./10080.*(-7.*a23zm3z-7.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3,-2))=0.
 coeff(ma3( 2,-3,-2))=0.
 coeff(ma3( 3,-3,-2))=0.
 coeff(ma3( 4,-3,-2))=0.
 coeff(ma3(-4,-2,-2))=0.
 coeff(ma3(-3,-2,-2))=0.
 coeff(ma3(-2,-2,-2))=0.
 coeff(ma3(-1,-2,-2))=0.
 coeff(ma3( 0,-2,-2))=1./10080.*(91.*a23zzm2+91.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-2))=0.
 coeff(ma3( 2,-2,-2))=0.
 coeff(ma3( 3,-2,-2))=0.
 coeff(ma3( 4,-2,-2))=0.
 coeff(ma3(-4,-1,-2))=0.
 coeff(ma3(-3,-1,-2))=0.
 coeff(ma3(-2,-1,-2))=0.
 coeff(ma3(-1,-1,-2))=0.
 coeff(ma3( 0,-1,-2))=1./10080.*(-665.*a23zzm2-665.*a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-2))=0.
 coeff(ma3( 2,-1,-2))=0.
 coeff(ma3( 3,-1,-2))=0.
 coeff(ma3( 4,-1,-2))=0.
 coeff(ma3(-4, 0,-2))=0.
 coeff(ma3(-3, 0,-2))=1./10080.*(-7.*a13m3zz-7.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0,-2))=1./10080.*(91.*a13m2zz+91.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-2))=1./10080.*(-665.*a13m1zz-665.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-2))=1./322560.*(-15120.*a33mh6-6300.*a33mh2-6300.*a33m3h2-1512.*a33ph4-1764.*a33m5h2-15120.*a33m3h6-7560.*a33mh4-7560.*a33m3h4-1764.*a33ph2-1512.*a33m5h4)/dr(2)**2*jac
 coeff(ma3( 1, 0,-2))=1./10080.*(665.*a13p1zz+665.*a13zzm2)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-2))=1./10080.*(-91.*a13zzm2-91.*a13p2zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0,-2))=1./10080.*(7.*a13zzm2+7.*a13p3zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 4, 0,-2))=0.
 coeff(ma3(-4, 1,-2))=0.
 coeff(ma3(-3, 1,-2))=0.
 coeff(ma3(-2, 1,-2))=0.
 coeff(ma3(-1, 1,-2))=0.
 coeff(ma3( 0, 1,-2))=1./10080.*(665.*a23zp1z+665.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-2))=0.
 coeff(ma3( 2, 1,-2))=0.
 coeff(ma3( 3, 1,-2))=0.
 coeff(ma3( 4, 1,-2))=0.
 coeff(ma3(-4, 2,-2))=0.
 coeff(ma3(-3, 2,-2))=0.
 coeff(ma3(-2, 2,-2))=0.
 coeff(ma3(-1, 2,-2))=0.
 coeff(ma3( 0, 2,-2))=1./10080.*(-91.*a23zp2z-91.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-2))=0.
 coeff(ma3( 2, 2,-2))=0.
 coeff(ma3( 3, 2,-2))=0.
 coeff(ma3( 4, 2,-2))=0.
 coeff(ma3(-4, 3,-2))=0.
 coeff(ma3(-3, 3,-2))=0.
 coeff(ma3(-2, 3,-2))=0.
 coeff(ma3(-1, 3,-2))=0.
 coeff(ma3( 0, 3,-2))=1./10080.*(7.*a23zp3z+7.*a23zzm2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3,-2))=0.
 coeff(ma3( 2, 3,-2))=0.
 coeff(ma3( 3, 3,-2))=0.
 coeff(ma3( 4, 3,-2))=0.
 coeff(ma3(-4, 4,-2))=0.
 coeff(ma3(-3, 4,-2))=0.
 coeff(ma3(-2, 4,-2))=0.
 coeff(ma3(-1, 4,-2))=0.
 coeff(ma3( 0, 4,-2))=0.
 coeff(ma3( 1, 4,-2))=0.
 coeff(ma3( 2, 4,-2))=0.
 coeff(ma3( 3, 4,-2))=0.
 coeff(ma3( 4, 4,-2))=0.
 coeff(ma3(-4,-4,-1))=0.
 coeff(ma3(-3,-4,-1))=0.
 coeff(ma3(-2,-4,-1))=0.
 coeff(ma3(-1,-4,-1))=0.
 coeff(ma3( 0,-4,-1))=1./10080.*(-9.*a23zzm1-9.*a23zm4z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-4,-1))=0.
 coeff(ma3( 2,-4,-1))=0.
 coeff(ma3( 3,-4,-1))=0.
 coeff(ma3( 4,-4,-1))=0.
 coeff(ma3(-4,-3,-1))=0.
 coeff(ma3(-3,-3,-1))=0.
 coeff(ma3(-2,-3,-1))=0.
 coeff(ma3(-1,-3,-1))=0.
 coeff(ma3( 0,-3,-1))=1./10080.*(110.*a23zzm1+110.*a23zm3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3,-1))=0.
 coeff(ma3( 2,-3,-1))=0.
 coeff(ma3( 3,-3,-1))=0.
 coeff(ma3( 4,-3,-1))=0.
 coeff(ma3(-4,-2,-1))=0.
 coeff(ma3(-3,-2,-1))=0.
 coeff(ma3(-2,-2,-1))=0.
 coeff(ma3(-1,-2,-1))=0.
 coeff(ma3( 0,-2,-1))=1./10080.*(-665.*a23zzm1-665.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2,-1))=0.
 coeff(ma3( 2,-2,-1))=0.
 coeff(ma3( 3,-2,-1))=0.
 coeff(ma3( 4,-2,-1))=0.
 coeff(ma3(-4,-1,-1))=0.
 coeff(ma3(-3,-1,-1))=0.
 coeff(ma3(-2,-1,-1))=0.
 coeff(ma3(-1,-1,-1))=0.
 coeff(ma3( 0,-1,-1))=1./10080.*(3052.*a23zm1z+3052.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1,-1))=0.
 coeff(ma3( 2,-1,-1))=0.
 coeff(ma3( 3,-1,-1))=0.
 coeff(ma3( 4,-1,-1))=0.
 coeff(ma3(-4, 0,-1))=1./10080.*(-9.*a13zzm1-9.*a13m4zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-3, 0,-1))=1./10080.*(110.*a13zzm1+110.*a13m3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0,-1))=1./10080.*(-665.*a13zzm1-665.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0,-1))=1./10080.*(3052.*a13m1zz+3052.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0,-1))=1./322560.*(6300.*a33ph2+6300.*a33m3h2+19530.*a33mh2+322560.*a33mh8+63.*a33m5h2+30240.*a33mh4+15120.*a33ph6+85680.*a33mh6+7560.*a33ph4+7560.*a33m3h4+15120.*a33m3h6+63.*a33p3h2)/dr(2)**2*jac
 coeff(ma3( 1, 0,-1))=1./10080.*(-3052.*a13p1zz-3052.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0,-1))=1./10080.*(665.*a13p2zz+665.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0,-1))=1./10080.*(-110.*a13zzm1-110.*a13p3zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 4, 0,-1))=1./10080.*(9.*a13p4zz+9.*a13zzm1)/(dr(0)*dr(2))*jac
 coeff(ma3(-4, 1,-1))=0.
 coeff(ma3(-3, 1,-1))=0.
 coeff(ma3(-2, 1,-1))=0.
 coeff(ma3(-1, 1,-1))=0.
 coeff(ma3( 0, 1,-1))=1./10080.*(-3052.*a23zp1z-3052.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1,-1))=0.
 coeff(ma3( 2, 1,-1))=0.
 coeff(ma3( 3, 1,-1))=0.
 coeff(ma3( 4, 1,-1))=0.
 coeff(ma3(-4, 2,-1))=0.
 coeff(ma3(-3, 2,-1))=0.
 coeff(ma3(-2, 2,-1))=0.
 coeff(ma3(-1, 2,-1))=0.
 coeff(ma3( 0, 2,-1))=1./10080.*(665.*a23zzm1+665.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2,-1))=0.
 coeff(ma3( 2, 2,-1))=0.
 coeff(ma3( 3, 2,-1))=0.
 coeff(ma3( 4, 2,-1))=0.
 coeff(ma3(-4, 3,-1))=0.
 coeff(ma3(-3, 3,-1))=0.
 coeff(ma3(-2, 3,-1))=0.
 coeff(ma3(-1, 3,-1))=0.
 coeff(ma3( 0, 3,-1))=1./10080.*(-110.*a23zzm1-110.*a23zp3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3,-1))=0.
 coeff(ma3( 2, 3,-1))=0.
 coeff(ma3( 3, 3,-1))=0.
 coeff(ma3( 4, 3,-1))=0.
 coeff(ma3(-4, 4,-1))=0.
 coeff(ma3(-3, 4,-1))=0.
 coeff(ma3(-2, 4,-1))=0.
 coeff(ma3(-1, 4,-1))=0.
 coeff(ma3( 0, 4,-1))=1./10080.*(9.*a23zp4z+9.*a23zzm1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 4,-1))=0.
 coeff(ma3( 2, 4,-1))=0.
 coeff(ma3( 3, 4,-1))=0.
 coeff(ma3( 4, 4,-1))=0.
 coeff(ma3(-4,-4, 0))=0.
 coeff(ma3(-3,-4, 0))=0.
 coeff(ma3(-2,-4, 0))=0.
 coeff(ma3(-1,-4, 0))=1./10080.*(-9.*a12m1zz-9.*a12zm4z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-4, 0))=1./322560.*(-63.*a22m3h2-225.*a22m7h2-63.*a22m5h2-225.*a22mh2)/dr(1)**2*jac
 coeff(ma3( 1,-4, 0))=1./10080.*(9.*a12zm4z+9.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-4, 0))=0.
 coeff(ma3( 3,-4, 0))=0.
 coeff(ma3( 4,-4, 0))=0.
 coeff(ma3(-4,-3, 0))=0.
 coeff(ma3(-3,-3, 0))=0.
 coeff(ma3(-2,-3, 0))=1./10080.*(-7.*a12m2zz-7.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-3, 0))=1./10080.*(110.*a12zm3z+110.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-3, 0))=1./322560.*(1512.*a22mh4+1512.*a22m5h4+630.*a22m3h2+1764.*a22mh2+225.*a22ph2+1764.*a22m5h2+225.*a22m7h2+560.*a22m3h6)/dr(1)**2*jac
 coeff(ma3( 1,-3, 0))=1./10080.*(-110.*a12zm3z-110.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-3, 0))=1./10080.*(7.*a12p2zz+7.*a12zm3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 3,-3, 0))=0.
 coeff(ma3( 4,-3, 0))=0.
 coeff(ma3(-4,-2, 0))=0.
 coeff(ma3(-3,-2, 0))=1./10080.*(-7.*a12m3zz-7.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma3(-2,-2, 0))=1./10080.*(91.*a12zm2z+91.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-2, 0))=1./10080.*(-665.*a12zm2z-665.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-2, 0))=1./322560.*(-1764.*a22ph2-1764.*a22m5h2-15120.*a22mh6-1512.*a22m5h4-7560.*a22mh4-7560.*a22m3h4-1512.*a22ph4-6300.*a22mh2-6300.*a22m3h2-15120.*a22m3h6)/dr(1)**2*jac
 coeff(ma3( 1,-2, 0))=1./10080.*(665.*a12p1zz+665.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-2, 0))=1./10080.*(-91.*a12p2zz-91.*a12zm2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 3,-2, 0))=1./10080.*(7.*a12zm2z+7.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 4,-2, 0))=0.
 coeff(ma3(-4,-1, 0))=1./10080.*(-9.*a12zm1z-9.*a12m4zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-3,-1, 0))=1./10080.*(110.*a12m3zz+110.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3(-2,-1, 0))=1./10080.*(-665.*a12zm1z-665.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1,-1, 0))=1./10080.*(3052.*a12m1zz+3052.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0,-1, 0))=1./322560.*(85680.*a22mh6+7560.*a22ph4+7560.*a22m3h4+63.*a22p3h2+63.*a22m5h2+322560.*a22mh8+30240.*a22mh4+6300.*a22ph2+6300.*a22m3h2+15120.*a22m3h6+15120.*a22ph6+19530.*a22mh2)/dr(1)**2*jac
 coeff(ma3( 1,-1, 0))=1./10080.*(-3052.*a12p1zz-3052.*a12zm1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2,-1, 0))=1./10080.*(665.*a12zm1z+665.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 3,-1, 0))=1./10080.*(-110.*a12zm1z-110.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 4,-1, 0))=1./10080.*(9.*a12zm1z+9.*a12p4zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-4, 0, 0))=1./322560.*(-63.*a11m3h2-225.*a11m7h2-225.*a11mh2-63.*a11m5h2)/dr(0)**2*jac
 coeff(ma3(-3, 0, 0))=1./322560.*(225.*a11ph2+630.*a11m3h2+1512.*a11m5h4+560.*a11m3h6+225.*a11m7h2+1512.*a11mh4+1764.*a11mh2+1764.*a11m5h2)/dr(0)**2*jac
 coeff(ma3(-2, 0, 0))=1./322560.*(-15120.*a11mh6-1512.*a11ph4-7560.*a11mh4-7560.*a11m3h4-1764.*a11ph2-1764.*a11m5h2-1512.*a11m5h4-6300.*a11mh2-6300.*a11m3h2-15120.*a11m3h6)/dr(0)**2*jac
 coeff(ma3(-1, 0, 0))=1./322560.*(63.*a11p3h2+19530.*a11mh2+15120.*a11m3h6+85680.*a11mh6+7560.*a11ph4+7560.*a11m3h4+322560.*a11mh8+15120.*a11ph6+63.*a11m5h2+30240.*a11mh4+6300.*a11ph2+6300.*a11m3h2)/dr(0)**2*jac
 coeff(ma3( 0, 0, 0))=(1./322560.*(-30240.*a11mh4-30240.*a11ph4-630.*a11m3h2-630.*a11p3h2-19530.*a11mh2-19530.*a11ph2-560.*a11p3h6-560.*a11m3h6-85680.*a11mh6-85680.*a11ph6-322560.*a11mh8-322560.*a11ph8)/dr(0)**2+1./322560.*(-85680.*a22mh6-85680.*a22ph6-630.*a22m3h2-630.*a22p3h2-322560.*a22mh8-322560.*a22ph8-30240.*a22mh4-30240.*a22ph4-560.*a22m3h6-19530.*a22mh2-19530.*a22ph2-560.*a22p3h6)/dr(1)**2+1./322560.*(-560.*a33m3h6-630.*a33m3h2-630.*a33p3h2-322560.*a33ph8-560.*a33p3h6-85680.*a33ph6-30240.*a33mh4-30240.*a33ph4-19530.*a33mh2-19530.*a33ph2-322560.*a33mh8-85680.*a33mh6)/dr(2)**2)*jac
 coeff(ma3( 1, 0, 0))=1./322560.*(15120.*a11mh6+85680.*a11ph6+7560.*a11mh4+7560.*a11p3h4+6300.*a11mh2+6300.*a11p3h2+63.*a11m3h2+19530.*a11ph2+15120.*a11p3h6+322560.*a11ph8+63.*a11p5h2+30240.*a11ph4)/dr(0)**2*jac
 coeff(ma3( 2, 0, 0))=1./322560.*(-6300.*a11ph2-6300.*a11p3h2-15120.*a11p3h6-1512.*a11mh4-1512.*a11p5h4-7560.*a11ph4-7560.*a11p3h4-1764.*a11mh2-1764.*a11p5h2-15120.*a11ph6)/dr(0)**2*jac
 coeff(ma3( 3, 0, 0))=1./322560.*(1512.*a11ph4+1764.*a11ph2+1764.*a11p5h2+225.*a11p7h2+225.*a11mh2+630.*a11p3h2+1512.*a11p5h4+560.*a11p3h6)/dr(0)**2*jac
 coeff(ma3( 4, 0, 0))=1./322560.*(-63.*a11p3h2-225.*a11p7h2-225.*a11ph2-63.*a11p5h2)/dr(0)**2*jac
 coeff(ma3(-4, 1, 0))=1./10080.*(9.*a12zp1z+9.*a12m4zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-3, 1, 0))=1./10080.*(-110.*a12zp1z-110.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-2, 1, 0))=1./10080.*(665.*a12zp1z+665.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 1, 0))=1./10080.*(-3052.*a12m1zz-3052.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 1, 0))=1./322560.*(322560.*a22ph8+63.*a22p5h2+30240.*a22ph4+6300.*a22mh2+6300.*a22p3h2+19530.*a22ph2+15120.*a22p3h6+85680.*a22ph6+63.*a22m3h2+7560.*a22mh4+7560.*a22p3h4+15120.*a22mh6)/dr(1)**2*jac
 coeff(ma3( 1, 1, 0))=1./10080.*(3052.*a12p1zz+3052.*a12zp1z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 1, 0))=1./10080.*(-665.*a12zp1z-665.*a12p2zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 3, 1, 0))=1./10080.*(110.*a12zp1z+110.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 4, 1, 0))=1./10080.*(-9.*a12zp1z-9.*a12p4zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-4, 2, 0))=0.
 coeff(ma3(-3, 2, 0))=1./10080.*(7.*a12zp2z+7.*a12m3zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-2, 2, 0))=1./10080.*(-91.*a12m2zz-91.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 2, 0))=1./10080.*(665.*a12m1zz+665.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 2, 0))=1./322560.*(-6300.*a22ph2-6300.*a22p3h2-7560.*a22ph4-1512.*a22p5h4-15120.*a22p3h6-1512.*a22mh4-15120.*a22ph6-7560.*a22p3h4-1764.*a22mh2-1764.*a22p5h2)/dr(1)**2*jac
 coeff(ma3( 1, 2, 0))=1./10080.*(-665.*a12p1zz-665.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 2, 0))=1./10080.*(91.*a12p2zz+91.*a12zp2z)/(dr(0)*dr(1))*jac
 coeff(ma3( 3, 2, 0))=1./10080.*(-7.*a12zp2z-7.*a12p3zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 4, 2, 0))=0.
 coeff(ma3(-4, 3, 0))=0.
 coeff(ma3(-3, 3, 0))=0.
 coeff(ma3(-2, 3, 0))=1./10080.*(7.*a12zp3z+7.*a12m2zz)/(dr(0)*dr(1))*jac
 coeff(ma3(-1, 3, 0))=1./10080.*(-110.*a12zp3z-110.*a12m1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 3, 0))=1./322560.*(1764.*a22ph2+1764.*a22p5h2+225.*a22p7h2+630.*a22p3h2+1512.*a22ph4+225.*a22mh2+560.*a22p3h6+1512.*a22p5h4)/dr(1)**2*jac
 coeff(ma3( 1, 3, 0))=1./10080.*(110.*a12p1zz+110.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 3, 0))=1./10080.*(-7.*a12p2zz-7.*a12zp3z)/(dr(0)*dr(1))*jac
 coeff(ma3( 3, 3, 0))=0.
 coeff(ma3( 4, 3, 0))=0.
 coeff(ma3(-4, 4, 0))=0.
 coeff(ma3(-3, 4, 0))=0.
 coeff(ma3(-2, 4, 0))=0.
 coeff(ma3(-1, 4, 0))=1./10080.*(9.*a12m1zz+9.*a12zp4z)/(dr(0)*dr(1))*jac
 coeff(ma3( 0, 4, 0))=1./322560.*(-225.*a22ph2-63.*a22p3h2-63.*a22p5h2-225.*a22p7h2)/dr(1)**2*jac
 coeff(ma3( 1, 4, 0))=1./10080.*(-9.*a12zp4z-9.*a12p1zz)/(dr(0)*dr(1))*jac
 coeff(ma3( 2, 4, 0))=0.
 coeff(ma3( 3, 4, 0))=0.
 coeff(ma3( 4, 4, 0))=0.
 coeff(ma3(-4,-4, 1))=0.
 coeff(ma3(-3,-4, 1))=0.
 coeff(ma3(-2,-4, 1))=0.
 coeff(ma3(-1,-4, 1))=0.
 coeff(ma3( 0,-4, 1))=1./10080.*(9.*a23zm4z+9.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-4, 1))=0.
 coeff(ma3( 2,-4, 1))=0.
 coeff(ma3( 3,-4, 1))=0.
 coeff(ma3( 4,-4, 1))=0.
 coeff(ma3(-4,-3, 1))=0.
 coeff(ma3(-3,-3, 1))=0.
 coeff(ma3(-2,-3, 1))=0.
 coeff(ma3(-1,-3, 1))=0.
 coeff(ma3( 0,-3, 1))=1./10080.*(-110.*a23zzp1-110.*a23zm3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3, 1))=0.
 coeff(ma3( 2,-3, 1))=0.
 coeff(ma3( 3,-3, 1))=0.
 coeff(ma3( 4,-3, 1))=0.
 coeff(ma3(-4,-2, 1))=0.
 coeff(ma3(-3,-2, 1))=0.
 coeff(ma3(-2,-2, 1))=0.
 coeff(ma3(-1,-2, 1))=0.
 coeff(ma3( 0,-2, 1))=1./10080.*(665.*a23zm2z+665.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 1))=0.
 coeff(ma3( 2,-2, 1))=0.
 coeff(ma3( 3,-2, 1))=0.
 coeff(ma3( 4,-2, 1))=0.
 coeff(ma3(-4,-1, 1))=0.
 coeff(ma3(-3,-1, 1))=0.
 coeff(ma3(-2,-1, 1))=0.
 coeff(ma3(-1,-1, 1))=0.
 coeff(ma3( 0,-1, 1))=1./10080.*(-3052.*a23zm1z-3052.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 1))=0.
 coeff(ma3( 2,-1, 1))=0.
 coeff(ma3( 3,-1, 1))=0.
 coeff(ma3( 4,-1, 1))=0.
 coeff(ma3(-4, 0, 1))=1./10080.*(9.*a13m4zz+9.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3(-3, 0, 1))=1./10080.*(-110.*a13zzp1-110.*a13m3zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0, 1))=1./10080.*(665.*a13zzp1+665.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 1))=1./10080.*(-3052.*a13m1zz-3052.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 1))=1./322560.*(63.*a33m3h2+15120.*a33mh6+85680.*a33ph6+7560.*a33mh4+7560.*a33p3h4+322560.*a33ph8+15120.*a33p3h6+63.*a33p5h2+30240.*a33ph4+6300.*a33mh2+6300.*a33p3h2+19530.*a33ph2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 1))=1./10080.*(3052.*a13p1zz+3052.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 1))=1./10080.*(-665.*a13p2zz-665.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0, 1))=1./10080.*(110.*a13p3zz+110.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3( 4, 0, 1))=1./10080.*(-9.*a13p4zz-9.*a13zzp1)/(dr(0)*dr(2))*jac
 coeff(ma3(-4, 1, 1))=0.
 coeff(ma3(-3, 1, 1))=0.
 coeff(ma3(-2, 1, 1))=0.
 coeff(ma3(-1, 1, 1))=0.
 coeff(ma3( 0, 1, 1))=1./10080.*(3052.*a23zp1z+3052.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 1))=0.
 coeff(ma3( 2, 1, 1))=0.
 coeff(ma3( 3, 1, 1))=0.
 coeff(ma3( 4, 1, 1))=0.
 coeff(ma3(-4, 2, 1))=0.
 coeff(ma3(-3, 2, 1))=0.
 coeff(ma3(-2, 2, 1))=0.
 coeff(ma3(-1, 2, 1))=0.
 coeff(ma3( 0, 2, 1))=1./10080.*(-665.*a23zzp1-665.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 1))=0.
 coeff(ma3( 2, 2, 1))=0.
 coeff(ma3( 3, 2, 1))=0.
 coeff(ma3( 4, 2, 1))=0.
 coeff(ma3(-4, 3, 1))=0.
 coeff(ma3(-3, 3, 1))=0.
 coeff(ma3(-2, 3, 1))=0.
 coeff(ma3(-1, 3, 1))=0.
 coeff(ma3( 0, 3, 1))=1./10080.*(110.*a23zp3z+110.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3, 1))=0.
 coeff(ma3( 2, 3, 1))=0.
 coeff(ma3( 3, 3, 1))=0.
 coeff(ma3( 4, 3, 1))=0.
 coeff(ma3(-4, 4, 1))=0.
 coeff(ma3(-3, 4, 1))=0.
 coeff(ma3(-2, 4, 1))=0.
 coeff(ma3(-1, 4, 1))=0.
 coeff(ma3( 0, 4, 1))=1./10080.*(-9.*a23zp4z-9.*a23zzp1)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 4, 1))=0.
 coeff(ma3( 2, 4, 1))=0.
 coeff(ma3( 3, 4, 1))=0.
 coeff(ma3( 4, 4, 1))=0.
 coeff(ma3(-4,-4, 2))=0.
 coeff(ma3(-3,-4, 2))=0.
 coeff(ma3(-2,-4, 2))=0.
 coeff(ma3(-1,-4, 2))=0.
 coeff(ma3( 0,-4, 2))=0.
 coeff(ma3( 1,-4, 2))=0.
 coeff(ma3( 2,-4, 2))=0.
 coeff(ma3( 3,-4, 2))=0.
 coeff(ma3( 4,-4, 2))=0.
 coeff(ma3(-4,-3, 2))=0.
 coeff(ma3(-3,-3, 2))=0.
 coeff(ma3(-2,-3, 2))=0.
 coeff(ma3(-1,-3, 2))=0.
 coeff(ma3( 0,-3, 2))=1./10080.*(7.*a23zzp2+7.*a23zm3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-3, 2))=0.
 coeff(ma3( 2,-3, 2))=0.
 coeff(ma3( 3,-3, 2))=0.
 coeff(ma3( 4,-3, 2))=0.
 coeff(ma3(-4,-2, 2))=0.
 coeff(ma3(-3,-2, 2))=0.
 coeff(ma3(-2,-2, 2))=0.
 coeff(ma3(-1,-2, 2))=0.
 coeff(ma3( 0,-2, 2))=1./10080.*(-91.*a23zzp2-91.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 2))=0.
 coeff(ma3( 2,-2, 2))=0.
 coeff(ma3( 3,-2, 2))=0.
 coeff(ma3( 4,-2, 2))=0.
 coeff(ma3(-4,-1, 2))=0.
 coeff(ma3(-3,-1, 2))=0.
 coeff(ma3(-2,-1, 2))=0.
 coeff(ma3(-1,-1, 2))=0.
 coeff(ma3( 0,-1, 2))=1./10080.*(665.*a23zm1z+665.*a23zzp2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 2))=0.
 coeff(ma3( 2,-1, 2))=0.
 coeff(ma3( 3,-1, 2))=0.
 coeff(ma3( 4,-1, 2))=0.
 coeff(ma3(-4, 0, 2))=0.
 coeff(ma3(-3, 0, 2))=1./10080.*(7.*a13m3zz+7.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3(-2, 0, 2))=1./10080.*(-91.*a13m2zz-91.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 2))=1./10080.*(665.*a13zzp2+665.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 2))=1./322560.*(-15120.*a33ph6-7560.*a33ph4-7560.*a33p3h4-1764.*a33mh2-1764.*a33p5h2-15120.*a33p3h6-1512.*a33mh4-6300.*a33ph2-6300.*a33p3h2-1512.*a33p5h4)/dr(2)**2*jac
 coeff(ma3( 1, 0, 2))=1./10080.*(-665.*a13zzp2-665.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 2))=1./10080.*(91.*a13p2zz+91.*a13zzp2)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0, 2))=1./10080.*(-7.*a13zzp2-7.*a13p3zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 4, 0, 2))=0.
 coeff(ma3(-4, 1, 2))=0.
 coeff(ma3(-3, 1, 2))=0.
 coeff(ma3(-2, 1, 2))=0.
 coeff(ma3(-1, 1, 2))=0.
 coeff(ma3( 0, 1, 2))=1./10080.*(-665.*a23zp1z-665.*a23zzp2)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 2))=0.
 coeff(ma3( 2, 1, 2))=0.
 coeff(ma3( 3, 1, 2))=0.
 coeff(ma3( 4, 1, 2))=0.
 coeff(ma3(-4, 2, 2))=0.
 coeff(ma3(-3, 2, 2))=0.
 coeff(ma3(-2, 2, 2))=0.
 coeff(ma3(-1, 2, 2))=0.
 coeff(ma3( 0, 2, 2))=1./10080.*(91.*a23zzp2+91.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 2))=0.
 coeff(ma3( 2, 2, 2))=0.
 coeff(ma3( 3, 2, 2))=0.
 coeff(ma3( 4, 2, 2))=0.
 coeff(ma3(-4, 3, 2))=0.
 coeff(ma3(-3, 3, 2))=0.
 coeff(ma3(-2, 3, 2))=0.
 coeff(ma3(-1, 3, 2))=0.
 coeff(ma3( 0, 3, 2))=1./10080.*(-7.*a23zzp2-7.*a23zp3z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 3, 2))=0.
 coeff(ma3( 2, 3, 2))=0.
 coeff(ma3( 3, 3, 2))=0.
 coeff(ma3( 4, 3, 2))=0.
 coeff(ma3(-4, 4, 2))=0.
 coeff(ma3(-3, 4, 2))=0.
 coeff(ma3(-2, 4, 2))=0.
 coeff(ma3(-1, 4, 2))=0.
 coeff(ma3( 0, 4, 2))=0.
 coeff(ma3( 1, 4, 2))=0.
 coeff(ma3( 2, 4, 2))=0.
 coeff(ma3( 3, 4, 2))=0.
 coeff(ma3( 4, 4, 2))=0.
 coeff(ma3(-4,-4, 3))=0.
 coeff(ma3(-3,-4, 3))=0.
 coeff(ma3(-2,-4, 3))=0.
 coeff(ma3(-1,-4, 3))=0.
 coeff(ma3( 0,-4, 3))=0.
 coeff(ma3( 1,-4, 3))=0.
 coeff(ma3( 2,-4, 3))=0.
 coeff(ma3( 3,-4, 3))=0.
 coeff(ma3( 4,-4, 3))=0.
 coeff(ma3(-4,-3, 3))=0.
 coeff(ma3(-3,-3, 3))=0.
 coeff(ma3(-2,-3, 3))=0.
 coeff(ma3(-1,-3, 3))=0.
 coeff(ma3( 0,-3, 3))=0.
 coeff(ma3( 1,-3, 3))=0.
 coeff(ma3( 2,-3, 3))=0.
 coeff(ma3( 3,-3, 3))=0.
 coeff(ma3( 4,-3, 3))=0.
 coeff(ma3(-4,-2, 3))=0.
 coeff(ma3(-3,-2, 3))=0.
 coeff(ma3(-2,-2, 3))=0.
 coeff(ma3(-1,-2, 3))=0.
 coeff(ma3( 0,-2, 3))=1./10080.*(7.*a23zzp3+7.*a23zm2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-2, 3))=0.
 coeff(ma3( 2,-2, 3))=0.
 coeff(ma3( 3,-2, 3))=0.
 coeff(ma3( 4,-2, 3))=0.
 coeff(ma3(-4,-1, 3))=0.
 coeff(ma3(-3,-1, 3))=0.
 coeff(ma3(-2,-1, 3))=0.
 coeff(ma3(-1,-1, 3))=0.
 coeff(ma3( 0,-1, 3))=1./10080.*(-110.*a23zm1z-110.*a23zzp3)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 3))=0.
 coeff(ma3( 2,-1, 3))=0.
 coeff(ma3( 3,-1, 3))=0.
 coeff(ma3( 4,-1, 3))=0.
 coeff(ma3(-4, 0, 3))=0.
 coeff(ma3(-3, 0, 3))=0.
 coeff(ma3(-2, 0, 3))=1./10080.*(7.*a13zzp3+7.*a13m2zz)/(dr(0)*dr(2))*jac
 coeff(ma3(-1, 0, 3))=1./10080.*(-110.*a13zzp3-110.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 3))=1./322560.*(225.*a33mh2+1764.*a33ph2+1764.*a33p5h2+630.*a33p3h2+1512.*a33ph4+560.*a33p3h6+1512.*a33p5h4+225.*a33p7h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 3))=1./10080.*(110.*a13zzp3+110.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 3))=1./10080.*(-7.*a13p2zz-7.*a13zzp3)/(dr(0)*dr(2))*jac
 coeff(ma3( 3, 0, 3))=0.
 coeff(ma3( 4, 0, 3))=0.
 coeff(ma3(-4, 1, 3))=0.
 coeff(ma3(-3, 1, 3))=0.
 coeff(ma3(-2, 1, 3))=0.
 coeff(ma3(-1, 1, 3))=0.
 coeff(ma3( 0, 1, 3))=1./10080.*(110.*a23zzp3+110.*a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 3))=0.
 coeff(ma3( 2, 1, 3))=0.
 coeff(ma3( 3, 1, 3))=0.
 coeff(ma3( 4, 1, 3))=0.
 coeff(ma3(-4, 2, 3))=0.
 coeff(ma3(-3, 2, 3))=0.
 coeff(ma3(-2, 2, 3))=0.
 coeff(ma3(-1, 2, 3))=0.
 coeff(ma3( 0, 2, 3))=1./10080.*(-7.*a23zzp3-7.*a23zp2z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 2, 3))=0.
 coeff(ma3( 2, 2, 3))=0.
 coeff(ma3( 3, 2, 3))=0.
 coeff(ma3( 4, 2, 3))=0.
 coeff(ma3(-4, 3, 3))=0.
 coeff(ma3(-3, 3, 3))=0.
 coeff(ma3(-2, 3, 3))=0.
 coeff(ma3(-1, 3, 3))=0.
 coeff(ma3( 0, 3, 3))=0.
 coeff(ma3( 1, 3, 3))=0.
 coeff(ma3( 2, 3, 3))=0.
 coeff(ma3( 3, 3, 3))=0.
 coeff(ma3( 4, 3, 3))=0.
 coeff(ma3(-4, 4, 3))=0.
 coeff(ma3(-3, 4, 3))=0.
 coeff(ma3(-2, 4, 3))=0.
 coeff(ma3(-1, 4, 3))=0.
 coeff(ma3( 0, 4, 3))=0.
 coeff(ma3( 1, 4, 3))=0.
 coeff(ma3( 2, 4, 3))=0.
 coeff(ma3( 3, 4, 3))=0.
 coeff(ma3( 4, 4, 3))=0.
 coeff(ma3(-4,-4, 4))=0.
 coeff(ma3(-3,-4, 4))=0.
 coeff(ma3(-2,-4, 4))=0.
 coeff(ma3(-1,-4, 4))=0.
 coeff(ma3( 0,-4, 4))=0.
 coeff(ma3( 1,-4, 4))=0.
 coeff(ma3( 2,-4, 4))=0.
 coeff(ma3( 3,-4, 4))=0.
 coeff(ma3( 4,-4, 4))=0.
 coeff(ma3(-4,-3, 4))=0.
 coeff(ma3(-3,-3, 4))=0.
 coeff(ma3(-2,-3, 4))=0.
 coeff(ma3(-1,-3, 4))=0.
 coeff(ma3( 0,-3, 4))=0.
 coeff(ma3( 1,-3, 4))=0.
 coeff(ma3( 2,-3, 4))=0.
 coeff(ma3( 3,-3, 4))=0.
 coeff(ma3( 4,-3, 4))=0.
 coeff(ma3(-4,-2, 4))=0.
 coeff(ma3(-3,-2, 4))=0.
 coeff(ma3(-2,-2, 4))=0.
 coeff(ma3(-1,-2, 4))=0.
 coeff(ma3( 0,-2, 4))=0.
 coeff(ma3( 1,-2, 4))=0.
 coeff(ma3( 2,-2, 4))=0.
 coeff(ma3( 3,-2, 4))=0.
 coeff(ma3( 4,-2, 4))=0.
 coeff(ma3(-4,-1, 4))=0.
 coeff(ma3(-3,-1, 4))=0.
 coeff(ma3(-2,-1, 4))=0.
 coeff(ma3(-1,-1, 4))=0.
 coeff(ma3( 0,-1, 4))=1./10080.*(9.*a23zzp4+9.*a23zm1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1,-1, 4))=0.
 coeff(ma3( 2,-1, 4))=0.
 coeff(ma3( 3,-1, 4))=0.
 coeff(ma3( 4,-1, 4))=0.
 coeff(ma3(-4, 0, 4))=0.
 coeff(ma3(-3, 0, 4))=0.
 coeff(ma3(-2, 0, 4))=0.
 coeff(ma3(-1, 0, 4))=1./10080.*(9.*a13zzp4+9.*a13m1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 0, 0, 4))=1./322560.*(-225.*a33ph2-63.*a33p3h2-63.*a33p5h2-225.*a33p7h2)/dr(2)**2*jac
 coeff(ma3( 1, 0, 4))=1./10080.*(-9.*a13zzp4-9.*a13p1zz)/(dr(0)*dr(2))*jac
 coeff(ma3( 2, 0, 4))=0.
 coeff(ma3( 3, 0, 4))=0.
 coeff(ma3( 4, 0, 4))=0.
 coeff(ma3(-4, 1, 4))=0.
 coeff(ma3(-3, 1, 4))=0.
 coeff(ma3(-2, 1, 4))=0.
 coeff(ma3(-1, 1, 4))=0.
 coeff(ma3( 0, 1, 4))=1./10080.*(-9.*a23zzp4-9.*a23zp1z)/(dr(1)*dr(2))*jac
 coeff(ma3( 1, 1, 4))=0.
 coeff(ma3( 2, 1, 4))=0.
 coeff(ma3( 3, 1, 4))=0.
 coeff(ma3( 4, 1, 4))=0.
 coeff(ma3(-4, 2, 4))=0.
 coeff(ma3(-3, 2, 4))=0.
 coeff(ma3(-2, 2, 4))=0.
 coeff(ma3(-1, 2, 4))=0.
 coeff(ma3( 0, 2, 4))=0.
 coeff(ma3( 1, 2, 4))=0.
 coeff(ma3( 2, 2, 4))=0.
 coeff(ma3( 3, 2, 4))=0.
 coeff(ma3( 4, 2, 4))=0.
 coeff(ma3(-4, 3, 4))=0.
 coeff(ma3(-3, 3, 4))=0.
 coeff(ma3(-2, 3, 4))=0.
 coeff(ma3(-1, 3, 4))=0.
 coeff(ma3( 0, 3, 4))=0.
 coeff(ma3( 1, 3, 4))=0.
 coeff(ma3( 2, 3, 4))=0.
 coeff(ma3( 3, 3, 4))=0.
 coeff(ma3( 4, 3, 4))=0.
 coeff(ma3(-4, 4, 4))=0.
 coeff(ma3(-3, 4, 4))=0.
 coeff(ma3(-2, 4, 4))=0.
 coeff(ma3(-1, 4, 4))=0.
 coeff(ma3( 0, 4, 4))=0.
 coeff(ma3( 1, 4, 4))=0.
 coeff(ma3( 2, 4, 4))=0.
 coeff(ma3( 3, 4, 4))=0.
 coeff(ma3( 4, 4, 4))=0.
#endMacro

