#include "Overture.h"
#include "SquareMapping.h"
#include "AnnulusMapping.h"
#include "Ogen.h"
#include "PlotStuff.h"

int 
main(int argc, char *argv[])
{
  Overture::start(argc,argv);  // initialize Overture

  printf(" -------------------------------------------------------------------------- \n");
  printf(" This example shows how to build an overlapping grid directly in a program, \n");
  printf(" rather than interactively.                                                 \n");
  printf(" -------------------------------------------------------------------------- \n");


  PlotStuff ps;               // for plotting
  PlotStuffParameters psp;

  // First build mappings
  SquareMapping map1(-1., 1., -1., 1.);        // Create a square
  map1.setGridDimensions(axis1,21); map1.setGridDimensions(axis2,21);

  AnnulusMapping map2;            // Create an Annulus
  map2.setRadii(.3,.6);
  map2.setGridDimensions(axis1,41); map2.setGridDimensions(axis2,7);
  // set outer boundary to be interpolation:
  map2.setBoundaryCondition(End,axis2,0);  

  // Put the mappings into a list
  MappingInformation mapInfo;
  mapInfo.graphXInterface=&ps;
  mapInfo.mappingList.addElement(map1);
  mapInfo.mappingList.addElement(map2);

  // Indicate which mappings should be used in the overlapping grid
  const int numberOfDimensions=2, numberOfGrids=2;
  IntegerArray mapList(numberOfGrids);
  mapList(0)=0;
  mapList(1)=1;

  // build an empty overlapping grid
  CompositeGrid cg;

  // Create an overlapping grid generator
  Ogen ogen(ps);

  // Put the mappings into the CompositeGrid
  ogen.buildACompositeGrid(cg,mapInfo,mapList);

  //  ogen.debug=3; // turn this on to show intermediate results

  // ** generate the overlapping grid**
  ogen.updateOverlap(cg);
  
  // Plot the overlapping grid.
  ps.erase();
  psp.set(GI_TOP_LABEL,"Grid after Ogen");  // set title
  PlotIt::plot(ps,cg,psp);                       // plot the grid


  Overture::finish();          
  return 0;
}
