*
* Circle in a channel (taking arguments)
*
*
* usage: ogen [noplot] cicArg -factor=<num> -order=[2/4/6/8] -interp=[e/i]
* 
* examples:
*     ogen noplot cicArg -factor=1 -order=2
*     ogen noplot cicArg -factor=1 -order=4
*     ogen noplot cicArg -factor=2 -order=4
*     ogen noplot cicArg -factor=2 -order=2 -interp=e    ( creates cice2.order2.hdf)
*     ogen noplot cicArg -factor=3 -order=2 -interp=e    ( creates cice3.order2.hdf)
*     ogen noplot cicArg -factor=4 -order=2 -interp=e    ( creates cice4.order2.hdf)
*     ogen noplot cicArg -factor=2 -order=4 -interp=e   
*     ogen noplot cicArg -factor=3 -order=4 -interp=e    ( creates cice4.order4.hdf)
*     ogen noplot cicArg -factor=12 -order=4 -interp=e   
* 
*     ogen noplot cicArg -factor=2 -order=2 -interp=e -xa=-4. -xb=4. -name="cice2L8.hdf"
*     ogen noplot cicArg -factor=4 -order=2 -interp=e -xa=-4. -xb=4. -name="cice4L8.hdf"
*
*
$order=2; $factor=1; $interp="i"; # default values
$orderOfAccuracy = "second order"; $ng=2; $interpType = "implicit for all grids";
$name=""; $xa=-2.; $xb=2.; $ya=-2.; $yb=2.; 
* 
* get command line arguments
GetOptions( "order=i"=>\$order,"factor=f"=> \$factor,"xa=f"=> \$xa,"xb=f"=> \$xb,"ya=f"=> \$ya,"yb=f"=> \$yb,\
            "interp=s"=> \$interp,"name=s"=> \$name);
* 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; }
* 
$suffix = ".order$order"; 
if( $name eq "" ){$name = "cic" . "$interp$factor" . $suffix . ".hdf";}
* 
$ds=.1/$factor;
* 
$dw = $order+1; $iw=$order+1; 
* parallel ghost lines: for ogen we need at least:
*       .5*( iw -1 )   : implicit interpolation 
*       .5*( iw+dw-2 ) : explicit interpolation
$parallelGhost=($iw-1)/2;
if( $interp eq "e" ){  $parallelGhost=($iw+$dw-2)/2; }
if( $parallelGhost<1 ){ $parallelGhost=1; } 
minimum number of distributed ghost lines
  $parallelGhost
create mappings
*
rectangle
  set corners
    $xa $xb $ya $yb
  lines
    $nx = int( ($xb-$xa)/$ds +1.5 ); 
    $ny = int( ($yb-$ya)/$ds +1.5 ); 
    $nx $ny
  boundary conditions
    1 1 1 1
  mappingName
  square
exit
*
Annulus
  $nr = 5+$ng;
  $innerRad=.5; $outerRad = $innerRad + ($nr-1)*$ds;
  inner and outer radii
    $innerRad $outerRad
  lines
    $nTheta = int( 2.*3.1415*($innerRad+$outerRad)*.5/$ds + 1.5 );
    $nTheta $nr
  boundary conditions
    -1 -1 1 0
exit
*
exit
generate an overlapping grid
    square
    Annulus
  done
  change parameters
    * choose implicit or explicit interpolation
    interpolation type
      $interpType
    order of accuracy 
      $orderOfAccuracy
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
  exit
*  display intermediate results
  compute overlap
**  display computed geometry
  exit
*
* save an overlapping grid
save a grid (compressed)
$name
cic
exit

