*
* Create the initial grid for a deforming disk. 
* Use this grid with cgmp for a fluid-structure example.
*
* Usage:
*         ogen [noplot] diskDeform [options]
* where options are
*     -factor=<num>     : grid spacing is .1 divided by this factor
*     -interp=[e/i]     : implicit or explicit interpolation
*     -name=<string>    : over-ride the default name  
*     -case=[inner|outer] : only build a grid for the inner or outer domain
*     -nExtra          : add extra lines in the normal direction on the boundary fitted grids
*
* Examples:
*
*      ogen noplot diskDeform -factor=.5
*      ogen noplot diskDeform -factor=1
*      ogen noplot diskDeform -factor=2
*      ogen noplot diskDeform -factor=1 -interp=e
*      ogen noplot diskDeform -factor=2 -interp=e
* Bigger outer domain: 
*   ogen noplot diskDeform -factor=2 -interp=e -xa=-2.5 -xb=2.5 -ya=-2.5 -yb=2.5 -name="diskDeformBig2e.hdf"
*   ogen noplot diskDeform -factor=4 -interp=e -xa=-2.5 -xb=2.5 -ya=-2.5 -yb=2.5 -name="diskDeformBig4e.hdf"
*   ogen noplot diskDeform -factor=8 -interp=e -xa=-2.5 -xb=2.5 -ya=-2.5 -yb=2.5 -nExtra=5 -name="diskDeformBig8e.hdf"
*   ogen noplot diskDeform -factor=16 -interp=e -xa=-2.5 -xb=2.5 -ya=-2.5 -yb=2.5 -nExtra=5 -name="diskDeformBig16e.hdf"
* 
* outer-domain only:
*     ogen noplot diskDeform -case=outer -factor=1 -interp=e
*
*
$factor=1; $name=""; $case=""; 
$interp="i"; $interpType = "implicit for all grids"; 
$order=2; $orderOfAccuracy = "second order"; $ng=2; 
$xa=-2.; $xb=2.; $ya=-2.; $yb=2.; $nExtra=0; 
* 
* get command line arguments
GetOptions("name=s"=> \$name,"order=i"=>\$order,"factor=f"=> \$factor,"interp=s"=> \$interp,"case=s"=> \$case,\
           "xa=f"=> \$xa,"xb=f"=> \$xb,"ya=f"=> \$ya,"yb=f"=> \$yb,"nExtra=i"=>\$nExtra);
* 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; }
if( $name eq "" ){$name = "diskDeform$case" . "$interp$factor" . ".hdf";}
*
$bcInterface0=100;  # bc for interfaces
$bcInterface1=101;  
$shareInterface=100;        # share value for interfaces
*
$Pi=4.*atan2(1.,1.);
*
$ds0 = .1; 
* target grid spacing:
$ds = $ds0/$factor;
*
    $width=.25; if( $factor<1. ){ $width = .25*$factor; }
    $rad=1.; 
* 
create mappings
*
  rectangle
    $nx=int( ($xb-$xa)/$ds+1.5 ); 
    $ny=int( ($yb-$ya)/$ds+1.5 ); 
    set corners
      $xa $xb $ya $yb 
    lines
      $nx $ny 
    mappingName
      outerSquare
    exit
*
  rectangle
    * make the inner square bigger for deforming grid problems
    * $xb = $rad-$width/$factor + 2.*$ds ;
    $xb = $rad + .5;
    $xa=-$xb; $ya=$xa; $yb=$xb; 
    $nx=int( ($xb-$xa)/$ds+1.5 ); 
    $ny=int( ($yb-$ya)/$ds+1.5 ); 
    set corners
      $xa $xb $ya $yb 
    lines
      $nx $ny 
    boundary conditions
      0 0 0 0 
    mappingName
      innerSquare
    exit
*
* Create a start curve for the interface
*
    $innerRadius=$rad; $outerRadius=$innerRadius + $width/$factor; 
    $averageRadius=($innerRadius+$outerRadius)/2.;
* 
  spline
    $n=51;
    enter spline points
      $n 
    $x0=0.; $y0=1.*$rad;
    $commands="";
    for( $i=0; $i<$n; $i++ ){ $theta=2.*$Pi*$i/($n-1.); $x0=$rad*cos($theta); $y0=$rad*sin($theta); \
                              $commands = $commands . "$x0 $y0\n"; }
    $commands
    lines
      $nr = int( 2.*$Pi*$rad/$ds+1.5 );
      $nr
    periodicity
      2
    * pause
    exit
*  
   $width = $width + $ds0*$nExtra;
* 
  hyperbolic
    * add a few extra points as the boundary deforms it gets longer
    $stretchFactor=1.25; 
    $dist = $width/$factor;     
    $ns = int( $width/$ds0 +1.5 );  if( $ns<3 ){ $ns=3; }
    distance to march $dist 
    lines to march $ns
    points on initial curve $nr
    generate
    mapping parameters
    Boundary Condition: bottom  $bcInterface0
    Share Value: bottom $shareInterface
    close mapping dialog
    name outerInterface
    * pause
    exit
*  
  hyperbolic
    * add a few extra points as the boundary deforms it gets longer
    $stretchFactor=1.25; 
    $dist = $width/$factor;     
    $ns = int( $width/$ds0 +1.5 ); if( $ns<3 ){ $ns=3; }
    backward
    distance to march $dist 
    lines to march $ns
    points on initial curve $nr
    generate
    mapping parameters
    Boundary Condition: bottom  $bcInterface0
    Share Value: bottom $shareInterface
    close mapping dialog
    name innerInterface
    exit
*
*
  exit this menu
*
generate an overlapping grid
if( $case eq "inner" ){ $gridList="innerSquare\n innerInterface"; }\
elsif( $case eq "outer" ){ $gridList="outerSquare\n outerInterface"; }\
else{ $gridList="outerSquare\n outerInterface\n innerSquare\n innerInterface";  }
  $gridList
  done choosing mappings
* 
  change parameters 
    * define the domains -- these will behave like independent overlapping grids
   if( $case eq "" ){ $cmd="specify a domain\n outerDomain\n outerSquare\n outerInterface\n done";}else{ $cmd="*"; }
      $cmd
   if( $case eq "" ){ $cmd="specify a domain\n innerDomain\n innerSquare\n innerInterface\n done";}else{ $cmd="*"; }
      $cmd
    order of accuracy
     $orderOfAccuracy
    interpolation type
      $interpType
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
    exit 
*
  compute overlap
*
exit
*
save an overlapping grid
  $name
  diskDeform
exit
