*
* "Square in a circle" (taking arguments)
*
*
* usage: ogen [noplot] sicArg -factor=<num> -order=[2/4/6/8] -interp=[e/i]
* 
* examples:
*     ogen noplot sicArg -factor=1 -order=2
*     ogen noplot sicArg -factor=2 -order=2 -interp=e    ( creates sice2.order2.hdf)
*     ogen noplot sicArg -factor=3 -order=2 -interp=e    ( creates sice3.order2.hdf)
*     ogen noplot sicArg -factor=4 -order=2 -interp=e    ( creates sice4.order2.hdf)
*     ogen noplot sicArg -factor=5 -order=2 -interp=e    ( creates sice5.order2.hdf)
*
*     ogen noplot sicArg -factor=1 -order=4
*     ogen noplot sicArg -factor=2 -order=4
*     ogen noplot sicArg -factor=2 -order=4 -interp=e
*     ogen noplot sicArg -factor=4 -order=4 -interp=e
*     ogen noplot sicArg -factor=8 -order=4 -interp=e
*
$order=2; $factor=1; $interp="i"; # default values
$orderOfAccuracy = "second order"; $ng=2; $interpType = "implicit for all grids";
* 
* get command line arguments
GetOptions( "order=i"=>\$order,"factor=i"=> \$factor,"interp=s"=> \$interp);
* 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; }
* 
$suffix = ".order$order"; 
$name = "sic" . "$interp$factor" . $suffix . ".hdf";
* 
$ds=.1/$factor;
* 
* Here is the radius of the circular boundary:
$outerRad=1.;
* 
create mappings
*
Annulus
  * keep the number of radial points on the annulus fixed:
  $nr = 5+$ng;
  $innerRad= $outerRad - ($nr-1)*$ds;
  inner and outer radii
    $innerRad $outerRad
  lines
    $nTheta = int( 2.*3.1415*($innerRad+$outerRad)*.5/$ds + 1.5 );
    $nTheta $nr
  boundary conditions
    -1 -1 0 1 
  mappingName
   Annulus
exit
*
rectangle
  set corners
    $xb =  $innerRad+($ng-1)*$ds;
    $xa = -$xb;
    $xa $xb $xa $xb
  lines
    $nx = int( ($xb-$xa)/$ds +1.5 ); $ny=$nx; 
    $nx $ny
  boundary conditions
    0 0 0 0
  mappingName
    square
exit
*
exit
generate an overlapping grid
    square
    Annulus
  done
  change parameters
    * choose implicit or explicit interpolation
    interpolation type
      $interpType
    order of accuracy 
      $orderOfAccuracy
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
  exit
*  display intermediate results
  compute overlap
  exit
*
save an overlapping grid
$name
sic
exit
