*
*  Solid sphere in a box (two-domain problem)
*
* usage: ogen [noplot] solidSphereInABox -factor=<num> -order=[2/4/6/8] -interp=[e/i] -nrExtra=<> -rgd=[fixed|var]
*
*  nrExtra: extra lines to add in the radial direction on the sphere grids 
*  -rgd : var=variable : decrease radial grid distance as grids are refined. fixed=fix radial grid distance
* 
* examples:
*     ogen noplot solidSphereInABox -factor=1 -order=2
*     ogen noplot solidSphereInABox -factor=1 -order=4
*     ogen noplot solidSphereInABox -factor=2 -order=4 -interp=e -nrExtra=8    (for cgmx : add extra grid lines in the normal direction)
* 
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=1 
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=2 
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=4 
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=8 
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=16
*
*  ** fixed radial distance (for convergence tests)
*     ogen noplot solidSphereInABox -order=2 -factor=1 -rgd=fixed  (backup)
*     ogen noplot solidSphereInABox -order=2 -factor=2 -rgd=fixed
*     ogen noplot solidSphereInABox -order=2 -factor=4 -rgd=fixed
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=2 -rgd=fixed  (explicit interp starts with factor=2)
*     ogen noplot solidSphereInABox -order=2 -interp=e -factor=4 -rgd=fixed  
* 
$xa=-2.; $xb=2.; $ya=-2.; $yb=2.; $za=-2.; $zb=2.; $nrExtra=2; $rgd="var";
$order=2; $factor=1; $interp="i"; # default values
$orderOfAccuracy = "second order"; $ng=2; $interpType = "implicit for all grids"; $dse=0.; 
* 
* get command line arguments
GetOptions( "order=i"=>\$order,"factor=i"=> \$factor,"nrExtra=i"=> \$nrExtra,"interp=s"=> \$interp,"rgd=s"=> \$rgd);
* 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; $dse=1.; }
* 
$prefix="solidSphereInABox";
if( $rgd eq "fixed" ){ $prefix = $prefix . "Fixed"; }
$suffix = ".order$order"; 
$name = $prefix . "$interp$factor" . $suffix . ".hdf";
* 
$ds=.15/$factor;
#
# Here is the radial width of the spherical grids -- this will be fixed if rgd=fixed
$deltaRadius0=.3; # do not make larger than .3 or troubles with cgmx
# matching interface grids should be given distinct share values for now for cgmx -- fix me -- cgmp is different
$interfaceShareSouthPole=100;  
$interfaceShareNorthPole=101;
* 
* ---------------------------------------
* turn off graphics
* ---------------------------------------
*
create mappings
*
*   ******** outer-sphere ********
* 
Sphere
  $nr=3+$order; if( $interp eq "e" ){ $nr=$nr+$order; } 
  $innerRad=1.; $outerRad=$innerRad+($nr-1)*$ds;
  if( $rgd eq "fixed" ){ $outerRad = $innerRad + $deltaRadius0; $nr = int( $deltaRadius0/$ds+3.5); }
  $nr=$nr + $nrExtra; 
  inner and outer radii
    $innerRad $outerRad
 mappingName
  outerSphere
exit
*
* now make a mapping for the north pole
*
reparameterize
  transform which mapping?
    outerSphere
  orthographic
    * sa=2 --> patches just match (not including ghost points)
    $sa = 2. + $order*$dse*$ds + ($order-2)*$ds*.5; $sb=$sa; 
    specify sa,sb
      $sa $sb
  exit
  lines
    $nTheta=int( 3.2*($innerRad+$outerRad)*.5/$ds +1.5 );    
    $nTheta $nTheta $nr
*    15 15 5
  boundary conditions
    0 0 0 0 1 0
  share
    0 0 0 0 $interfaceShareNorthPole 0
  mappingName
    outerNorthPole
exit
*
* now make a mapping for the south pole
*
reparameterize
  transform which mapping?
    outerSphere
  orthographic
    choose north or south pole
      -1
    specify sa,sb
      $sa $sb
  exit
  lines
    $nTheta $nTheta $nr
  boundary conditions
    0 0 0 0 1 0
  share
    0 0 0 0 $interfaceShareSouthPole 0
  mappingName
    outerSouthPole
exit
*
* Here is the box
*
Box
  set corners
    $xa $xb $ya $yb $za $zb
  lines
    $nx = int( ($xb-$xa)/$ds +1.5);
    $ny = int( ($yb-$ya)/$ds +1.5);
    $nz = int( ($zb-$za)/$ds +1.5);
    $nx $ny $nz
  mappingName
    outerBox
  exit
*
*   ******** inner-sphere ********
* 
Sphere
  $nr=3+$order; if( $interp eq "e" ){ $nr=$nr+$order; } 
  $outerRad=1.; $innerRad=$outerRad-($nr-1)*$ds;
  if( $rgd eq "fixed" ){ $innerRad = $outerRad - $deltaRadius0; $nr = int( $deltaRadius0/$ds+3.5); }
  $nr=$nr + $nrExtra; 
  inner and outer radii
    $innerRad $outerRad
 mappingName
  innerSphere
exit
*
* now make a mapping for the north pole
*
reparameterize
  transform which mapping?
    innerSphere
  orthographic
    specify sa,sb
      $sa $sb
  exit
  lines
    $nTheta $nTheta $nr
  boundary conditions
    0 0 0 0 0 1
  share
    0 0 0 0 0 $interfaceShareNorthPole 
  mappingName
    innerNorthPole
exit
*
* now make a mapping for the south pole
*
reparameterize
  transform which mapping?
    innerSphere
  orthographic
    choose north or south pole
      -1
    specify sa,sb
      $sa $sb
  exit
  lines
    $nTheta $nTheta $nr
  boundary conditions
    0 0 0 0 0 1
  share
    0 0 0 0 0 $interfaceShareSouthPole 
  mappingName
    innerSouthPole
exit
*
* Here is the box
*
Box
  set corners
    * $xa = -( $innerRad + ($order-2)*$ds + $order*$dse );
    $xa = -( $innerRad + ($order-2 + $dse*1.5 )*$ds );
    $xb=-$xa;  $ya=$xa; $yb=$xb; $za=$xa; $zb=$xb; 
    $xa $xb $ya $yb $za $zb
  lines
    $nx = int( ($xb-$xa)/$ds +1.5);
    $ny = int( ($yb-$ya)/$ds +1.5);
    $nz = int( ($zb-$za)/$ds +1.5);
    $nx $ny $nz
  boundary conditions
    0 0 0 0 0 0
  mappingName
    innerBox
  exit
***********************************
exit
*
generate an overlapping grid
  outerBox
  outerNorthPole
  outerSouthPole
* 
  innerBox
  innerNorthPole
  innerSouthPole
  done
* 
  change parameters
    specify a domain
      * domain name:
      outerDomain 
      * grids in the domain:
      outerBox
      outerNorthPole
      outerSouthPole
      done
    specify a domain
      * domain name:
      innerDomain 
      * grids in the domain:
      innerBox
      innerNorthPole
      innerSouthPole
      done
* 
    interpolation type
      $interpType
    order of accuracy 
      $orderOfAccuracy
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
  exit
*
  compute overlap
*
exit
* save an overlapping grid
save a grid (compressed)
$name
sib
exit
