#
#  Ogen: Two spheres in a channel 
#
# usage: ogen [noplot] twoSpheresInAChannel -factor=<num> -order=[2/4/6/8] -interp=[e/i] -nrExtra=<>
#
#  nrExtra: extra lines to add in the radial direction on the sphere grids 
# 
# examples:
#     ogen noplot twoSpheresInAChannel -factor=1 -order=2
#     ogen noplot twoSpheresInAChannel -factor=1 -order=4
#     ogen noplot twoSpheresInAChannel -factor=2 -order=4 -interp=e -nrExtra=8    (for cgmx : add extra grid lines in the normal direction)
# 
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=1 
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=2 
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=4 -xb=3.
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=8 
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=16
#
# -- bigger domain
#     ogen noplot twoSpheresInAChannel -order=2 -interp=e -factor=4 -xa=-2.5 -xb=7.5 -ya=-2.5 -yb=2.5 -za=-2.5 -zb=2.5 
#
$xa=-1.5; $xb=1.5; $ya=-1.5; $yb=1.5; $za=-1.5; $zb=1.5; $nrExtra=3; $loadBalance=0;
$order=2; $factor=1; $interp="i"; # default values
$orderOfAccuracy = "second order"; $ng=2; $interpType = "implicit for all grids"; $dse=0.; 
# 
# get command line arguments
GetOptions( "order=i"=>\$order,"factor=i"=> \$factor,"nrExtra=i"=> \$nrExtra,"interp=s"=> \$interp,\
            "loadBalance=i"=>\$loadBalance,"xa=f"=> \$xa,"xb=f"=> \$xb,"ya=f"=> \$ya,"yb=f"=> \$yb,\
            "za=f"=> \$za,"zb=f"=> \$zb);
# 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; $dse=1.; }
# 
$suffix = ".order$order"; 
$name = "twoSpheresInAChannel" . "$interp$factor" . $suffix . ".hdf";
# 
$ds=.1/$factor;
$pi=4.*atan2(1.,1.);
# 
create mappings
# 
# add extra points for stretching 
$nr=3+ $nrExtra +$order; if( $interp eq "e" ){ $nr=$nr+$order; } 
#
$gridNames="*"; 
$numberOfSpheres=0;  $sphereShare=1; $sphereStretchb=10.; 
$sphereRadius=.5; 
#
# sphere 1: 
# 
# $xSphere=-.5; $ySphere=-.4; $zSphere=-.4; 
$xSphere=-.6; $ySphere=-.6; $zSphere=-.6; 
include sphere.h
#
# sphere 2: 
# $xSphere= .5; $ySphere= .4; $zSphere= .4; 
$xSphere= .6; $ySphere= .6; $zSphere= .6; 
include sphere.h
#
# Here is the box
#
Box
  set corners
    $xa $xb $ya $yb $za $zb
  lines
    $nx = int( ($xb-$xa)/$ds +1.5);
    $ny = int( ($yb-$ya)/$ds +1.5);
    $nz = int( ($zb-$za)/$ds +1.5);
    $nx $ny $nz
  mappingName
    box
  exit
exit
#
generate an overlapping grid
  box
  $gridNames
#  sphere1
#  northPole1
#  southPole1
  done
  change parameters
    * improve quality of interpolation
    interpolation type
      $interpType
    order of accuracy 
      $orderOfAccuracy
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
  exit
# change the plot
# open graphics
 compute overlap
exit
* save an overlapping grid
save a grid (compressed)
$name
twoSpheresInAChannel
exit




#
# make a second sphere
#
Sphere
  inner and outer radii
    .5 .7
  centre for sphere
    +.5 +.4  .4  +.5 +.5  .5
  mappingName
    sphere2
exit
#
# now make a mapping for the north pole
#
reparameterize
  transform which mapping?
    sphere2
  orthographic
    specify sa,sb
      2. 2. 2.5 2.5
  exit
  lines
    65 65 17 33 33 9  15 15 5
  boundary conditions
    0 0 0 0 1 0
  share
    0 0 0 0 2 0
  mappingName
    unstretched-north-pole2
exit
#
# now make a mapping for the south pole
#
reparameterize
  transform which mapping?
    sphere2
  orthographic
    choose north or south pole
      -1
    specify sa,sb
      2. 2. 2.5 2.5
  exit
  lines
    65 65 17 33 33 9  15 15 5
  boundary conditions
    0 0 0 0 1 0
  share
    0 0 0 0 2 0
  mappingName
    unstretched-south-pole2
exit
#
  stretch coordinates
    transform which mapping?
      unstretched-north-pole1
    Stretch r3:itanh
    STP:stretch r3 itanh: layer 0 1 10 0 (id>=0,weight,exponent,position)
    stretch grid
    name north-pole1
    exit
#
#
  stretch coordinates
    transform which mapping?
      unstretched-south-pole1
    Stretch r3:itanh
    STP:stretch r3 itanh: layer 0 1 10 0 (id>=0,weight,exponent,position)
    stretch grid
    name south-pole1
    exit
#
#
  stretch coordinates
    transform which mapping?
      unstretched-north-pole2
    Stretch r3:itanh
    STP:stretch r3 itanh: layer 0 1 10 0 (id>=0,weight,exponent,position)
    stretch grid
    name north-pole2
    exit
#
#
  stretch coordinates
    transform which mapping?
      unstretched-south-pole2
    Stretch r3:itanh
    STP:stretch r3 itanh: layer 0 1 10 0 (id>=0,weight,exponent,position)
    stretch grid
    name south-pole2
    exit
#
# Here is the box
#
Box
  specify corners
    -2 -2 -2 4 2 2 
  lines
    145 97 97   73 49 49     97 65 65     51 51   21 21 21  31 31 31
  mappingName
    box
  exit
#
# pause
exit
#
generate an overlapping grid
  box
  north-pole2
  south-pole2
  north-pole1
  south-pole1
 done
  change parameters
 # improve quality
    ghost points
      all
      2 2 2 2 2 2
  exit
 # pause
 compute overlap
#  change the plot
#
# pause
exit
save an overlapping grid
twoSpheresInAChannel.hdf
twoSpheresInAChannel
exit



